/****************************************************************************
*																			*
*					Certificate Attribute Management Routines				*
*						Copyright Peter Gutmann 1996-1998					*
*																			*
****************************************************************************/

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#if defined( INC_ALL ) ||  defined( INC_CHILD )
  #include "asn1.h"
  #include "asn1oid.h"
  #include "cert.h"
  #include "certattr.h"
#else
  #include "keymgmt/asn1.h"
  #include "keymgmt/asn1oid.h"
  #include "keymgmt/cert.h"
  #include "keymgmt/certattr.h"
#endif /* Compiler-specific includes */

/* Prototypes for functions in lib_cert.c */

int textToOID( const char *oid, const int oidLength, BYTE *binaryOID );

/****************************************************************************
*																			*
*						Attribute Type Checking and Mapping					*
*																			*
****************************************************************************/

/* Get the attribute information for a given OID */

ATTRIBUTE_INFO *oidToAttribute( const ATTRIBUTE_TYPE attributeType,
								const BYTE *oid )
	{
	ATTRIBUTE_INFO *attributeInfoPtr = ( attributeType == ATTRIBUTE_CMS ) ? \
		( ATTRIBUTE_INFO * ) cmsAttributeInfo : ( ATTRIBUTE_INFO * ) extensionInfo;
	const int length = sizeofOID( oid );

	while( attributeInfoPtr->fieldID != CRYPT_ERROR )
		{
		if( attributeInfoPtr->oid != NULL && \
			sizeofOID( attributeInfoPtr->oid ) == length && \
			!memcmp( attributeInfoPtr->oid, oid, length ) )
			return( attributeInfoPtr );
		attributeInfoPtr++;
		}

	/* It's an unknown attribute */
	return( NULL );
	}

/* Get the attribute and attributeID for a field ID */

static const ATTRIBUTE_INFO *fieldIDToAttribute( const ATTRIBUTE_TYPE attributeType,
		const CRYPT_CERTINFO_TYPE fieldID, const CRYPT_CERTINFO_TYPE subFieldID,
		CRYPT_CERTINFO_TYPE *attributeID )
	{
	const ATTRIBUTE_INFO *attributeInfoPtr = \
		( attributeType == ATTRIBUTE_CMS ) ? cmsAttributeInfo : extensionInfo;
	int i;

	/* Find the information on this attribute field */
	for( i = 0; attributeInfoPtr[ i ].fieldID != CRYPT_ERROR; i++ )
		{
		/* If the previous entry doesn't have more data following it, the
		   current entry is the start of a complete attribute and therefore
		   contains the attribute ID */
		if( attributeID != NULL && \
			( !i || !( attributeInfoPtr[ i - 1 ].flags & FL_MORE ) ) )
			{
			int j = i;

			/* Usually the attribute ID is the fieldID for the first entry,
			   however in some cases the attributeID is the same as the
			   fieldID and isn't specified until later on (denoted by the
			   fieldID being FIELDID_FOLLOWS), so we have to look ahead to
			   find it */
			*attributeID = attributeInfoPtr[ i ].fieldID;
			for( j = i + 1; *attributeID == FIELDID_FOLLOWS; j++ )
				*attributeID = attributeInfoPtr[ j ].fieldID;
			}

		/* Check whether the field ID for this entry matches the one we want */
		if( attributeInfoPtr[ i ].fieldID == fieldID )
			{
			/* If we're after a subfield match as well, try and match the
			   subfield */
			if( subFieldID != CRYPT_CERTINFO_NONE )
				{
				for( i = 0; generalNameInfo[ i ].fieldID != CRYPT_ERROR; i++ )
					if( generalNameInfo[ i ].fieldID == subFieldID )
						return( &generalNameInfo[ i ] );

				return( NULL );
				}

			return( &attributeInfoPtr[ i ] );
			}
		}

	return( NULL );
	}

/* Validate and preprocess a set of attributes in preparation for writing
   them to a certificate or CRL and set up links to the information in the
   attribute information table prior to encoding the attributes.  This is a
   rather complex process which relies on stepping through the list of
   attribute fields and the attribute information table in sync and making
   sure the list of fields is consistent with the attribute information
   table.  In addition we set up sync points between the list and table which
   are used during the encoding process.  For example assume we have the
   following attribute:

	attribute ::= SEQUENCE {
		foo		BOOLEAN DEFAULT TRUE,
		bar		SEQUENCE OF OBJECT IDENTIFIER
		}

   The attribute information table would encode this attribute as:

	t1:	OID	SEQUENCE	MORE
	t2:		BOOLEAN		MORE	OPTIONAL
	t3:		SEQUENCE	MORE
	t4:		OID

   The first table entry t1 contains the OID, the SEQUENCE wrapper, and a
   continuation flag.  For the purposes of comparison with the list, this is
   a nop and can be skipped since it's only used for encoding purposes.  The
   next table entry t2 contains the first attribute field, an optional
   boolean and a continuation flag.  The next table entry t3 contains another
   SEQUENCE wrapper which again is only used for encoding and can be skipped
   for comparing with the list, and a continuation flag.  Finally, the last
   table entry t4 contains the second attribute field, an OID.

   Assuming the attribute list contains the following:

	BOOLEAN	FALSE	-> t1
	OID		xxx		-> t3

   The attribute validation process sets the sync point for the first
   attribute list entry to point to t1, and the second one to point to t3.
   When we encode the attribute, we encode t1 (the OID, critical flag, and
   SEQUENCE wrapper), since the field ID's won't match we step to t2 and use
   that to encode the boolean.  We then do the same for t3 with the SEQUENCE
   and OID.

   If the attribute list instead contained only:

	OID		xxx		-> t1

   then this time the attribute validation process sets the sync point to t1.
   When encoding we encode t1 as before, step to t2, the field ID's won't
   match but t2 is optional so we skip it, then encode t3 as for t1 and
   finally encode the OID using t4.

   At this point we also evaluate the encoded size of each attribute.  For
   invidual fields, we just store their encoded size.  For constructed
   objects, we stack the attribute list entry where the constructed object
   starts and, until we reach the end of the constructed object, accumulate
   the total size of the fields which make up the object.  When we reach the
   end of the object, we unstack the pointer to the attribute list and store
   the total size in it.

   To handle nested constructed objects, we only update the size of the
   topmost item on the stack.  When this is unstacked, we add the size of
   that entry, plus the size of its tag and length information, to the next
   entry on the stack.

   In addition to updating the size, we also record the sequence of table
   entries which are required to encode the constructed item.  A worst-case
   sequence of entries would be:

	SEQUENCE {
		SEQUENCE OPTIONAL { ... }		| Not encoded
		SEQUENCE {
			SEQUENCE OPTIONAL { ... }	| Not encoded
			SEQUENCE {
				value
				}
			}
		}

   which contains an alternating sequence of encoded and non-encoded fields.
   Because of this, the validation check performs the complex task of
   recording which table entries are used for the encoding by stacking and
   unstacking them and discarding the ones which evaluate to a zero size
   during the unstacking process.

   Each entry in the stack contains the list item it applies to, the table
   entry which is used to encode the stacked item, and the size of the item */

#define ATTRIBUTE_STACKSIZE		10

typedef struct {
	ATTRIBUTE_LIST *attributeListPtr;	/* List entry that this applies to */
	ATTRIBUTE_INFO *attributeInfoPtr;	/* Encoding point for sequence */
	int size;							/* Size of sequence */
	} ATTRIBUTE_STACK;

#include <assert.h>					/*!!!!!!!!*/

/* Once we reach the end of the constructed item, we need to unwind the stack
   and update everything we've gone past.  If it's an optional item (so that
   nothing gets encoded), we don't do anything */

static void updateStackedInfo( ATTRIBUTE_STACK *stack, int *stackPosPtr,
							   int count )
	{
	int stackPos = *stackPosPtr;

assert( count <= stackPos );		/*!!!!!!!!*/

	while( count-- )
		{
		ATTRIBUTE_LIST *attributeFifoPtr = stack[ --stackPos ].attributeListPtr;
		const ATTRIBUTE_INFO *attributeInfoPtr = stack[ stackPos ].attributeInfoPtr;
		const int size = stack[ stackPos ].size;

assert( attributeInfoPtr != NULL );	/*!!!!!!!!*/
assert( size >= 0 );				/*!!!!!!!!*/
assert( count >= 0 );				/*!!!!!!!!*/
assert( stackPos >= -1 );			/*!!!!!!!!*/

		/* If there's something to encode, update the attribute list item.
		   There's a special case here where even if the sequence is of zero
		   length, if there's a member with a default value present (resulting
		   in nothing being encoded) we still have to encode a zero-length
		   sequence.

		   In addition sometimes we can reach the end of the attribute list
		   but there are further actions defined in the encoding table (for
		   example cleanup actions in nested sequences).  In this case the
		   stacked attributeFifoPtr is NULL and the size is zero, so we
		   perform an additional check to make sure the pointer is non-null */
		if( size || ( attributeFifoPtr != NULL && \
					  attributeFifoPtr->isDefaultValue ) )
			{
assert( attributeFifoPtr != NULL );	/*!!!!!!!!*/

			/* Remember the size and table entry used to encode this stack
			   entry */
			attributeFifoPtr->sizeFifo[ attributeFifoPtr->fifoEnd ] = size;
			attributeFifoPtr->encodingFifo[ attributeFifoPtr->fifoEnd++ ] = \
										stack[ stackPos ].attributeInfoPtr;

			/* If there are further items on the stack, percolate the size
			   information up to the previous entry */
			if( stackPos )
				{
				/* If it's a non-constructed field which is part of an
				   attributeTypeAndValue, or a field not used to encode user
				   data, add the length of the type and value fields,
				   otherwise add the length of the encapsulated data */
				if( attributeInfoPtr->fieldType == FIELDTYPE_IDENTIFIER )
					stack[ stackPos - 1 ].size += \
							sizeofOID( attributeInfoPtr->oid ) + size;
				else
					if( attributeInfoPtr->flags & FL_NONENCODING )
						stack[ stackPos - 1 ].size += \
							sizeofShortInteger( attributeInfoPtr->defaultValue ) + size;
					else
						stack[ stackPos - 1 ].size += ( int ) sizeofObject( size );
				}
			}
		}

	*stackPosPtr = stackPos;
	}

/* Some attributes contain a sequence of items of the attributeTypeAndValue
   form (ie OID, ANY DEFINED BY OID).  To process these, a check is made to
   determine whether the named value component in the attribute list is
   present in the current attributeTypeAndValue definition.  If it isn't, the
   item is given a zero length which means it's never encoded since the field
   is marked as optional.  The following function checks whether a named
   value component is present in the item */

static BOOLEAN checkComponentPresent( const CRYPT_CERTINFO_TYPE fieldID,
									  ATTRIBUTE_INFO **attributeInfoPtrPtr )
	{
	ATTRIBUTE_INFO *attributeInfoPtr = *attributeInfoPtrPtr;
	int nestLevel = 0;

	/* Check each field we find until we find the end of the
	   attributeTypeAndValue */
	while( TRUE )
		{
		/* Adjust the nesting level depending on whether we're entering or
		   leaving a sequence */
		if( attributeInfoPtr->fieldType == BER_SEQUENCE )
			nestLevel++;
		nestLevel -= decodeNestingLevel( attributeInfoPtr->flags );

		/* If the field is present in this attributeTypeAndValue, return */
		if( attributeInfoPtr->fieldID == fieldID )
			return( TRUE );

		/* If we're at the end of the attribute or the attributeTypeAndValue,
		   exit the loop before adjusting the attributeInfoPtr so that we're
		   still pointing at the end-of-attribute field */
		if( nestLevel <= 0 || !( attributeInfoPtr->flags & FL_MORE ) )
			break;

		attributeInfoPtr++;
		}

	/* The field isn't present, update the pointer to the next
	   attributeTypeAndValue or the end of the attribute */
	*attributeInfoPtrPtr = attributeInfoPtr;
	return( FALSE );
	}

/* State machine for checking a CHOICE.  When we get to the start of a
   CHOICE, we move from CHOICE_NONE to CHOICE_START.  Once we've checked one
   of the CHOICE options, we move to CHOICE_DONE.  If a further option is
   found in the CHOICE_DONE state, we record an error.  This is a somewhat
   crude mechanism which works because the only CHOICE fields which can't be
   handled by rewriting them as alternative representations are complete
   attributes so that the CHOICE applies over the entire attribute.  If a
   CHOICE is ever present as an attribute subfield, the checking would be
   handled by recursively checking it a la GeneralName */

typedef enum { CHOICE_NONE, CHOICE_START, CHOICE_DONE } CHOICE_STATE;

/* Check an entry in the attribute table.  While we're performing the check
   we need to pass a lot of state information around, this is contained in
   the following structure */

typedef struct {
	/* State information.  When we're encoding a GeneralName (using the
	   alternative encoding table), we need to remember the field ID of
	   the GeneralName to both tell the encoding routines that we're
	   using the GN encoding table, and to remember the overall GN ID so
	   we don't treat two adjacent GN subfields as though they were part
	   of the same GN.  If we're not currently encoding a GN, this field
	   is set to CRYPT_CERTINFO_NONE */
	ATTRIBUTE_LIST *attributeListPtr;	/* Position in attribute list */
	ATTRIBUTE_INFO *attributeInfoPtr;	/* Position in attribute table */
	CRYPT_CERTINFO_TYPE generalNameType;/* Field ID of GN being processed */
	CHOICE_STATE choiceState;			/* State of CHOICE processing */

	/* Encoding stack.  When we're encoding subfields, the stack contains
	   items from both the subfield and the encapsulating field, so we also
	   record the current stack top to make sure we don't go past this level
	   when popping items after we've finished encoding a subfield */
	ATTRIBUTE_STACK *stack;			/* Encoding stack */
	int stackPos;					/* Encoding stack position */
	int stackTop;

	/* Error information */
	CRYPT_CERTINFO_TYPE errorLocus;	/* Error locus */
	CRYPT_CERTERROR_TYPE errorType;	/* Error type */
	} ATTRIBUTE_CHECK_INFO;

static int checkAttribute( ATTRIBUTE_CHECK_INFO *attributeCheckInfo );

static int checkAttributeEntry( ATTRIBUTE_CHECK_INFO *attributeCheckInfo )
	{
	ATTRIBUTE_LIST *attributeListPtr = attributeCheckInfo->attributeListPtr;
	ATTRIBUTE_INFO *attributeInfoPtr = attributeCheckInfo->attributeInfoPtr;
	ATTRIBUTE_STACK *stack = attributeCheckInfo->stack;
	CRYPT_CERTINFO_TYPE fieldID;

	/* Determine the fieldID for the current attribute field */
	if( attributeListPtr == NULL )
		/* If we've reached the end of the list of attributes, use a non-ID
		   which doesn't match any table entry */
		fieldID = CRYPT_UNUSED;
	else
		/* If we're encoding a GeneralName, the fieldID is the field ID
		   within the GeneralName, or the subFieldID */
		if( attributeCheckInfo->generalNameType == attributeListPtr->fieldID )
			fieldID = attributeListPtr->subFieldID;
		else
			/* It's a standard attribute field */
			fieldID = attributeListPtr->fieldID;

	/* If the field in the attribute list matches the one in the table,
	   process it and move on to the next one */
	if( attributeListPtr != NULL && attributeInfoPtr->fieldID == fieldID )
		{
		/* If it's a GeneralName field, check the GeneralName components
		   using their own encoding table */
		if( attributeInfoPtr->fieldType == FIELDTYPE_GENERALNAME )
			{
			int status;

			/* Switch to the GeneralName encoding table and record the fact
			   that we've done this, and set the new stack top to the level at
			   which we start encoding the GeneralName */
			attributeCheckInfo->attributeInfoPtr = ( ATTRIBUTE_INFO * ) generalNameInfo;
			attributeCheckInfo->generalNameType = attributeListPtr->fieldID;
			attributeCheckInfo->stackTop = attributeCheckInfo->stackPos;
			status = checkAttribute( attributeCheckInfo );
			attributeCheckInfo->attributeInfoPtr = attributeInfoPtr;
			attributeCheckInfo->generalNameType = CRYPT_CERTINFO_NONE;
            attributeCheckInfo->stackTop = 0;
			return( status );
			}

		/* If there's an extended validation function attached to this field,
		   call it */
		if( attributeInfoPtr->validationFunction != NULL )
			{
			attributeCheckInfo->errorType = \
				attributeInfoPtr->validationFunction( attributeListPtr );
			if( attributeCheckInfo->errorType != CRYPT_CERTERROR_NONE )
				return( CRYPT_INVALID );
			}

		/* If this is an optional field and the value is the same as the
		   default value, remember that it doesn't get encoded */
		if( ( attributeInfoPtr->flags & FL_DEFAULT ) && \
			( attributeInfoPtr->defaultValue == attributeListPtr->value ) )
			{
			attributeListPtr->isDefaultValue = TRUE;
			attributeCheckInfo->attributeListPtr = attributeListPtr->next;
			return( CRYPT_OK );
			}

		/* Remember the encoded size of this field */
		attributeListPtr->attributeInfoPtr = attributeInfoPtr;
		attributeListPtr->encodedSize = writeAttributeField( NULL, attributeListPtr );
		if( attributeCheckInfo->stackPos )
			stack[ attributeCheckInfo->stackPos - 1 ].size += attributeListPtr->encodedSize;

		/* If this is a CHOICE field, update the choice state */
		if( attributeCheckInfo->choiceState != CHOICE_NONE )
			{
			if( attributeCheckInfo->choiceState == CHOICE_DONE )
				{
				/* If we've already processed one of the CHOICE options, there
				   can't be another one present */
				attributeCheckInfo->errorType = CRYPT_CERTERROR_PRESENT;
				return( CRYPT_INVALID );
				}
			if( attributeCheckInfo->choiceState == CHOICE_START )
				/* Remember that we've seen a CHOICE option */
				attributeCheckInfo->choiceState = CHOICE_DONE;
			}

		/* Move on to the next attribute field */
		attributeCheckInfo->attributeListPtr = attributeListPtr->next;
		return( CRYPT_OK );
		}

	/* If it's an attributeTypeAndValue sequence, check whether it contains
	   the field we want */
	if( attributeInfoPtr->flags & FL_IDENTIFIER )
		{
		if( checkComponentPresent( fieldID, &attributeInfoPtr ) )
			{
			/* Stack the position of the sequence start and the following
			   OID */
			stack[ attributeCheckInfo->stackPos ].size = 0;
			stack[ attributeCheckInfo->stackPos ].attributeListPtr = attributeListPtr;
			stack[ attributeCheckInfo->stackPos++ ].attributeInfoPtr = attributeInfoPtr++;
			stack[ attributeCheckInfo->stackPos ].size = 0;
			stack[ attributeCheckInfo->stackPos ].attributeListPtr = attributeListPtr;
			stack[ attributeCheckInfo->stackPos++ ].attributeInfoPtr = attributeInfoPtr;
			}

		/* Since we've jumped over several items we may be pointing at an
		   end-of-sequence flag for which no sequence start was stacked, so
		   we skip the stack update step */
		attributeCheckInfo->attributeInfoPtr = attributeInfoPtr;
		return( OK_SPECIAL );
		}

	/* If it's a sequence or a non-encoding value then it's a nop entry used
	   only for encoding purposes and can be skipped, however we need to
	   remember it for later encoding */
	if( attributeInfoPtr->fieldType == BER_SEQUENCE || \
		attributeInfoPtr->fieldType == FIELDTYPE_CHOICE || \
		attributeInfoPtr->flags & FL_NONENCODING )
		{
		/* If it's the start of a CHOICE, remember this */
		if( attributeInfoPtr->fieldType == FIELDTYPE_CHOICE )
			attributeCheckInfo->choiceState = CHOICE_START;

		/* Stack the sequence or value start position in the attribute list */
		stack[ attributeCheckInfo->stackPos ].size = 0;
		stack[ attributeCheckInfo->stackPos ].attributeListPtr = attributeListPtr;
		stack[ attributeCheckInfo->stackPos++ ].attributeInfoPtr = attributeInfoPtr;
		return( CRYPT_OK );
		}

	/* If it's a non-optional field and the attribute field doesn't match,
	   it's an error - attribute attributeID is missing field
	   attributeInfoPtr->fieldID (optional subfield
	   attributeInfoPtr->subFieldID) */
	if( !( attributeInfoPtr->flags & FL_OPTIONAL ) )
		{
		attributeCheckInfo->errorType = CRYPT_CERTERROR_ABSENT;
		return( CRYPT_NOTINITED );
		}

	return( CRYPT_OK );
	}

/* Check an individual attribute */

static int checkAttribute( ATTRIBUTE_CHECK_INFO *attributeCheckInfo )
	{
	BOOLEAN attributeContinues;

	/* Step through the attribute comparing the fields which are present in
	   the attribute list with the fields which should be present according
	   to the table, and set encoding sync points as required */
	do
		{
		int status;

		/* Check the current encoding table entry */
		status = checkAttributeEntry( attributeCheckInfo );
		if( cryptStatusError( status ) )
			{
			attributeCheckInfo->errorLocus = attributeCheckInfo->attributeInfoPtr->fieldID;
			return( status );
			}

		/* If this is the end of a constructed item, unstack it and update
		   the attribute list entry with the length information.  If it's a
		   sequence with all fields optional (so that nothing gets encoded),
		   we don't do anything */
		if( status != OK_SPECIAL )
			updateStackedInfo( attributeCheckInfo->stack,
				&attributeCheckInfo->stackPos,
				decodeNestingLevel( attributeCheckInfo->attributeInfoPtr->flags ) );

		/* Move on to the next table entry.  We have to check the
		   continuation flag before we move to the next table entry in order
		   to include processing of the last field in an attribute */
		attributeContinues = ( attributeCheckInfo->attributeInfoPtr->flags & FL_MORE ) ? TRUE : FALSE;
		attributeCheckInfo->attributeInfoPtr++;
		}
	while( attributeContinues );
	attributeCheckInfo->choiceState = CHOICE_NONE;

	/* We've reached the end of the attribute, if there are still constructed
	   objects stacked, unstack them and update their length information.  If
	   it's a sequence with all fields optional (so that nothing gets
	   encoded), we don't do anything */
	updateStackedInfo( attributeCheckInfo->stack,
					   &attributeCheckInfo->stackPos,
					   attributeCheckInfo->stackPos - attributeCheckInfo->stackTop );

	return( CRYPT_OK );
	}

/* Check the entire list of attributes */

int checkAttributes( const ATTRIBUTE_TYPE attributeType,
					 const ATTRIBUTE_LIST *listHeadPtr,
					 CRYPT_CERTINFO_TYPE *errorLocus,
					 CRYPT_CERTERROR_TYPE *errorType )
	{
	ATTRIBUTE_CHECK_INFO attributeCheckInfo;
	const ATTRIBUTE_INFO *attributeInfoStartPtr = \
		( attributeType == ATTRIBUTE_CMS ) ? cmsAttributeInfo : extensionInfo;
	ATTRIBUTE_LIST *attributeListPtr;
	ATTRIBUTE_STACK stack[ ATTRIBUTE_STACKSIZE ];

	/* If we've already done a validation pass, some of the fields will
	   contain values which were previously set, so before we begin we walk
	   down the last resetting the fields which are updated by this
	   function */
	for( attributeListPtr = ( ATTRIBUTE_LIST * ) listHeadPtr;
		 attributeListPtr != NULL && attributeListPtr->fieldID != CRYPT_CERTINFO_NONE;
		 attributeListPtr = attributeListPtr->next )
		{
		attributeListPtr->attributeInfoPtr = NULL;
		attributeListPtr->encodedSize = attributeListPtr->fifoPos = \
			attributeListPtr->fifoEnd = 0;
		attributeListPtr->isDefaultValue = FALSE;
		}

	/* Set up the attribute-checking state information */
	memset( &attributeCheckInfo, 0, sizeof( ATTRIBUTE_CHECK_INFO ) );
	attributeCheckInfo.attributeListPtr = ( ATTRIBUTE_LIST * ) listHeadPtr;
	attributeCheckInfo.attributeInfoPtr = ( ATTRIBUTE_INFO * ) attributeInfoStartPtr;
	attributeCheckInfo.stack = stack;

	/* Walk down the list of known attributes checking each one for
	   consistency */
	while( attributeCheckInfo.attributeListPtr != NULL && \
		   attributeCheckInfo.attributeListPtr->fieldID != CRYPT_CERTINFO_NONE )
		{
		int status;

		/* Find the start of this attribute in the attribute info table and
		   remember it as an encoding sync point.  Comparing the field ID
		   with the attribute ID is usually valid because the attribute info
		   table always begins the series of entries for an attribute with
		   the attribute ID.  The one exception is where the attribute ID is
		   the same as the field ID but they're separate entries in the
		   table, in which case the first entries will contain a
		   FIELDID_FOLLOWS code to indicate that a following field contains
		   the attribute/fieldID */
		while( attributeCheckInfo.attributeInfoPtr->fieldID != \
			   attributeCheckInfo.attributeListPtr->attributeID )
			attributeCheckInfo.attributeInfoPtr++;
		while( attributeCheckInfo.attributeInfoPtr != attributeInfoStartPtr && \
			   attributeCheckInfo.attributeInfoPtr[ -1 ].fieldID == FIELDID_FOLLOWS )
			attributeCheckInfo.attributeInfoPtr--;

		/* Check this attribute */
		status = checkAttribute( &attributeCheckInfo );
		if( cryptStatusError( status ) )
			{
			*errorLocus = attributeCheckInfo.errorLocus;
			*errorType = attributeCheckInfo.errorType;
			return( status );
			}
		}

	return( CRYPT_OK );
	}

/****************************************************************************
*																			*
*					Attribute Location/Cursor Movement Routines				*
*																			*
****************************************************************************/

/* Find the start and end of an attribute from a field within the
   attribute */

static ATTRIBUTE_LIST *findAttributeStart( const ATTRIBUTE_LIST *attributeListPtr )
	{
	ATTRIBUTE_LIST *attributeListCursor = ( ATTRIBUTE_LIST * ) attributeListPtr;
	CRYPT_CERTINFO_TYPE attributeID;

	if( attributeListCursor == NULL )
		return( NULL );
	attributeID = attributeListCursor->attributeID;

	/* Move backwards until we find the start of the attribute */
	while( attributeListCursor->prev != NULL && \
		   attributeListCursor->prev->attributeID == attributeID )
		attributeListCursor = attributeListCursor->prev;

	return( attributeListCursor );
	}

static ATTRIBUTE_LIST *findAttributeEnd( ATTRIBUTE_LIST *attributeListPtr )
	{
	CRYPT_CERTINFO_TYPE attributeID;

	if( attributeListPtr == NULL )
		return( NULL );
	attributeID = attributeListPtr->attributeID;

	/* Move forwards until we find the start of the next attribute */
	while( attributeListPtr->next != NULL && \
		   attributeListPtr->next->attributeID == attributeID )
		attributeListPtr = attributeListPtr->next;

	return( attributeListPtr );
	}

/* Find an attribute in a list of certificate attributes by object identifier
   (for blob-type attributes) or by field and subfield ID (for known
   attributes), with extended handling for fields with default values */

ATTRIBUTE_LIST *findAttributeByOID( const ATTRIBUTE_LIST *listHead,
									const BYTE *oid )
	{
	ATTRIBUTE_LIST *attributeListPtr;

	/* Find the position of this component in the list */
	for( attributeListPtr = ( ATTRIBUTE_LIST * ) listHead;
		 attributeListPtr != NULL; attributeListPtr = attributeListPtr->next )
		if( sizeofOID( attributeListPtr->oid ) == sizeofOID( oid ) && \
			!memcmp( attributeListPtr->oid, oid, sizeofOID( oid ) ) )
			break;

	return( attributeListPtr );
	}

ATTRIBUTE_LIST *findAttributeField( const ATTRIBUTE_LIST *listHead,
									const CRYPT_CERTINFO_TYPE fieldID,
									const CRYPT_CERTINFO_TYPE subFieldID )
	{
	ATTRIBUTE_LIST *attributeListPtr;

	/* Find the position of this component in the list */
	for( attributeListPtr = ( ATTRIBUTE_LIST * ) listHead;
		 attributeListPtr != NULL && attributeListPtr->fieldID != fieldID;
		 attributeListPtr = attributeListPtr->next );
	if( subFieldID == CRYPT_CERTINFO_NONE )
		return( attributeListPtr );

	/* Find the subfield in the field */
	while( attributeListPtr != NULL && \
		   attributeListPtr->fieldID == fieldID )
		{
		if( attributeListPtr->subFieldID == subFieldID )
			return( attributeListPtr );
		attributeListPtr = attributeListPtr->next;
		}

	return( NULL );
	}

ATTRIBUTE_LIST *findAttributeFieldEx( const ATTRIBUTE_LIST *listHead,
									  const CRYPT_CERTINFO_TYPE fieldID )
	{
	const ATTRIBUTE_TYPE attributeType = ( fieldID >= CRYPT_FIRST_CMS ) ? \
									ATTRIBUTE_CMS : ATTRIBUTE_CERTIFICATE;
	ATTRIBUTE_LIST *attributeListPtr;

	/* Find the position of this component in the list */
	for( attributeListPtr = ( ATTRIBUTE_LIST * ) listHead;
		 attributeListPtr != NULL && attributeListPtr->fieldID != fieldID;
		 attributeListPtr = attributeListPtr->next );

	/* If the field isn't present in the list of attributes, check whether
	   the attribute itself is present and whether this field has a default
	   value */
	if( attributeListPtr == NULL )
		{
		static const ATTRIBUTE_LIST defaultField = { 0, CRYPT_ERROR, 0 };
		static const ATTRIBUTE_LIST completeAttribute = { CRYPT_ERROR, 0, 0 };
		CRYPT_CERTINFO_TYPE attributeID;
		const ATTRIBUTE_INFO *attributeInfoPtr = fieldIDToAttribute( attributeType,
								fieldID, CRYPT_CERTINFO_NONE, &attributeID );
		ATTRIBUTE_LIST *attributeListSearchPtr;

		/* If there's no attribute containing this field, exit */
		if( attributeInfoPtr == NULL )
			return( NULL );

		/* Check whether any part of the attribute which contains the given
		   field is present in the list of attribute fields */
		for( attributeListSearchPtr = ( ATTRIBUTE_LIST * ) listHead;
			 attributeListSearchPtr != NULL && \
				attributeListSearchPtr->attributeID != attributeID;
			 attributeListSearchPtr = attributeListSearchPtr->next );
		if( attributeListSearchPtr == NULL )
			return( NULL );

		/* Some other part of the attribute containing the given field is
		   present in the list.  If this field wasn't found that could either
		   be a default value (in which case we return an entry which denotes
		   that this field is absent but has a default setting) or a field
		   which denotes an entire constructed attribute (in which case we
		   return an entry which denotes this) */
		if( attributeInfoPtr->flags & FL_DEFAULT )
			return( ( ATTRIBUTE_LIST * ) &defaultField );
		if( attributeInfoPtr->fieldType == BER_SEQUENCE )
			return( ( ATTRIBUTE_LIST * ) &completeAttribute );
		}

	return( attributeListPtr );
	}

ATTRIBUTE_LIST *findAttribute( const ATTRIBUTE_LIST *listHead,
							   const CRYPT_CERTINFO_TYPE fieldID )
	{
	ATTRIBUTE_LIST *attributeListPtr;

	/* Finding an overall attribute is a bit more complex than finding a
	   field since we may be given an attribute ID (which denotes the entire
	   attribute) or a field ID (which denotes one field in the attribute).
	   First we check for a match on the attribute ID */
	for( attributeListPtr = ( ATTRIBUTE_LIST * ) listHead;
		 attributeListPtr != NULL && attributeListPtr->attributeID != fieldID;
		 attributeListPtr = attributeListPtr->next );
	if( attributeListPtr != NULL )
		return( attributeListPtr );

	/* It wasn't a (present) attribute ID, try again looking for a field
	   within an attribute */
	attributeListPtr = findAttributeField( listHead, fieldID,
										   CRYPT_CERTINFO_NONE );
	if( attributeListPtr == NULL )
		return( NULL );

	/* Find the start of the attribute which contains this field */
	return( findAttributeStart( attributeListPtr ) );
	}

/* Get the default value for an optional field of an attribute */

int getDefaultFieldValue( const CRYPT_CERTINFO_TYPE fieldID )
	{
	const ATTRIBUTE_TYPE attributeType = ( fieldID >= CRYPT_FIRST_CMS ) ? \
									ATTRIBUTE_CMS : ATTRIBUTE_CERTIFICATE;
	const ATTRIBUTE_INFO *attributeInfoPtr = fieldIDToAttribute( attributeType,
										fieldID, CRYPT_CERTINFO_NONE, NULL );

	return( ( int ) attributeInfoPtr->defaultValue );
	}

/* Determine whether an attribute field is valid for a given certificate
   object type */

BOOLEAN isValidField( const CRYPT_CERTINFO_TYPE fieldID,
					  const CRYPT_CERTTYPE_TYPE certType )
	{
	const ATTRIBUTE_TYPE attributeType = \
					( certType == CRYPT_CERTTYPE_CMS_ATTRIBUTES ) ? \
					ATTRIBUTE_CMS : ATTRIBUTE_CERTIFICATE;
	CRYPT_CERTINFO_TYPE attributeID;
	ATTRIBUTE_INFO *attributeInfoPtr;
	const int certTypeFlag = ( certType == CRYPT_CERTTYPE_CERTIFICATE ) ? \
					   FL_VALID_CERT : ( certType == CRYPT_CERTTYPE_CRL ) ? \
					   FL_VALID_CRL : ( certType == CRYPT_CERTTYPE_CERTREQUEST ) ? \
					   FL_VALID_CERTREQ : 0;

	/* Find the attribute this field is a part of and check whether it's
	   allowed for this cert object */
	if( fieldIDToAttribute( attributeType, fieldID, CRYPT_CERTINFO_NONE,
							&attributeID ) == NULL )
		return( FALSE );
	attributeInfoPtr = ( ATTRIBUTE_INFO * ) fieldIDToAttribute( attributeType,
									attributeID, CRYPT_CERTINFO_NONE, NULL );
	return( ( attributeInfoPtr->flags & certTypeFlag ) ? TRUE : FALSE );
	}

/* Move the attribute cursor relative to the current cursor position.  This
   moves as far as possible in the direction given and then returns an
   appropriate return code, either CRYPT_OK or CRYPT_DATA_NOTFOUND if no
   movement is possible.  Note that it's possible to both return
   CRYPT_DATA_NOTFOUND and update the cursor, for example if we're halfway
   through the first attribute then a position code of CRYPT_CURSOR_FIRST
   will move to the start of the attribute, but the error will be returned to
   indicate that it wasn't possible to move back before the current
   attribute */

int moveAttributeCursor( ATTRIBUTE_LIST **currentCursor,
						 const BOOLEAN moveByField, const int position )
	{
	CRYPT_CERTINFO_TYPE attributeID = ( *currentCursor )->attributeID;
	ATTRIBUTE_LIST *newCursor = *currentCursor, *lastCursor = NULL;
	int count;

	/* Set the amount we want to move by based on the position code.  This
	   means we can handle the movement in a simple while loop instead of
	   having to special-case it for moves by one item */
	count = ( position == CRYPT_CURSOR_FIRST || \
			  position == CRYPT_CURSOR_LAST ) ? INT_MAX : 1;

	/* Moving by field is relatively simple, just move backwards or forwards
	   until we either run out of fields or the next field belongs to a
	   different attribute */
	if( moveByField )
		{
		if( position == CRYPT_CURSOR_FIRST || position == CRYPT_CURSOR_PREVIOUS )
			while( count-- && newCursor->prev != NULL && \
				   newCursor->prev->attributeID == attributeID )
				newCursor = newCursor->prev;
		else
			while( count-- && newCursor->next != NULL && \
				   newCursor->next->attributeID == attributeID )
				newCursor = newCursor->next;

		if( *currentCursor == newCursor )
			return( CRYPT_DATA_NOTFOUND );
		*currentCursor = newCursor;
		return( CRYPT_OK );
		}

	/* Moving by attribute is a bit more complex.  First we find the start or
	   end of the current attribute, if there's another attribute before or
	   after it move back/forwards into that attribute and find the start of
	   that.  This has the effect of moving us from anywhere in the current
	   attribute to the start of the preceding or following attribute.
	   Finally, we repeat this as required */
	if( position == CRYPT_CURSOR_FIRST || position == CRYPT_CURSOR_PREVIOUS )
		while( count-- && newCursor != NULL )
			{
			lastCursor = newCursor;
			newCursor = findAttributeStart( findAttributeStart( newCursor )->prev );
			}
	else
		while( count-- && newCursor != NULL && \
			   newCursor->attributeID != CRYPT_CERTINFO_NONE )
			{
			lastCursor = newCursor;
			newCursor = findAttributeEnd( newCursor )->next;
			}

	/* We've gone as far as we can.  If the new cursor is NULL, we've reached
	   the start or end of the attribute list, in case we haven't moved at all
	   we call findAttributeStart() to move to the first field.  If the new
	   cursor isn't NULL, it's located on the next or previous attribute */
	if( newCursor == NULL )
		*currentCursor = findAttributeStart( lastCursor );
	else
		*currentCursor = newCursor;
	if( *currentCursor == NULL )
		/* Positioning in null attribute lists is always unsuccessful */
		return( CRYPT_DATA_NOTFOUND );
	if( position == CRYPT_CURSOR_FIRST || position == CRYPT_CURSOR_LAST )
		/* Absolute positioning is always successful */
		return( CRYPT_OK );
	return( ( ( *currentCursor )->attributeID == attributeID ) ? \
			CRYPT_DATA_NOTFOUND : CRYPT_OK );
	}

/****************************************************************************
*																			*
*							Attribute Management Routines					*
*																			*
****************************************************************************/

/* Insert an element into an attribute list after the given insertion point */

static void insertListElements( ATTRIBUTE_LIST **listHeadPtr,
								ATTRIBUTE_LIST *insertPoint,
								ATTRIBUTE_LIST *newStartElement,
								ATTRIBUTE_LIST *newEndElement )
	{
	/* If it's an empty list, make this the new list */
	if( *listHeadPtr == NULL )
		{
		*listHeadPtr = newStartElement;
		return;
		}

	/* If we're inserting at the start of the list, make this the new first
	   element */
	if( insertPoint == NULL )
		{
		/* Insert the element at the start of the list */
		newEndElement->next = *listHeadPtr;
		( *listHeadPtr )->prev = newEndElement;
		*listHeadPtr = newStartElement;
		return;
		}

	/* Insert the element in the middle or end of the list.  Update the links
	   for the next element */
	newEndElement->next = insertPoint->next;
	if( insertPoint->next != NULL )
		insertPoint->next->prev = newEndElement;

	/* Update the links for the previous element */
	insertPoint->next = newStartElement;
	newStartElement->prev = insertPoint;
	}

/* Add a blob-type attribute to a list of attributes */

int addAttribute( const ATTRIBUTE_TYPE attributeType,
				  ATTRIBUTE_LIST **listHeadPtr, const BYTE *oid,
				  const BOOLEAN criticalFlag, const void *data,
				  const int dataLength )
	{
	ATTRIBUTE_LIST *newElement, *insertPoint;

	/* If this attribute type is already handled as a non-blob attribute,
	   don't allow it to be added as a blob as well.  This avoids problems
	   with the same attribute being added twice, once as a blob and once as
	   a non-blob.  In addition it forces the caller to use the (recommended)
	   normal attribute handling mechanism which allows for proper type
	   checking */
	if( oidToAttribute( attributeType, oid ) != NULL )
		return( CRYPT_NOPERM );

	/* Find the correct place in the list to insert the new element */
	if( *listHeadPtr != NULL )
		{
		ATTRIBUTE_LIST *prevElement = NULL;

		for( insertPoint = *listHeadPtr; insertPoint != NULL;
			 insertPoint = insertPoint->next )
			{
			/* Make sure this blob attribute isn't already present */
			if( insertPoint->attributeID != CRYPT_CERTINFO_NONE && \
				sizeofOID( insertPoint->oid ) == sizeofOID( oid ) && \
				!memcmp( insertPoint->oid, oid, sizeofOID( oid ) ) )
				return( CRYPT_INITED );

			prevElement = insertPoint;
			}
		insertPoint = prevElement;
		}

	/* Allocate memory for the new element and copy the information across */
	if( ( newElement  = ( ATTRIBUTE_LIST * ) malloc( sizeof( ATTRIBUTE_LIST ) ) ) == NULL )
		return( CRYPT_NOMEM );
	memset( newElement, 0, sizeof( ATTRIBUTE_LIST ) );
	memcpy( newElement->oid, oid, sizeofOID( oid ) );
	newElement->isCritical = criticalFlag;
	if( dataLength <= CRYPT_MAX_TEXTSIZE )
		memcpy( newElement->smallData, data, dataLength );
	else
		{
		if( ( newElement->data = malloc( dataLength ) ) == NULL )
			{
			free( newElement );
			return( CRYPT_NOMEM );
			}
		memcpy( newElement->data, data, dataLength );
		}
	newElement->dataLength = dataLength;
	insertListElements( listHeadPtr, insertPoint, newElement, newElement );

	return( CRYPT_OK );
	}

/* Add an attribute field at the appropriate location to a list of
   attributes after checking its validity */

static int checkAttributeField( const ATTRIBUTE_LIST *attributeList,
								const CRYPT_CERTINFO_TYPE fieldID,
								const CRYPT_CERTINFO_TYPE subFieldID,
								const void *data, const int dataLength,
								const BOOLEAN isBlob,
								CRYPT_CERTERROR_TYPE *errorType )
	{
	const ATTRIBUTE_TYPE attributeType = ( fieldID >= CRYPT_FIRST_CMS ) ? \
									ATTRIBUTE_CMS : ATTRIBUTE_CERTIFICATE;
	const ATTRIBUTE_INFO *attributeInfoPtr = fieldIDToAttribute( attributeType,
												fieldID, subFieldID, NULL );
	ATTRIBUTE_LIST *attributeListSearchPtr;

	/* Make sure that a valid field has been specified, and that this field
	   isn't already present as a non-default entry */
	if( attributeInfoPtr == NULL )
		return( CRYPT_BADPARM2 );
	attributeListSearchPtr = findAttributeField( attributeList, fieldID,
												 subFieldID );
	if( attributeListSearchPtr != NULL )
		{
		if( errorType != NULL )
			*errorType = CRYPT_CERTERROR_PRESENT;
		return( CRYPT_INITED );
		}

	/* If we're treating it as a blob, skip the type checking */
	if( isBlob )
		return( CRYPT_OK );

	/* If it's an identifier or special-case field, make sure all parameters
	   are CRYPT_UNUSED */
	if( attributeInfoPtr->fieldType == FIELDTYPE_IDENTIFIER || \
		attributeInfoPtr->fieldType == FIELDTYPE_DN )
		{
		/* Make sure we were called with the correct function */
		if( dataLength != CRYPT_UNUSED )
			return( CRYPT_BADPARM2 );

		/* Make sure the data value is correct */
		if( *( ( int * ) data ) != CRYPT_UNUSED )
			return( CRYPT_BADPARM3 );

		return( CRYPT_OK );
		}

	/* If it's an OID, convert it to a binary OID (which performs a syntax
	   check on the OID) */
	if( attributeInfoPtr->fieldType == BER_OBJECT_IDENTIFIER )
		{
		BYTE binaryOID[ CRYPT_MAX_TEXTSIZE ];

		/* Make sure we were called with the correct function */
		if( dataLength == CRYPT_UNUSED )
			return( CRYPT_BADPARM2 );

		if( !textToOID( data, dataLength, binaryOID ) )
			{
			if( errorType != NULL )
				*errorType = CRYPT_CERTERROR_VALUE;
			return( CRYPT_BADPARM3 );
			}

		return( CRYPT_OK );
		}

	/* If it's an integer type, type check it */
	if( attributeInfoPtr->fieldType == BER_INTEGER || \
		attributeInfoPtr->fieldType == BER_ENUMERATED || \
		attributeInfoPtr->fieldType == BER_BITSTRING || \
		attributeInfoPtr->fieldType == BER_BOOLEAN )
		{
		int value = *( ( int * ) data );

		/* Make sure we were called with the correct function */
		if( dataLength != CRYPT_UNUSED )
			return( CRYPT_BADPARM2 );

		/* Convert BOOLEAN data to the correct range */
		if( attributeInfoPtr->fieldType == BER_BOOLEAN )
			value = ( value ) ? TRUE : FALSE;

		/* Check that the data size and range is valid */
		if( value < attributeInfoPtr->lowRange || \
			value > attributeInfoPtr->highRange )
			{
			if( errorType != NULL )
				*errorType = CRYPT_CERTERROR_VALUE;
			return( CRYPT_BADPARM3 );
			}

		return( CRYPT_OK );
		}

	/* It's a string type, make sure we were called with the correct
	   function */
	if( dataLength == CRYPT_UNUSED )
		return( CRYPT_BADPARM2 );

	/* Type check the value */
	if( dataLength < attributeInfoPtr->lowRange || \
		dataLength > attributeInfoPtr->highRange )
		{
		if( errorType != NULL )
			*errorType = CRYPT_CERTERROR_SIZE;
		return( CRYPT_BADPARM4 );
		}
	if( attributeInfoPtr->fieldType == BER_STRING_NUMERIC )
		{
		const char *dataPtr = data;
		int i;

		for( i = 0; i < dataLength; i++ )
			if( !isdigit( dataPtr[ i ] ) )
				{
				if( errorType != NULL )
					*errorType = CRYPT_CERTERROR_VALUE;
				return( CRYPT_BADPARM4 );
				}
		}
	if( attributeInfoPtr->fieldType == BER_STRING_ISO646 || \
		attributeInfoPtr->fieldType == BER_STRING_IA5 )
		{
		const char *dataPtr = data;
		int i;

		for( i = 0; i < dataLength; i++ )
			if( !isprint( dataPtr[ i ] ) )
				{
				if( errorType != NULL )
					*errorType = CRYPT_CERTERROR_VALUE;
				return( CRYPT_BADPARM4 );
				}
		}

	return( CRYPT_OK );
	}

int addAttributeField( ATTRIBUTE_LIST **listHeadPtr,
					   const CRYPT_CERTINFO_TYPE fieldID,
					   const CRYPT_CERTINFO_TYPE subFieldID,
					   const void *data, const int dataLength,
					   const BOOLEAN criticalFlag, const BOOLEAN isBlob,
					   CRYPT_CERTINFO_TYPE *errorLocus,
					   CRYPT_CERTERROR_TYPE *errorType )
	{
	const ATTRIBUTE_TYPE attributeType = ( fieldID >= CRYPT_FIRST_CMS ) ? \
									ATTRIBUTE_CMS : ATTRIBUTE_CERTIFICATE;
	CRYPT_CERTINFO_TYPE attributeID;
	const ATTRIBUTE_INFO *attributeInfoPtr = fieldIDToAttribute( attributeType,
										fieldID, subFieldID, &attributeID );
	ATTRIBUTE_LIST *newElement, *insertPoint, *prevElement = NULL;
	int status;

	if( attributeInfoPtr == NULL )
		return( CRYPT_ERROR );	/* Internal error, should never happen */

	/* Check the fields validity if necessary */
	status = checkAttributeField( *listHeadPtr, fieldID, subFieldID, data,
								  dataLength, isBlob, errorType );
	if( cryptStatusError( status ) )
		{
		if( errorType != NULL && *errorType != CRYPT_CERTERROR_NONE )
			/* If we encountered an error which sets the error type, record
			   the locus */
			*errorLocus = fieldID;
		return( status );
		}

	/* Find the location at which to insert this attribute field.  For now
	   we assume that the fieldID's are defined in sorted order, we may need
	   to change this and add internal mapping if new fieldID's are added out
	   of order.

	   This loop has a somewhat complex double test, the simpler one is done
	   for basic fields which are identified by the field ID (exit as soon as
	   we find a higher field ID), the more complex one is done for composite
	   fields which can have multiple fields with the same field ID. In this
	   case we exit if the overall field ID is greater (the component belongs
	   to a different field entirely) or if the field ID is the same, when the
	   subfield ID is greater (if the component belongs to the same field) */
	insertPoint = *listHeadPtr;
	while( insertPoint != NULL && \
		   insertPoint->fieldID != CRYPT_CERTINFO_NONE )
		{
		if( subFieldID == CRYPT_CERTINFO_NONE )
			{
			if( insertPoint->fieldID >= fieldID )
				break;
			}
		else
			if( insertPoint->fieldID > fieldID || \
				( insertPoint->fieldID == fieldID && \
				  insertPoint->subFieldID >= subFieldID ) )
				break;

		prevElement = insertPoint;
		insertPoint = insertPoint->next;
		}
	insertPoint = prevElement;

	/* Allocate memory for the new element and copy the information across.
	   If it's a simple type we can assign it to the simple value in the
	   element itself, otherwise we either copy into the storage in the
	   element or allocate seperate storage and copy it into that */
	if( ( newElement  = ( ATTRIBUTE_LIST * ) malloc( sizeof( ATTRIBUTE_LIST ) ) ) == NULL )
		return( CRYPT_NOMEM );
	memset( newElement, 0, sizeof( ATTRIBUTE_LIST ) );
	newElement->attributeID = attributeID;
	newElement->fieldID = fieldID;
	newElement->subFieldID = subFieldID;
	newElement->isCritical = criticalFlag;
	if( subFieldID == CRYPT_CERTINFO_DIRECTORYNAME )
		newElement->isDN = TRUE;
	if( attributeInfoPtr->fieldType == BER_INTEGER || \
		attributeInfoPtr->fieldType == BER_ENUMERATED || \
		attributeInfoPtr->fieldType == BER_BITSTRING || \
		attributeInfoPtr->fieldType == BER_BOOLEAN )
		{
		newElement->value = *( ( int * ) data );
		if( attributeInfoPtr->fieldType == BER_BOOLEAN )
			/* Force it to the correct type if it's a boolean */
			newElement->value = ( newElement->value ) ? TRUE : FALSE;
		}
	else
		if( attributeInfoPtr->fieldType == BER_OBJECT_IDENTIFIER )
			newElement->dataLength = textToOID( data, dataLength,
												newElement->smallData );
		else
			if( attributeInfoPtr->fieldType != FIELDTYPE_IDENTIFIER && \
				attributeInfoPtr->fieldType != FIELDTYPE_DN )
				{
				if( dataLength <= CRYPT_MAX_TEXTSIZE )
					memcpy( newElement->smallData, data, dataLength );
				else
					{
					if( ( newElement->data = malloc( dataLength ) ) == NULL )
						{
						free( newElement );
						return( CRYPT_NOMEM );
						}
					memcpy( newElement->data, data, dataLength );
					}
				newElement->dataLength = dataLength;
				}
	insertListElements( listHeadPtr, insertPoint, newElement, newElement );

	return( CRYPT_OK );
	}

/* Copy an attribute from one attribute list to another.  This is an all-or-
   nothing copy in that it either copies a complete attribute or nothing at
   all */

static int copyAttributeField( ATTRIBUTE_LIST **destAttributeField,
							   const ATTRIBUTE_LIST *srcAttributeField )
	{
	ATTRIBUTE_LIST *newElement;
	int status = CRYPT_OK;

	/* Allocate memory for the new element and copy the information across.
	   If it's a simple type we can assign it to the simple value in the
	   element itself, otherwise we either copy into the storage in the
	   element or allocate seperate storage and copy it into that */
	*destAttributeField = NULL;
	if( ( newElement = ( ATTRIBUTE_LIST * ) malloc( sizeof( ATTRIBUTE_LIST ) ) ) == NULL )
		return( CRYPT_NOMEM );
	memcpy( newElement, srcAttributeField, sizeof( ATTRIBUTE_LIST ) );
	if( srcAttributeField->dataLength > CRYPT_MAX_TEXTSIZE )
		{
		/* If the payload doesn't fit into the attribute structure, copy the
		   separate data payload across */
		if( ( newElement->data = malloc( srcAttributeField->dataLength ) ) == NULL )
			status = CRYPT_NOMEM;
		else
			memcpy( newElement->data, srcAttributeField->data,
					srcAttributeField->dataLength );
		}
	if( srcAttributeField->isDN )
		/* If the field contains a DN, copy the DN across */
		status = copyDN( ( DN_COMPONENT ** ) &newElement->data,
						 srcAttributeField->data );
	if( cryptStatusError( status ) )
		{
		free( newElement );
		return( status );
		}
	newElement->next = newElement->prev = NULL;
	*destAttributeField = newElement;

	return( CRYPT_OK );
	}

static int copyAttribute( ATTRIBUTE_LIST **listHeadPtr,
						  const ATTRIBUTE_LIST *attributeListPtr,
						  const BOOLEAN subjectToIssuer )
	{
	const CRYPT_CERTINFO_TYPE attributeID = attributeListPtr->attributeID;
	CRYPT_CERTINFO_TYPE newAttributeID = attributeID, newFieldID = attributeID;
	ATTRIBUTE_LIST *newAttributeListHead = NULL, *newAttributeListTail;
	ATTRIBUTE_LIST *insertPoint, *prevElement = NULL;

	/* If we're copying from an issuer to a subject attribute list and the
	   field is an altName or keyIdentifier, change the field type from
	   issuer.subjectAltName to subject.issuerAltName or
	   issuer.subjectKeyIdentifier to subject.authorityKeyIdentifier */
	if( subjectToIssuer )
		{
		if( attributeID == CRYPT_CERTINFO_SUBJECTALTNAME )
			newAttributeID = newFieldID = CRYPT_CERTINFO_ISSUERALTNAME;
		if( attributeID == CRYPT_CERTINFO_SUBJECTKEYIDENTIFIER )
			{
			newAttributeID = CRYPT_CERTINFO_AUTHORITYKEYIDENTIFIER;
			newFieldID = CRYPT_CERTINFO_AUTHORITY_KEYIDENTIFIER;
			}
		}

	/* Find the location at which to insert this attribute.  For now we
	   assume that the fieldID's are defined in sorted order, we may need to
	   change this and add internal mapping if new fieldID's are added out of
	   order */
	for( insertPoint = *listHeadPtr;
		 insertPoint != NULL && insertPoint->attributeID < newAttributeID && \
			insertPoint->fieldID != CRYPT_CERTINFO_NONE;
		 insertPoint = insertPoint->next )
		prevElement = insertPoint;
	insertPoint = prevElement;

	/* Build a new attribute list containing the attribute fields */
	while( attributeListPtr != NULL && \
		   attributeListPtr->attributeID == attributeID )
		{
		ATTRIBUTE_LIST *newAttributeField;
		int status;

		/* Copy the field across, append it to the new attribute list, and
		   adjust the type for issuer->subject copying if necessary */
		status = copyAttributeField( &newAttributeField, attributeListPtr );
		if( cryptStatusError( status ) )
			{
			deleteAttributes( &newAttributeListHead );
			return( CRYPT_NOMEM );
			}
		if( newAttributeListHead == NULL )
			newAttributeListHead = newAttributeListTail = newAttributeField;
		else
			{
			newAttributeListTail->next = newAttributeField;
			newAttributeField->prev = newAttributeListTail;
			newAttributeListTail = newAttributeField;
			}
		if( newAttributeID != attributeID )
			{
			newAttributeField->attributeID = newAttributeID;
			newAttributeField->fieldID = newFieldID;
			}

		/* Move on to the next field */
		attributeListPtr = attributeListPtr->next;
		}

	/* Link the new list into the existing list at the appropriate position */
	insertListElements( listHeadPtr, insertPoint, newAttributeListHead,
						newAttributeListTail );

	return( CRYPT_OK );
	}

/* Copy a complete attribute list up to the blob-type attributes */

int copyAttributes( ATTRIBUTE_LIST **destListHeadPtr,
					ATTRIBUTE_LIST *srcListPtr,
					CRYPT_CERTINFO_TYPE *errorLocus,
					CRYPT_CERTERROR_TYPE *errorType )
	{
	ATTRIBUTE_LIST *attributeListCursor = srcListPtr;

	/* Make a first pass down the list checking that the attribute to copy
	   isn't already present, first for recognised attributes and then for
	   unrecognised ones.  We have to do this separately since once we begin
	   the copy process it's rather hard to undo it */
	while( attributeListCursor != NULL && \
		   !isBlobAttribute( attributeListCursor ) )
		{
		if( findAttributeField( *destListHeadPtr,
				attributeListCursor->fieldID, CRYPT_CERTINFO_NONE ) != NULL )
			{
			*errorLocus = attributeListCursor->fieldID;
			*errorType = CRYPT_CERTERROR_PRESENT;
			return( CRYPT_DATA_DUPLICATE );
			}
		attributeListCursor = attributeListCursor->next;
		}
	while( attributeListCursor != NULL )
		{
		if( findAttributeByOID( *destListHeadPtr, attributeListCursor->oid ) != NULL )
			{
			/* We can't set the locus for blob-type attributes since it's not
			   a known attribute */
			*errorLocus = CRYPT_CERTINFO_NONE;
			*errorType = CRYPT_CERTERROR_PRESENT;
			return( CRYPT_DATA_DUPLICATE );
			}
		attributeListCursor = attributeListCursor->next;
		}

	/* Make a second pass copying everything across */
	while( srcListPtr != NULL && !isBlobAttribute( srcListPtr ) )
		{
		CRYPT_CERTINFO_TYPE attributeID = srcListPtr->attributeID;
		int status;

		/* Copy the complete attribute across */
		status = copyAttribute( destListHeadPtr, srcListPtr, FALSE );
		if( cryptStatusError( status ) )
			return( status );

		/* Move on to the next attribute */
		while( srcListPtr != NULL && srcListPtr->attributeID == attributeID )
			srcListPtr = srcListPtr->next;
		}

	/* If there are blob-type attributes left at the end of the list, copy
	   them across last */
	if( srcListPtr != NULL )
		{
		ATTRIBUTE_LIST *insertPoint;

		/* Find the end of the destination list */
		for( insertPoint = *destListHeadPtr;
			 insertPoint != NULL && insertPoint->next != NULL;
			 insertPoint = insertPoint->next );

		/* Copy all remaining attributes across */
		while( srcListPtr != NULL )
			{
			ATTRIBUTE_LIST *newAttribute;
			int status;

			status = copyAttributeField( &newAttribute, srcListPtr );
			if( cryptStatusError( status ) )
				return( status );
			insertListElements( destListHeadPtr, insertPoint, newAttribute,
								newAttribute );
			srcListPtr = srcListPtr->next;
			}
		}

	return( CRYPT_OK );
	}

/* Copy attributes which are propagated down cert chains from an issuer to a
   subject cert, changing the field types from subject to issuer at the same
   time if required */

int copyIssuerAttributes( ATTRIBUTE_LIST **destListHeadPtr,
						  const ATTRIBUTE_LIST *srcListPtr,
						  CRYPT_CERTINFO_TYPE *errorLocus,
						  CRYPT_CERTERROR_TYPE *errorType,
						  const CRYPT_CERTTYPE_TYPE type )
	{
	ATTRIBUTE_LIST *attributeListPtr;
	int status;

	/* If the destination is a CA cert and the source has name constraints,
	   copy them over to the destination */
	attributeListPtr = findAttribute( *destListHeadPtr, CRYPT_CERTINFO_CA );
	if( attributeListPtr != NULL && attributeListPtr->value )
		{
		ATTRIBUTE_LIST *srcPermittedSubtrees, *srcExcludedSubtrees;

		srcPermittedSubtrees = findAttributeField( *destListHeadPtr,
					CRYPT_CERTINFO_PERMITTEDSUBTREES, CRYPT_CERTINFO_NONE );
		srcExcludedSubtrees = findAttributeField( *destListHeadPtr,
        			CRYPT_CERTINFO_EXCLUDEDSUBTREES, CRYPT_CERTINFO_NONE );

		/* If we're copying permitted or excluded subtrees, they can't
		   already be present */
		if( srcPermittedSubtrees != NULL && \
			findAttributeField( *destListHeadPtr, \
            		CRYPT_CERTINFO_PERMITTEDSUBTREES, CRYPT_CERTINFO_NONE ) != NULL )
			{
			*errorLocus = CRYPT_CERTINFO_PERMITTEDSUBTREES;
			*errorType = CRYPT_CERTERROR_PRESENT;
			return( CRYPT_DATA_DUPLICATE );
			}
		if( srcExcludedSubtrees != NULL && \
			findAttributeField( *destListHeadPtr,
            		CRYPT_CERTINFO_EXCLUDEDSUBTREES, CRYPT_CERTINFO_NONE ) != NULL )
			{
			*errorLocus = CRYPT_CERTINFO_EXCLUDEDSUBTREES;
			*errorType = CRYPT_CERTERROR_PRESENT;
			return( CRYPT_DATA_DUPLICATE );
			}

		/* Copy the fields across */
		if( srcPermittedSubtrees != NULL )
			status = copyAttribute( destListHeadPtr, srcPermittedSubtrees, FALSE );
		if( cryptStatusOK( status ) && srcExcludedSubtrees != NULL )
			status = copyAttribute( destListHeadPtr, srcExcludedSubtrees, FALSE );
		if( cryptStatusError( status ) )
			return( status );
		}

	/* If it's an attribute certificate, that's all we can copy */
	if( type == CRYPT_CERTTYPE_ATTRIBUTE_CERT )
		return( CRYPT_OK );

	/* Copy the altName and keyIdentifier if these are present.  We don't
	   have to check for their presence in the destination cert since they're
	   read-only fields and can't be added by the user */
	attributeListPtr = findAttributeField( srcListPtr,
					CRYPT_CERTINFO_SUBJECTALTNAME, CRYPT_CERTINFO_NONE );
	if( attributeListPtr != NULL )
		{
		status = copyAttribute( destListHeadPtr, attributeListPtr, TRUE );
		if( cryptStatusError( status ) )
			return( status );
		}
	attributeListPtr = findAttributeField( srcListPtr,
					CRYPT_CERTINFO_SUBJECTKEYIDENTIFIER, CRYPT_CERTINFO_NONE );
	if( attributeListPtr != NULL )
		{
		status = copyAttribute( destListHeadPtr, attributeListPtr, TRUE );
		if( cryptStatusError( status ) )
			return( status );
		}

	return( CRYPT_OK );
	}

/* Delete an attribute/attribute field from a list of attributes, updating
   the list cursor at the same time (this is a somewhat ugly kludge, it's
   not really possible to do this cleanly) */

void deleteAttributeField( ATTRIBUTE_LIST **listHeadPtr,
						   ATTRIBUTE_LIST **listCursorPtr,
						   ATTRIBUTE_LIST *listItem )
	{
	ATTRIBUTE_LIST *listPrevPtr = listItem->prev;
	ATTRIBUTE_LIST *listNextPtr = listItem->next;

	/* If we're about to delete the field which is pointed to by the
	   attribute cursor, advance the cursor to the next field.  If there's no
	   next field, move it to the previous field.  This behaviour is the most
	   logically consistent, it means we can do things like deleting an
	   entire attribute list by repeatedly deleting a field */
	if( listCursorPtr != NULL && *listCursorPtr == listItem )
		*listCursorPtr = ( listNextPtr != NULL ) ? listNextPtr : listPrevPtr;

	/* Remove the item from the list */
	if( listItem == *listHeadPtr )
		{
		/* Special case for first item */
		*listHeadPtr = listNextPtr;
		if( listNextPtr != NULL )
			listNextPtr->prev = NULL;
		}
	else
		{
		/* Delete from the middle or the end of the chain */
		listPrevPtr->next = listNextPtr;
		if( listNextPtr != NULL )
			listNextPtr->prev = listPrevPtr;
		}

	/* Clear all data in the item and free the memory */
	if( listItem->isDN )
		deleteDN( ( DN_COMPONENT ** ) &listItem->data );
	else
		if( listItem->dataLength > CRYPT_MAX_TEXTSIZE )
			{
			zeroise( listItem->data, listItem->dataLength );
			free( listItem->data );
			}
	zeroise( listItem, sizeof( ATTRIBUTE_LIST ) );
	free( listItem );
	}

void deleteAttribute( ATTRIBUTE_LIST **listHeadPtr,
					  ATTRIBUTE_LIST **listCursorPtr,
					  ATTRIBUTE_LIST *listItem )
	{
	CRYPT_CERTINFO_TYPE attributeID;
	ATTRIBUTE_LIST *attributeListPtr;

	/* If it's a blob-type attribute, everything is contained in this one
	   list item so we only need to destroy that */
	if( isBlobAttribute( listItem ) )
		{
		deleteAttributeField( listHeadPtr, listCursorPtr, listItem );
		return;
		}

	/* If it's a field which denotes an entire (constructed) attribute, it
	   won't have an entry in the list, so we find the first field of the
	   constructed attribute which is present in the list and start deleting
	   from that point */
	if( isCompleteAttribute( listItem ) )
		{
		for( attributeListPtr = *listHeadPtr; attributeListPtr != NULL && \
			 attributeListPtr->attributeID != listItem->value;
			 attributeListPtr = attributeListPtr->next );
		if( attributeListPtr == NULL )
			return;	/* Internal error, should never happen */
		}
	else
		/* The list item is a field in the attribute, find the start of the
		   fields in this attribute */
		attributeListPtr = findAttributeStart( listItem );
	attributeID = attributeListPtr->attributeID;

	/* It's an item with multiple fields, destroy each field separately */
	while( attributeListPtr != NULL && \
		   attributeListPtr->attributeID == attributeID )
		{
		ATTRIBUTE_LIST *itemToFree = attributeListPtr;

		attributeListPtr = attributeListPtr->next;
		deleteAttributeField( listHeadPtr, listCursorPtr, itemToFree );
		}
	}

/* Delete a certificate attribute component list */

void deleteAttributes( ATTRIBUTE_LIST **listHeadPtr )
	{
	ATTRIBUTE_LIST *attributeListPtr = *listHeadPtr;

	/* If the list was empty, return now */
	if( attributeListPtr == NULL )
		return;

	/* Destroy any remaining list items */
	while( attributeListPtr != NULL )
		{
		ATTRIBUTE_LIST *itemToFree = attributeListPtr;

		attributeListPtr = attributeListPtr->next;
		deleteAttributeField( listHeadPtr, NULL, itemToFree );
		}
	}
