/****************************************************************************
*																			*
*							cryptlib PKCS #11 Routines						*
*						Copyright Peter Gutmann 1998-1999					*
*																			*
****************************************************************************/

/* This file contains its own version of the various PKCS #11 definitions and
   values to avoid potential copyright problems with redistributing the
   PKCS #11 header files */

#include <stdlib.h>
#include <string.h>
#if defined( INC_ALL )
  #include "crypt.h"
  #include "cryptctx.h"
  #include "device.h"
#elif defined( INC_CHILD )
  #include "../crypt.h"
  #include "../cryptctx.h"
  #include "device.h"
#else
  #include "crypt.h"
  #include "cryptctx.h"
  #include "misc/device.h"
#endif /* Compiler-specific includes */

/* Data types and constants */

typedef unsigned char CK_BYTE;			/* Unsigned 8-bit */
typedef CK_BYTE CK_CHAR;				/* Unsigned character */
typedef CK_BYTE CK_BOOL;				/* Boolean */
typedef unsigned short int CK_USHORT;	/* Unsigned 16-bit */
typedef unsigned long int CK_ULONG;		/* Unsigned 32-bit */
typedef CK_ULONG CK_FLAGS;				/* Bitflags */

typedef CK_BYTE *CK_BYTE_PTR;
typedef CK_CHAR *CK_CHAR_PTR;
typedef CK_USHORT *CK_USHORT_PTR;
typedef void *CK_VOID_PTR;				/* Pointers to the above values */

#define NULL_PTR	NULL

/* General information types */

typedef struct {
	CK_BYTE major;						/* Major version number */
	CK_BYTE minor;						/* Minor version number */
	} CK_VERSION;

typedef struct {
	CK_VERSION version;					/* Version info */
	CK_CHAR manufacturerID[ 32 ];		/* Blank-padded PKCS #11 lib.vendor */
	CK_FLAGS flags;
	} CK_INFO;

typedef CK_INFO *CK_INFO_PTR;			/* Pointer to CK_INFO */

typedef enum { CKN_SURRENDER, CKN_COMPLETE, CKN_DEVICE_REMOVED } CK_NOTIFICATION;

/* Slot and token types */

typedef CK_ULONG CK_SLOT_ID;			/* Slot number */
typedef CK_SLOT_ID *CK_SLOT_ID_PTR;		/* Pointer to CK_SLOT_ID */

typedef struct {
	CK_CHAR slotDescription[ 64 ];		/* Blank-padded description of slot */
	CK_CHAR manufacturerID[ 32 ];		/* Blank-padded slot vendor */
	CK_FLAGS flags;
	} CK_SLOT_INFO;

typedef CK_SLOT_INFO *CK_SLOT_INFO_PTR;	/* Pointer to CK_SLOT_INFO */

typedef struct {
	CK_CHAR label[ 32 ];				/* Blank-padded label for token */
	CK_CHAR manufacturerID[ 32 ];		/* Blank-padded manuf.ID for device */
	CK_CHAR model[ 16 ];				/* Blank-padded model number */
	CK_CHAR serialNumber[ 16 ];			/* Blank-padded serial number */
	CK_FLAGS flags;						/* Device capabilities, see below */
	CK_USHORT usMaxSessionCount;		/* Max.concurrent sessions */
	CK_USHORT usSessionCount;			/* Current session count */
	CK_USHORT usMaxRwSessionCount;		/* Max.R/W sessions */
	CK_USHORT usRwSessionCount;			/* Current R/W sessions */
	CK_USHORT usMaxPinLen;				/* Max.bytes in PIN */
	CK_USHORT usMinPinLen;				/* Min.bytes in PIN */
	CK_ULONG ulTotalPublicMemory;		/* Memory used for public objects */
	CK_ULONG ulFreePublicMemory;		/* Memory avail.for public objects */
	CK_ULONG ulTotalPrivateMemory;		/* Memory used for private objects */
	CK_ULONG ulFreePrivateMemory;		/* Memory avail.for private objects */
	} CK_TOKEN_INFO;

typedef CK_TOKEN_INFO *CK_TOKEN_INFO_PTR;	/* Pointer to CK_TOKEN_INFO */

#define CKF_RNG					0x0001	/* Device has an RNG */
#define CKF_WRITE_PROTECTED		0x0002	/* Device is write-protected */
#define CKF_LOGIN_REQUIRED		0x0004	/* User must log in to use device */
#define CKF_USER_PIN_INITIALIZED 0x0008	/* Users PIN has been initialised */
#define CKF_EXCLUSIVE_EXISTS	0x0010	/* Exclusive session to dev.exists */

/* Session types */

typedef CK_ULONG CK_SESSION_HANDLE;		/* Session handle */

typedef CK_SESSION_HANDLE *CK_SESSION_HANDLE_PTR;	/* Pointer to above */

typedef enum { CKU_SO, CKU_USER } CK_USER_TYPE;
typedef enum { CKS_RW_PUBLIC_SESSION, CKS_RW_USER_FUNCTIONS,
			   CKS_RO_PUBLIC_SESSION, CKS_RO_SO_FUNCTIONS,
			   CKS_RO_USER_FUNCTIONS } CK_STATE;

typedef struct {
	CK_SLOT_ID slotID;					/* Slot ID for token */
	CK_STATE state;						/* Session state */
	CK_FLAGS flags;						/* Flags for session, see below */
	CK_USHORT usDeviceError;			/* Device-specific error code */
	} CK_SESSION_INFO;

typedef CK_SESSION_INFO *CK_SESSION_INFO_PTR;	/* Pointer to above */

#define CKF_EXCLUSIVE_SESSION	0x0001	/* Session is exclusive */
#define CKF_RW_SESSION			0x0002	/* Session is R/W */
#define CKF_SERIAL_SESSION		0x0004	/* Crypto ops are non-async */

/* Object types */

typedef CK_ULONG CK_OBJECT_HANDLE;		/* Object handle */

typedef CK_OBJECT_HANDLE *CK_OBJECT_HANDLE_PTR;	/* Pointer to above */

typedef CK_USHORT CK_OBJECT_CLASS;		/* Object class */

typedef CK_OBJECT_CLASS *CK_OBJECT_CLASS_PTR;	/* Pointer to above */

#define CKO_DATA				0x0000	/* Data object */
#define CKO_CERTIFICATE			0x0001	/* Cert.object */
#define CKO_PUBLIC_KEY			0x0002	/* Public key object */
#define CKO_PRIVATE_KEY			0x0003	/* Private key object */
#define CKO_SECRET_KEY			0x0004	/* Secret key object */
#define CKO_VENDOR_DEFINED		0x8000	/* >= vendor-defined obj.*/

typedef CK_USHORT CK_KEY_TYPE;			/* Key type */

#define CKK_RSA					0x0000	/* RSA */
#define CKK_DSA					0x0001	/* DSA */
#define CKK_DH					0x0002	/* DH */
#define CKK_GENERIC_SECRET		0x0010
#define CKK_RC2					0x0011	/* RC2 */
#define CKK_RC4					0x0012	/* RC4 */
#define CKK_DES					0x0013	/* DES */
#define CKK_DES2				0x0014	/* Two-key triple DES */
#define CKK_DES3				0x0015	/* Three-key triple DES */
#define CKK_VENDOR_DEFINED		0x8000	/* >= vendor-defined key type */

typedef CK_USHORT CK_CERTIFICATE_TYPE;	/* Cert.type */

#define CK_X_509				0x0000	/* X.509 */
#define CK_VENDOR_DEFINED		0x8000	/* >= vendor-defined cert.type */

typedef CK_USHORT CK_ATTRIBUTE_TYPE;	/* Attribute type */

#define CKA_CLASS				0x0000
#define CKA_TOKEN				0x0001
#define CKA_PRIVATE				0x0002
#define CKA_LABEL				0x0003
#define CKA_APPLICATION			0x0010
#define CKA_VALUE				0x0011
#define CKA_CERTIFICATE_TYPE	0x0080
#define CKA_ISSUER				0x0081
#define CKA_SERIAL_NUMBER		0x0082
#define CKA_KEY_TYPE			0x0100
#define CKA_SUBJECT				0x0101
#define CKA_ID					0x0102
#define CKA_SENSITIVE			0x0103
#define CKA_ENCRYPT				0x0104
#define CKA_DECRYPT				0x0105
#define CKA_WRAP				0x0106
#define CKA_UNWRAP				0x0107
#define CKA_SIGN				0x0108
#define CKA_SIGN_RECOVER		0x0109
#define CKA_VERIFY				0x010A
#define CKA_VERIFY_RECOVER		0x010B
#define CKA_DERIVE				0x010C
#define CKA_MODULUS				0x0120
#define CKA_MODULUS_BITS		0x0121
#define CKA_PUBLIC_EXPONENT		0x0122
#define CKA_PRIVATE_EXPONENT	0x0123
#define CKA_PRIME_1				0x0124
#define CKA_PRIME_2				0x0125
#define CKA_EXPONENT_1			0x0126
#define CKA_EXPONENT_2			0x0127
#define CKA_COEFFICIENT			0x0128
#define CKA_PRIME				0x0130
#define CKA_SUBPRIME			0x0131
#define CKA_BASE				0x0132
#define CKA_VALUE_BITS			0x0160
#define CKA_VALUE_LEN			0x0161
#define CKA_VENDOR_DEFINED		0x8000

typedef struct {
	CK_ATTRIBUTE_TYPE type;				/* Attribute type */
	CK_VOID_PTR pValue;					/* Attribute value */
	CK_USHORT usValueLen;				/* Attribute length */
	} CK_ATTRIBUTE;

typedef CK_ATTRIBUTE *CK_ATTRIBUTE_PTR;	/* Pointer to the above */

typedef struct {
	CK_CHAR year[ 4 ];					/* Year, 1900-9999 */
	CK_CHAR month[ 2 ];					/* Month, 01-12 */
	CK_CHAR day[ 2 ];					/* Day, 01-31 */
	} CK_DATE;

/* Mechanisms */

typedef CK_USHORT CK_MECHANISM_TYPE;	/* Mechanism type */

typedef CK_MECHANISM_TYPE *CK_MECHANISM_TYPE_PTR;	/* Pointer to above */

#define CKM_RSA_PKCS_KEY_PAIR_GEN 0x0000 /* Generate RSA key pair */
#define CKM_RSA_PKCS			0x0001	/* RSA enc/dec, PKCS #1 padding */
#define CKM_RSA_9796			0x0002	/* RSA enc/dec, ISO 9796 padding */
#define CKM_RSA_X_509			0x0003	/* RSA sign, X.509 */
#define CKM_RSA_DSA_KEY_PAIR_GEN 0x0010	/* Generate DSA key pair */
#define CKM_DSA					0x0011	/* DSA sign */
#define CKM_DH_PKCS_KEY_PAIR_GEN 0x0020	/* Generate DH key pair */
#define CKM_DH_PKCS_DERIVE		0x0021	/* Derive DH key */
#define CKM_RC2_KEY_GEN			0x0100	/* Generate RC2 key */
#define CKM_RC2_ECB				0x0101	/* RC2/ECB */
#define CKM_RC2_CBC				0x0102	/* RC2/CBC */
#define CKM_RC2_MAC				0x0103	/* RC2/MAC */
#define CKM_RC4_KEY_GEN			0x0110	/* Generate RC4 key */
#define CKM_RC4					0x0111	/* RC4 */
#define CKM_DES_KEY_GEN			0x0120	/* Generate DES key */
#define CKM_DES_ECB				0x0121	/* DES/ECB */
#define CKM_DES_CBC				0x0122	/* DES/CBC */
#define CKM_DES_MAC				0x0123	/* DES/MAC */
#define CKM_DES2_KEY_GEN		0x0130	/* Generate two-key triple DES key */
#define CKM_DES3_KEY_GEN		0x0131	/* Generate three-key triple DES key */
#define CKM_DES3_ECB			0x0132	/* 3DES/ECB */
#define CKM_DES3_CBC			0x0133	/* 3DES/CBC */
#define CKM_DES3_MAC			0x0134	/* 3DES/MAC */
#define CKM_MD2					0x0200	/* MD2 */
#define CKM_MD5					0x0210	/* MD5 */
#define CKM_SHA_1				0x0220	/* SHA1 */
#define CKM_VENDOR_DEFINED		0x8000	/* >= vendor-defined mech.type */

typedef struct {
	CK_MECHANISM_TYPE mechanism;		/* Mechanism type */
	CK_VOID_PTR pParameter;				/* Optional parameter(s) */
	CK_USHORT usParameterLen;			/* Length of optional params */
	} CK_MECHANISM;

typedef CK_MECHANISM *CK_MECHANISM_PTR;	/* Pointer to the above */

typedef struct {
	CK_ULONG ulMinKeySize;				/* Min.key size for mechanism */
	CK_ULONG ulMaxKeySize;				/* Max.key size for mechanism */
	CK_FLAGS flags;						/* Mechanism flags (see below) */
	} CK_MECHANISM_INFO;

typedef CK_MECHANISM_INFO *CK_MECHANISM_INFO_PTR;	/* Ptr.to the above */

#define CKE_HW					0x0001	/* Mechanism implem.in hardware */
#define CKE_EXTENSION			0x8000	/* Extension to flags */

typedef struct {
	CK_USHORT usEffectiveBits;			/* Effective RC2 key bits */
	CK_BYTE iv[ 8 ];					/* IV for CBC */
	} CK_RC2_CBC;

/* Functions */

#ifdef __WINDOWS__
  #ifdef __WIN16__
	#define CK_ENTRY	FAR PASCAL _export
  #else
	#define CK_ENTRY	__stdcall
  #endif /* Windows version-specific entry types */
#else
  #define CK_ENTRY
#endif /* OS-specific entry types */

typedef CK_USHORT CK_RV;

#define CKR_OK							0x0000
#define CKR_CANCEL						0x0001
#define CKR_HOST_MEMORY					0x0002
#define CKR_SLOT_ID_INVALID				0x0003
#define CKR_FLAGS_INVALID				0x0004
#define CKR_ATTRIBUTE_READ_ONLY			0x0010
#define CKR_ATTRIBUTE_SENSITIVE			0x0011
#define CKR_ATTRIBUTE_TYPE_INVALID		0x0012
#define CKR_ATTRIBUTE_VALUE_INVALID		0x0013
#define CKR_DATA_INVALID				0x0020
#define CKR_DATA_LEN_RANGE				0x0021
#define CKR_DEVICE_ERROR				0x0030
#define CKR_DEVICE_MEMORY				0x0031
#define CKR_DEVICE_REMOVED				0x0032
#define CKR_ENCRYPTED_DATA_INVALID		0x0040
#define CKR_ENCRYPTED_DATA_LEN_RANGE	0x0041
#define CKR_FUNCTION_CANCELLED			0x0050
#define CKR_FUNCTION_NOT_PARALLEL		0x0051
#define CKR_FUNCTION_PARALLEL			0x0052
#define CKR_KEY_HANDLE_INVALID			0x0060
#define CKR_KEY_SENSITIVE				0x0061
#define CKR_KEY_SIZE_RANGE				0x0062
#define CKR_KEY_TYPE_INCONSISTENT		0x0063
#define CKR_MECHANISM_INVALID			0x0070
#define CKR_MECHANISM_PARAM_INVALID		0x0071
#define CKR_OBJECT_CLASS_INCONSISTENT	0x0080
#define CKR_OBJECT_CLASS_INVALID		0x0081
#define CKR_OBJECT_HANDLE_INVALID		0x0082
#define CKR_OPERATION_ACTIVE			0x0090
#define CKR_OPERATION_NOT_INITIALIZED	0x0091
#define CKR_PIN_INCORRECT				0x00A0
#define CKR_PIN_INVALID					0x00A1
#define CKR_PIN_LEN_RANGE				0x00A2
#define CKR_SESSION_CLOSED				0x00B0
#define CKR_SESSION_COUNT				0x00B1
#define CKR_SESSION_EXCLUSIVE_EXISTS	0x00B2
#define CKR_SESSION_HANDLE_INVALID		0x00B3
#define CKR_SESSION_PARALLEL_NOT_SUPPORTED 0x00B4
#define CKR_SESSION_READ_ONLY			0x00B5
#define CKR_SIGNATURE_INVALID			0x00C0
#define CKR_SIGNATURE_LEN_RANGE			0x00C1
#define CKR_TEMPLATE_INCOMPLETE			0x00D0
#define CKR_TEMPLATE_INCONSISTENT		0x00D1
#define CKR_TOKEN_NOT_PRESENT			0x00E0
#define CKR_TOKEN_NOT_RECOGNISED		0x00E1
#define CKR_TOKEN_WRITE_PROTECTED		0x00E2
#define CKR_UNWRAPPING_KEY_HANDLE_INVALID 0x00F0
#define CKR_UNWRAPPING_KEY_SIZE_RANGE	0x00F1
#define CKR_USER_ALREADY_LOGGED_IN		0x0100
#define CKR_USER_NOT_LOGGED_IN			0x0101
#define CKR_USER_PIN_NOT_INITIALIZED	0x0102
#define CKR_USER_TYPE_INVALID			0x0103
#define CKR_WRAPPED_KEY_INVALID			0x0110
#define CKR_WRAPPED_KEY_LEN_RANGE		0x0112
#define CKR_WRAPPING_KEY_HANDLE_INVALID	0x0113
#define CKR_WRAPPING_KEY_SIZE_RANGE		0x0114
#define CKR_WRAPPING_KEY_TYPE_INCONSISTENT 0x0115
#define CKR_VENDOR_DEFINED				0x8000

/* Function prototypes */

CK_RV CK_ENTRY C_CloseSession( CK_SESSION_HANDLE hSession );
CK_RV CK_ENTRY C_GetTokenInfo( CK_SESSION_HANDLE hSession,
							   CK_TOKEN_INFO_PTR pInfo );
CK_RV CK_ENTRY C_Initialize( CK_VOID_PTR pReserved );
CK_RV CK_ENTRY C_InitToken( CK_SLOT_ID slotID, CK_CHAR_PTR pPin,
							CK_USHORT usPinLen, CK_CHAR_PTR pLabel );
CK_RV CK_ENTRY C_Login( CK_SESSION_HANDLE hSession, CK_USER_TYPE userType,
						CK_CHAR_PTR pPin, CK_USHORT usPinLen );
CK_RV CK_ENTRY C_Logout( CK_SESSION_HANDLE hSession );
CK_RV CK_ENTRY C_OpenSession( CK_SLOT_ID slotID, CK_FLAGS flags,
							  CK_VOID_PTR pApplication,
							  CK_RV ( CK_ENTRY *Notify )( CK_SESSION_HANDLE hSession,
									CK_NOTIFICATION event, CK_VOID_PTR pApplication ),
							  CK_SESSION_HANDLE_PTR phSession );
CK_RV CK_ENTRY C_SetPIN( CK_SESSION_HANDLE hSession, CK_CHAR_PTR pOldPin,
						 CK_USHORT usOldLen, CK_CHAR_PTR pNewPin,
						 CK_USHORT usNewLen );

/****************************************************************************
*																			*
*						 		Init/Shutdown Routines						*
*																			*
****************************************************************************/

#ifdef __WINDOWS__

/* Global function pointers.  These are necessary because the functions need
   to be dynamically linked since not all systems contain the necessary
   DLL's.  Explicitly linking to them will make cryptlib unloadable on most
   systems */

#define NULL_HINSTANCE	( HINSTANCE ) NULL

static HINSTANCE hPKCS11 = NULL_HINSTANCE;

typedef CK_RV ( CK_ENTRY *C_CLOSESESSION )( CK_SESSION_HANDLE hSession );
typedef CK_RV ( CK_ENTRY *C_GETTOKENINFO )( CK_SESSION_HANDLE hSession,
											CK_TOKEN_INFO_PTR pInfo );
typedef CK_RV ( CK_ENTRY *C_INITIALIZE )( CK_VOID_PTR pReserved );
typedef CK_RV ( CK_ENTRY *C_INITTOKEN )( CK_SLOT_ID slotID, CK_CHAR_PTR pPin,
										 CK_USHORT usPinLen,
										 CK_CHAR_PTR pLabel );
typedef CK_RV ( CK_ENTRY *C_LOGIN )( CK_SESSION_HANDLE hSession,
									 CK_USER_TYPE userType, CK_CHAR_PTR pPin,
									 CK_USHORT usPinLen );
typedef CK_RV ( CK_ENTRY *C_LOGOUT )( CK_SESSION_HANDLE hSession );
typedef CK_RV ( CK_ENTRY *C_OPENSESSION )( CK_SLOT_ID slotID, CK_FLAGS flags,
							CK_VOID_PTR pApplication,
							CK_RV ( CK_ENTRY *Notify )( CK_SESSION_HANDLE hSession,
											CK_NOTIFICATION event, CK_VOID_PTR pApplication ),
							CK_SESSION_HANDLE_PTR phSession );
typedef CK_RV ( CK_ENTRY *C_SETPIN )( CK_SESSION_HANDLE hSession,
									  CK_CHAR_PTR pOldPin, CK_USHORT usOldLen, CK_CHAR_PTR pNewPin,
									  CK_USHORT usNewLen );
static C_CLOSESESSION pC_CloseSession = NULL;
static C_GETTOKENINFO pC_GetTokenInfo = NULL;
static C_INITIALIZE pC_Initialize = NULL;
static C_INITTOKEN pC_InitToken = NULL;
static C_LOGIN pC_Login = NULL;
static C_LOGOUT pC_Logout = NULL;
static C_OPENSESSION pC_OpenSession = NULL;
static C_SETPIN pC_SetPIN = NULL;

/* The use of dynamically bound function pointers vs statically linked
   functions requires a bit of sleight of hand since we can't give the
   pointers the same names as prototyped functions.  To get around this we
   redefine the actual function names to the names of the pointers */

#define C_CloseSession	pC_CloseSession
#define C_GetTokenInfo	pC_GetTokenInfo
#define C_Initialize	pC_Initialize
#define C_InitToken		pC_InitToken
#define C_Login			pC_Login
#define C_Logout		pC_Logout
#define C_OpenSession	pC_OpenSession
#define C_SetPIN		pC_SetPIN

/* Depending on whether we're running under Win16 or Win32 we load the device
   driver under a different name */

#ifdef __WIN16__
  #define PKCS11_LIBNAME	"PKCS11.DLL"
#else
  #define PKCS11_LIBNAME	"PKCS11_32.DLL"
#endif /* __WIN16__ */

/* Dynamically load and unload any necessary card drivers */

void deviceInitPKCS11( void )
	{
#ifdef __WIN16__
	UINT errorMode;
#endif /* __WIN16__ */
	static BOOLEAN initCalled = FALSE;

	/* If we've previously tried to init the drivers, don't try it again */
	if( initCalled )
		return;
	initCalled = TRUE;

	/* Obtain a handle to the device driver module */
#ifdef __WIN16__
	errorMode = SetErrorMode( SEM_NOOPENFILEERRORBOX );
	hPKCS11 = LoadLibrary( PKCS11_LIBNAME );
	SetErrorMode( errorMode );
	if( hPKCS11 < HINSTANCE_ERROR )
		{
		hPKCS11 = NULL_HINSTANCE;
		return;
		}
#else
	if( ( hPKCS11 = LoadLibrary( PKCS11_LIBNAME ) ) == NULL_HINSTANCE )
		return;
#endif /* __WIN32__ */

	/* Now get pointers to the functions */
	pC_CloseSession = ( C_CLOSESESSION ) GetProcAddress( hPKCS11, "C_CloseSession" );
	pC_GetTokenInfo = ( C_GETTOKENINFO ) GetProcAddress( hPKCS11, "C_GetTokenInfo" );
	pC_Initialize = ( C_INITIALIZE ) GetProcAddress( hPKCS11, "C_Initialize" );
	pC_InitToken = ( C_INITTOKEN ) GetProcAddress( hPKCS11, "C_InitToken" );
	pC_Login = ( C_LOGIN ) GetProcAddress( hPKCS11, "C_Login" );
	pC_Logout = ( C_LOGOUT ) GetProcAddress( hPKCS11, "C_Logout" );
	pC_OpenSession = ( C_OPENSESSION ) GetProcAddress( hPKCS11, "C_OpenSession" );
	pC_SetPIN = ( C_SETPIN ) GetProcAddress( hPKCS11, "C_SetPIN" );

	/* Make sure we got valid pointers for every device function */
	if( pC_CloseSession == NULL || pC_GetTokenInfo == NULL ||
		pC_Initialize == NULL || pC_InitToken == NULL || pC_Login == NULL ||
		pC_Logout == NULL || pC_OpenSession == NULL || pC_SetPIN == NULL )
		{
		/* Free the library reference and reset the handle */
		FreeLibrary( hPKCS11 );
		hPKCS11 = NULL_HINSTANCE;
		}

	/* Initialise the PKCS #11 library */
	if( pC_Initialize( NULL_PTR ) != CKR_OK )
		{
		/* Free the library reference and reset the handle */
		FreeLibrary( hPKCS11 );
		hPKCS11 = NULL_HINSTANCE;
		}
	}

void deviceEndPKCS11( void )
	{
	if( hPKCS11 != NULL_HINSTANCE )
		FreeLibrary( hPKCS11 );
	hPKCS11 = NULL_HINSTANCE;
	}
#endif /* __WINDOWS__ */

/****************************************************************************
*																			*
*						 		Utility Routines							*
*																			*
****************************************************************************/

/* Map a PKCS #11-specific error to a cryptlib error */

static int mapError( DEVICE_INFO *deviceInfo, const int errorCode,
					 const int defaultError )
	{
	deviceInfo->errorCode = errorCode;
	switch( errorCode )
		{
		case CKR_OK:
			return( CRYPT_OK );
		case CKR_HOST_MEMORY:
		case CKR_DEVICE_MEMORY:
			return( CRYPT_NOMEM );
		case CKR_DEVICE_ERROR:
		case CKR_DEVICE_REMOVED:
			return( CRYPT_SIGNALLED );
		case CKR_PIN_INCORRECT:
		case CKR_PIN_INVALID:
		case CKR_PIN_LEN_RANGE:
			return( CRYPT_WRONGKEY );
		case CKR_SIGNATURE_INVALID:
			return( CRYPT_BADSIG );
		}

	return( defaultError );
	}

/****************************************************************************
*																			*
*					Device Init/Shutdown/Device Control Routines			*
*																			*
****************************************************************************/

/* Close a previously-opened session with the device.  We have to have this
   before the init function since it may be called by it if the init process
   fails */

static void shutdownDeviceFunction( DEVICE_INFO *deviceInfo )
	{
	UNUSED( deviceInfo );

	/* Close the session with the device */
	C_CloseSession( deviceInfo->deviceHandle );
	}

/* Open a session with the device */

static int initDeviceFunction( DEVICE_INFO *deviceInfo )
	{
	CK_SESSION_HANDLE hSession;
	int status;

	/* Open a session with the device */
	status = C_OpenSession( 0, CKF_EXCLUSIVE_SESSION, NULL_PTR, NULL_PTR,
							&hSession );
	if( status != CKR_OK )
		return( mapError( deviceInfo, status, CRYPT_DATA_OPEN ) );
	deviceInfo->deviceHandle = ( int ) hSession;

	return( CRYPT_OK );
	}

/* Handle device control functions */

static int controlFunction( DEVICE_INFO *deviceInfo,
							const CRYPT_DEVICECONTROL_TYPE type,
							const void *data1, const int data1Length,
							const void *data2, const int data2Length )
	{
	int status;

	/* Perform any additional parameter checking.  We do this before doing
	   anything else to allow better error reporting */
	if( type == CRYPT_DEVICECONTROL_AUTH_USER || \
		type == CRYPT_DEVICECONTROL_AUTH_SUPERVISOR || \
		type == CRYPT_DEVICECONTROL_SET_AUTH_USER || \
		type == CRYPT_DEVICECONTROL_SET_AUTH_SUPERVISOR || \
		type == CRYPT_DEVICECONTROL_ZEROISE )
		{
		CK_TOKEN_INFO tokenInfo;

		/* Make sure the PIN size is valid */
		status = C_GetTokenInfo( deviceInfo->deviceHandle, &tokenInfo );
		if( status != CKR_OK )
			return( mapError( deviceInfo, status, CRYPT_ERROR ) );
		if( data1Length < tokenInfo.usMinPinLen || \
			data1Length > tokenInfo.usMaxPinLen )
			return( CRYPT_BADPARM4 );
		if( ( type == CRYPT_DEVICECONTROL_SET_AUTH_USER || \
			  type == CRYPT_DEVICECONTROL_SET_AUTH_SUPERVISOR ) && \
			( data1Length < tokenInfo.usMinPinLen || \
			  data1Length > tokenInfo.usMaxPinLen ) )
			return( CRYPT_BADPARM6 );
		}

	/* Handle user authorisation */
	if( type == CRYPT_DEVICECONTROL_AUTH_USER || \
		type == CRYPT_DEVICECONTROL_AUTH_SUPERVISOR )
		{
		/* If the user is already logged in, log them out before we try
		   logging in with a new authentication value */
		if( deviceInfo->loggedIn )
			{
			C_Logout( deviceInfo->deviceHandle );
			deviceInfo->loggedIn = FALSE;
			}

		/* Authenticate the user to the device */
		status = C_Login( deviceInfo->deviceHandle,
						  ( type == CRYPT_DEVICECONTROL_AUTH_USER ) ? \
						  CKU_USER : CKU_SO, ( CK_CHAR_PTR ) data1, 
						  ( CK_USHORT ) data1Length );
		if( status == CKR_OK )
			deviceInfo->loggedIn = TRUE;
		return( mapError( deviceInfo, status, CRYPT_ERROR ) );
		}

	/* Handle authorisation value change */
	if( type == CRYPT_DEVICECONTROL_SET_AUTH_USER || \
		type == CRYPT_DEVICECONTROL_SET_AUTH_SUPERVISOR )
		{
		status = C_SetPIN( deviceInfo->deviceHandle, ( CK_CHAR_PTR ) data1, 
						   ( CK_USHORT ) data1Length, ( CK_CHAR_PTR ) data2, 
						   ( CK_USHORT ) data2Length );
		return( mapError( deviceInfo, status, CRYPT_ERROR ) );
		}

	/* Handle zeroisation */
	if( type == CRYPT_DEVICECONTROL_ZEROISE )
		{
		CK_CHAR label[ 32 ];

		memset( label, ' ', 32 );
		status = C_InitToken( 0, ( CK_CHAR_PTR ) data1, 
							  ( CK_USHORT ) data1Length, label );
		return( mapError( deviceInfo, status, CRYPT_ERROR ) );
		}

	/* Anything else isn't handled */
	return( CRYPT_BADPARM2 );
	}

/****************************************************************************
*																			*
*						 	Capability Interface Routines					*
*																			*
****************************************************************************/

/****************************************************************************
*																			*
*						 	Device Capability Routines						*
*																			*
****************************************************************************/

/* The capability information for this device */

#define bits(x)	bitsToBytes(x)

static const CAPABILITY_INFO FAR_BSS capabilities[] = {
	/* The end-of-list marker */
	{ CRYPT_ALGO_NONE, CRYPT_MODE_NONE, CRYPT_ERROR, "", "",
		0, 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, CRYPT_ERROR }
	};

/* Get the capability information for a given algorithm and mode */

static int findCapabilityFunction( DEVICE_INFO *deviceInfo,
								   const void FAR_BSS **capabilityInfoPtr,
								   const CRYPT_ALGO cryptAlgo,
								   const CRYPT_MODE cryptMode )
	{
	int index, status = CRYPT_NOALGO;

	UNUSED( deviceInfo );

	/* Find the capability corresponding to the requested algorithm/mode */
	for( index = 0; capabilities[ index + 1 ].blockSize != CRYPT_ERROR; index++ )
		if( capabilities[ index ].cryptAlgo == cryptAlgo )
			{
			status = CRYPT_NOMODE;
			if( capabilities[ index ].cryptMode == cryptMode || \
				cryptMode == CRYPT_UNUSED )
				{
				*capabilityInfoPtr = &capabilities[ index ];
				status = CRYPT_OK;
				break;
				}
			}
	return( status );
	}

/* Create a context using the capabilities of the device */

static int createContextFunction( DEVICE_INFO *deviceInfo,
								  CRYPT_CONTEXT *cryptContext,
								  const CRYPT_ALGO cryptAlgo,
								  const CRYPT_MODE cryptMode )
	{
	const CAPABILITY_INFO *capabilityInfoPtr;
	int status;

	UNUSED( deviceInfo );

	/* Find the capability corresponding to the requested algorithm/mode */
	status = findCapabilityFunction( deviceInfo,
				( const void ** ) &capabilityInfoPtr, cryptAlgo, cryptMode );
	if( cryptStatusError( status ) )
		return( status );

	/* Create the context */
	return( createContext( cryptContext, capabilityInfoPtr, NULL, 0 ) );
	}

/****************************************************************************
*																			*
*						 	Device Access Routines							*
*																			*
****************************************************************************/

/* Set up the function pointers to the device methods */

int setDevicePKCS11( DEVICE_INFO *deviceInfo )
	{
#ifdef __WINDOWS__
	/* Load the PKCS #11 driver DLL's if they aren't already loaded */
	if( hPKCS11 == NULL_HINSTANCE )
		{
		deviceInitPKCS11();
		if( hPKCS11 == NULL_HINSTANCE )
			return( CRYPT_DATA_OPEN );
		}
#else
	/* Initialise the PKCS #11 library */
	if( C_Initialize( NULL_PTR ) != CKR_OK )
		return( CRYPT_DATA_OPEN );
#endif /* __WINDOWS__ */

	deviceInfo->initDeviceFunction = initDeviceFunction;
	deviceInfo->shutdownDeviceFunction = shutdownDeviceFunction;
	deviceInfo->controlFunction = controlFunction;
	deviceInfo->findCapabilityFunction = findCapabilityFunction;
	deviceInfo->createContextFunction = createContextFunction;

	return( CRYPT_OK );
	}
