/* $Id: amain.c,v 1.1 1998/12/07 20:42:30 fubobb Exp $ */

#include "arpc/misc.h"
#include "arpc/queue.h"
#include "arpc/dns.h"
#include "fdlim.h"

#ifdef UPT
#define arpc_select upt_select
#else
#define arpc_select select
#endif

#define FD_SETSIZE_ROUND (sizeof (long))/* # Bytes to which to round fd_sets */
int fd_set_bytes;                       /* Size in bytes of a [wide] fd_set */
int maxfd;
jmp_buf amain_jmp;
int amain_jmpset;

/* Callback structure for timeouts */
struct tmo {
  time_t t_time;                      /* Next timeout time */
  void (*t_f) (void *);               /* Timeout function */
  void *t_a;                          /* Function argument */
  LIST_ENTRY (tmo) t_link;            /* Linked-list entry */
};
LIST_HEAD (tmo_list, tmo);
static struct tmo_list tmos;

/* Callback structure for child processes */
struct chld {
  LIST_ENTRY (chld) c_link;
  pid_t c_pid;
  void (*c_f) (void *, int);
  void *c_a;
};
LIST_HEAD (chld_list, chld);
static struct chld_list chlds;

/* Callback structure for asynchronous events */
struct cb {
  u_short cb_fd;                      /* Actual file descriptor for callback */
  u_char cb_w;                        /* 1 if we select for writing */
  LIST_ENTRY(cb) cb_p;                /* Linked list in hash table */
  void (*cb_f) (void *);              /* Function to call back */
  void *cb_a;                         /* Argument to pass */
};
LIST_HEAD (cb_list, cb);

static fd_set *rfdsp, *wfdsp, *tfdsp;
static int chld_waiting;
static struct timeval selwait;

#define CBHSZ (1<<7)
#define cbhash(fd) ((fd) & (CBHSZ-1))
static struct cb_list cbh[CBHSZ];

int
readytowrite (int fd)
{
  static struct timeval tv;
  int n;

  FD_SET (fd, tfdsp);
  n = arpc_select (fd + 1, NULL, tfdsp, NULL, &tv);
  FD_CLR (fd, tfdsp);
  if (n < 0)
    warn ("writeable: %s\n", strerror (errno));
  return n > 0;
}

void
chld_add (pid_t pid, void (*cb) (void *, int), void *arg)
{
  struct chld *chp = xmalloc (sizeof (*chp));
  chp->c_pid = pid;
  chp->c_f = cb;
  chp->c_a = arg;
  LIST_INSERT_HEAD (&chlds, chp, c_link);
}

static inline struct cb *
cb_lookup (int fd, int write)
{
  struct cb *c;

  for (c = cbh[cbhash (fd)].lh_first;
       c && (c->cb_fd != fd || c->cb_w != write);
       c = c->cb_p.le_next)
    ;
  return (c);
}

void
cb_enable (int fd, int write, int enable)
{
  assert (cb_lookup (fd, write));
  if (enable)
    FD_SET (fd, write ? wfdsp : rfdsp);
  else
    FD_CLR (fd, write ? wfdsp : rfdsp);
}

void
cb_free (int fd, int write)
{
  struct cb *c;

  c = cb_lookup (fd, write);
  if (c) {
    LIST_REMOVE (c, cb_p);
    xfree (c);
  }
  FD_CLR (fd, write ? wfdsp : rfdsp);
}

/* Register callbacks on a particular file descriptor */
void
cb_add (int fd, int write, void (*f)(void *), void *a)
{
  struct cb *c;

  if (fd >= maxfd)
    fatal ("cb_add: file descriptor too large!\n");
  c = cb_lookup (fd, write);
  if (! c) {
    c = xmalloc (sizeof (*c));
    c->cb_fd = fd;
    c->cb_w = write;
    LIST_INSERT_HEAD (&cbh[cbhash (fd)], c, cb_p);
  }
  c->cb_f = f;
  c->cb_a = a;
  FD_SET (fd, write ? wfdsp : rfdsp);
}

static struct timeval *
tmo_next (void)
{
  static struct timeval tv;
  struct tmo *tp = tmos.lh_first;

  if (tp) {
    time_t now = time (NULL);
    tv.tv_usec = 0;
    tv.tv_sec = (tp->t_time > now) ? tp->t_time - now : 0;
  }
  else {
    /* Can't actually return NULL, as that would mess up sigchld.  So
     * return 10 minutes which should be fine. */
    tv.tv_sec = 600;
    tv.tv_usec = 0;
  }
  return &tv;
}

void
tmo_add (time_t time, void (*f)(void *), void *a)
{
  struct tmo *t, *tp, *tn;

  t = xmalloc (sizeof (*t));
  t->t_time = time;
  t->t_f = f;
  t->t_a = a;

  tp = tmos.lh_first;
  if (!tp || time <= tp->t_time) {
    LIST_INSERT_HEAD (&tmos, t, t_link);
    return;
  }
  for (;;) {
    tn = tp->t_link.le_next;
    if (!tn || time <= tn->t_time)
      break;
    tp = tn;
  }
  LIST_INSERT_AFTER (tp, t, t_link);
}

void
printaddrs (void *null, struct hostent *h)
{
  char **aliases;
  struct in_addr **addrs;

  if (! h) {
    printf ("error\n");
    return;
  }

  aliases = h->h_aliases;
  addrs = (struct in_addr **) h->h_addr_list;

  printf ("Name: %s\n", h->h_name);
  while (*aliases)
    printf ("Alias: %s\n", *aliases++);
  while (*addrs)
    printf ("Address: %s\n", inet_ntoa (**addrs++));
}

static void
chld_check (void)
{
  for (;;) {
    struct chld *chp;
    pid_t pid;
    int status;

    chld_waiting = 0;
#ifdef HAVE_WAITPID
    pid = waitpid (0, &status, WNOHANG);
#else
    pid = wait3 (&status, WNOHANG, NULL);
#endif
    if (pid == 0 || pid == -1)
      return;

    for (chp = chlds.lh_first; chp; chp = chp->c_link.le_next) {
      if (pid == chp->c_pid) {
	LIST_REMOVE (chp, c_link);
	chp->c_f (chp->c_a, status);
	xfree (chp);
	break;
      }
    }
  }
}

/* Select for I/O events and perform the callbacks. */
static void
cb_check (void)
{
    static fd_set *rp;
    static fd_set *wp;
    int n, i;

    if (!rp) {
      rp = xmalloc (fd_set_bytes);
      wp = xmalloc (fd_set_bytes);
    }
    memcpy (rp, rfdsp, fd_set_bytes);
    memcpy (wp, wfdsp, fd_set_bytes);
    selwait = *tmo_next ();
    if (chld_waiting)
      selwait.tv_sec = selwait.tv_usec = 0;
    n = arpc_select (maxfd, rp, wp, NULL, &selwait);
    if (n < 0) {
      if (errno != EINTR) {
	int oerrno = errno;
	warn ("select: %s\n", strerror (errno));
	if (oerrno == EBADF) {
	  int i;

	  warn ("rfdsp:");
	  for (i = 0; i < maxfd; i++)
	    if (FD_ISSET (i, rfdsp))
	      ferr (" %d", i);
	  ferr ("\n");

	  warn ("wfdsp:");
	  for (i = 0; i < maxfd; i++)
	    if (FD_ISSET (i, wfdsp))
	      ferr (" %d", i);
	  ferr ("\n");

	  err_flush ();

	  assert (0);
	}
      }
      return;
    }
    for (i = 0; n && i < maxfd; i++) {
      struct cb *c;
      if (FD_ISSET (i, rp)) {
	n--;
	c = cb_lookup (i, 0);
	if (c)
	  c->cb_f (c->cb_a);
	else if (FD_ISSET (i, rfdsp))
	  fatal ("main: lost a reader!\n");
      }
      if (FD_ISSET (i, wp)) {
	n--;
	c = cb_lookup (i, 1);
	if (c)
	  c->cb_f (c->cb_a);
	else if (FD_ISSET (i, wfdsp))
	  fatal ("main: lost a writer!\n");
      }
    }
}

static void
tmo_check (void)
{
  static time_t last;
  time_t now;
  struct tmo *tp, *tn;

  if (time (&now) == last)
    return;
  last = now;
  tp = tmos.lh_first; 
  while (tp && tp->t_time <= now) {
    tn = tp->t_link.le_next;
    LIST_REMOVE (tp, t_link);
    tp->t_f (tp->t_a);
    xfree (tp);
    tp = tn;
  }
}

void
acheck (void)
{
  assert (maxfd);
  cb_check ();
  tmo_check ();
  chld_check ();
}

void
amain (void)
{
  assert (maxfd);
  for (;;) {
    cb_check ();
    if (!amain_jmpset) {
      amain_jmpset = 1;
      setjmp (amain_jmp);
    }
    tmo_check ();
    chld_check ();
  }
}

static void
sigchld (int sig)
{
  chld_waiting = 1;
  selwait.tv_sec = selwait.tv_usec = 0;
}

void
ainit (void)
{
  struct sigaction sa;
  sigset_t sm;

  err_init ();

  sigemptyset(&sm);
  sigaddset(&sm, SIGPIPE);  
  sigprocmask (SIG_BLOCK, &sm, NULL);

  sa.sa_handler = sigchld;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_NOCLDSTOP;
  sigaction (SIGCHLD, &sa, NULL);

#ifndef HAVE_WIDE_SELECT
  fdlim_set (min (FD_SETSIZE, fdlim_get (1)));
  maxfd = (fdlim_get (0));
  fd_set_bytes = sizeof (fd_set);
#else /* HAVE_WIDE_SELECT */
  fdlim_set (fdlim_get (1));
  maxfd = (fdlim_get (0));
  fd_set_bytes = (maxfd+7)/8;
  if (fd_set_bytes % FD_SETSIZE_ROUND)
    fd_set_bytes += FD_SETSIZE_ROUND - (fd_set_bytes % FD_SETSIZE_ROUND);
#endif /* HAVE_WIDE_SELECT */
  rfdsp = xmalloc (fd_set_bytes);
  bzero (rfdsp, fd_set_bytes);
  wfdsp = xmalloc (fd_set_bytes);
  bzero (wfdsp, fd_set_bytes);
  tfdsp = xmalloc (fd_set_bytes);
  bzero (tfdsp, fd_set_bytes);

  ar_init ();

  signal (SIGCHLD, sigchld);
}
