/* $Id: arpcclient.c,v 1.1 1998/12/07 20:42:34 fubobb Exp $ */

#include "arpc/arpc.h"
#include "backoff.h"
#include "xdr_suio.h"
#include "dns.h"

struct rpccb {
  u_int rb_xid;                /* XID for this call */
  struct aclnt *rb_clnt;       /* Client structure for this call */
  char *rb_rawin;              /* Raw call data */
  u_int rb_inlen;              /* Length of call data */
  xdrproc_t rb_outproc;        /* XDR for call result */
  void *rb_out;                /* Return structure */
  clnt_cb rb_callback;         /* Callback when call complete */
  void *rb_arg;                /* First argument to callback function */
  LIST_ENTRY (rpccb) rb_link;  /* linked list */
  LIST_ENTRY (rpccb) rb_clink; /* Per client link */
  struct sockaddr_in rb_sa;    /* Remote socket address if not connected */
  TMOQ_ENTRY (rpccb) rb_tmo;   /* Retry queue if not reliable */
};
LIST_HEAD (rpccb_list, rpccb);

struct aclnt_xprt {
  void (*acx_eofcb) (void *);  /* Callback for end of file/error */
  void *acx_eofarg;            /* Argument to callback */
  struct axh *acx_xh;          /* Transport handle */
  char acx_reliable;           /* Transport is reliable */
  char acx_eof;                /* Transport has failed */
};
struct aclnt_udp {
  struct sockaddr_in acu_addr; /* Socket address for sending */
};
struct aclnt {
  struct rpccb_list ac_rpccbs; /* List of rpccbs for this client */
  u_int32_t ac_prog;           /* RPC Program number */
  u_int32_t ac_vers;           /* RPC Program version */
  struct rpcgen_table *ac_tbl; /* XDR table */
  int ac_nproc;                /* Size of xdr table */
  enum {AC_UDP, AC_XPRT} ac_type;
  union {
    struct aclnt_xprt x;
    struct aclnt_udp u;
  } ac_un;
#define ac_xh ac_un.x.acx_xh
#define ac_eof ac_un.x.acx_eof
#define ac_addr ac_un.u.acu_addr
};

#define aclnt_xh(c) (c && c->ac_type == AC_XPRT ? c->ac_un.x.acx_xh : udpxprt)
#define rpccb_xh(rp) (aclnt_xh (rp->rb_clnt))
#define aclnt_isconnected(c) ((c) && (c)->ac_type == AC_XPRT)
#define rpccb_isreliable(rp)				\
  ((rp)->rb_clnt && (rp)->rb_clnt->ac_type == AC_XPRT	\
   && (rp)->rb_clnt->ac_un.x.acx_reliable)

static u_short sunrpc_port;    /* Portmap port in network byte order */
static AUTH *auth_none;        /* Default (NULL) auth */
struct axh *udpxprt;           /* Default (UDP) transport */

static u_int32_t xidctr;       /* Last used XID */

#define NXIDBUCKETS (1<<4)
#define xidhash(x) ((x) & (NXIDBUCKETS-1))

static struct rpccb_list rb_free;
static struct rpccb_list rb_hash[NXIDBUCKETS];

static TMOQ_HEAD(rpccb) rtoq;

void
arpcclient_init (void)
{
  int udpsock;
  struct servent *sp;

  xidctr = (u_int32_t) time (NULL) ^ (getpid () << 16);
  if ((sp = getservbyname ("sunrpc", "udp")))
    sunrpc_port = sp->s_port;
  else
    sunrpc_port = htons (PMAPPORT);
  TMOQ_INIT (rtoq);
  auth_none = authnone_create ();
  udpsock = inetsocket_resvport (SOCK_DGRAM, INADDR_ANY);
  if (udpsock < 0)
    fatal ("inetsocket: %s\n", strerror (errno));
  close_on_exec (udpsock);
  make_async (udpsock);
  udpxprt = dx_alloc (udpsock);
  xh_setcb (udpxprt, aclnt_reap, NULL);
}

static inline struct rpccb *
arpc_lookup (u_int32_t xid, struct axh *xh)
{
  static struct rpccb *rp;

  for (rp = rb_hash[xidhash (xid)].lh_first; rp; rp = rp->rb_link.le_next)
    if (rp->rb_xid == xid && rpccb_xh(rp) == xh)
      return rp;
  return NULL;
}

static inline void
arpc_send (struct rpccb *rp)
{
  xh_send (rpccb_xh (rp), rp->rb_rawin, rp->rb_inlen, &rp->rb_sa);
}

static void
arpc_free (struct rpccb *rp)
{
  LIST_REMOVE (rp, rb_link);
  if (rp->rb_clnt)
    LIST_REMOVE (rp, rb_clink);
  if (!rpccb_isreliable (rp))
    TMOQ_REMOVE (rtoq, rp, rb_tmo);
  xfree (rp->rb_rawin);
  LIST_INSERT_HEAD (&rb_free, rp, rb_link);
}

static inline void
arpc_purge (struct rpccb *rp, enum clnt_stat err)
{
  clnt_cb cb = rp->rb_callback;
  void *arg = rp->rb_arg;
  void *res = rp->rb_out;
  arpc_free (rp);
  cb (arg, err, res);
}

static inline void
arpc_expire (struct rpccb *rp)
{
  arpc_purge (rp, RPC_TIMEDOUT);
}
static void
arpc_timeout (void *arg)
{
  TMOQ_RUN (rtoq, rpccb, rb_tmo, time (NULL), arpc_send, arpc_expire,
	    arpc_timeout, arg);
}

static inline void
arpc_marshall (u_int32_t xid, AUTH *auth,
	       u_int32_t prognum, u_int32_t versnum, u_int32_t procnum,
	       xdrproc_t inproc, void *in,
	       char **inraw, int *rawlen)
{
  u_int32_t *dp;
  XDR x;

  xdrsuio_create (&x, XDR_ENCODE);

  dp = xdrsuio_inline (&x, 6*4);
  *dp++ = htonl (xid);
  *dp++ = htonl (CALL);
  *dp++ = htonl (RPC_MSG_VERSION);
  *dp++ = htonl (prognum);
  *dp++ = htonl (versnum);
  *dp++ = htonl (procnum);
  if (!AUTH_MARSHALL (auth ? auth : auth_none, &x))
    fatal ("failed to marshall auth crap\n");
  if (!inproc (&x, in))
    fatal ("arg marshalling failed (prog %d, vers %d, proc %d)\n",
	   prognum, versnum, prognum);
  *rawlen = xsuio (&x)->uio_resid;
  *inraw = suio_flatten (xsuio (&x));

  xdr_destroy (&x);
}

static struct rpccb *
arpc_alloc (struct aclnt *c, struct sockaddr_in *dest, AUTH *auth,
	   u_int32_t prognum, u_int32_t versnum, u_int32_t procnum,
	   xdrproc_t inproc, void *in, xdrproc_t outproc, void *out,
	   clnt_cb cb, void *arg)
{
  struct rpccb *rp;

  if ((rp = rb_free.lh_first)) {
    LIST_REMOVE (rp, rb_link);
  }
  else
    rp = xmalloc (sizeof (*rp));

  rp->rb_xid = xidctr++;
  rp->rb_clnt = c;
  rp->rb_outproc = outproc;
  rp->rb_out = out;
  rp->rb_callback = cb;
  rp->rb_arg = arg;

  if (!c) {
    if (dest)
      rp->rb_sa = *dest;
    else
      /* Means caller doesn't yet know the destination address, but in
       * arguments are on the stack and thus must be marshalled now.
       * A sin_family of 0 will ensure that bad packets never trick us
       * into "reaping" this rpccb before the caller has set the
       * address and arpc_start'ed it. */
      bzero (&rp->rb_sa, sizeof (rp->rb_sa));
  }
  else if (c->ac_type == AC_UDP)
    rp->rb_sa = c->ac_addr;

  arpc_marshall (rp->rb_xid, auth, prognum, versnum, procnum,
		 inproc, in, &rp->rb_rawin, &rp->rb_inlen);

  LIST_INSERT_HEAD (&rb_hash[xidhash (rp->rb_xid)], rp, rb_link);
  if (c)
    LIST_INSERT_HEAD (&c->ac_rpccbs, rp, rb_clink);

  return rp;
}

static void
arpc_start (struct rpccb *rp)
{
  if (!rpccb_isreliable (rp))
    TMOQ_ADD (rtoq, rpccb, rp, rb_tmo, arpc_timeout);
  arpc_send (rp);
}

static void
arpc_call (struct aclnt *c, struct sockaddr_in *dest, AUTH *auth,
	   u_int32_t prognum, u_int32_t versnum, u_int32_t procnum,
	   xdrproc_t inproc, void *in, xdrproc_t outproc, void *out,
	   clnt_cb cb, void *arg)
{
  struct rpccb *rp;

  rp = arpc_alloc (c, dest, auth, prognum, versnum, procnum,
		   inproc, in, outproc, out, cb, arg);
  arpc_start (rp);
}

/* Purge old calls for a client */
static inline void
aclnt_purge (struct aclnt *c)
{
  struct rpccb *rp;

  c->ac_eof = 1;
  if (aclnt_isconnected (c))
    while ((rp = c->ac_rpccbs.lh_first))
      arpc_purge (rp, RPC_CANTRECV);
}

static inline void
aclnt_eof (struct aclnt *c)
{
  if (c && c->ac_type == AC_XPRT && !c->ac_eof) {
    aclnt_purge (c);
    if (c->ac_un.x.acx_eofcb)
      c->ac_un.x.acx_eofcb (c->ac_un.x.acx_eofarg);
  }
}

void
aclnt_reap (void *_c, char *out, int len, void *_sap, int sal)
{
  struct aclnt *c = _c;
  struct sockaddr_in *sap = _sap;
  int xid;
  struct rpccb *rp;
  XDR x;
  struct rpc_msg m;
  int res;
  struct rpc_err re;
  clnt_cb cb;
  void *arg;
  xdrproc_t outproc;
  void *outmem;

  if (len <= 0) {
    aclnt_eof (c);
    return;
  }
  xid = ntohl (*((u_int32_t *) out));

  if (!(rp = arpc_lookup (xid, aclnt_xh (c))))
    return;

  if ((!c || c->ac_type != AC_XPRT)
      && (sal != sizeof (rp->rb_sa)
	  || sap->sin_family != rp->rb_sa.sin_family
	  || sap->sin_port != rp->rb_sa.sin_port
	  || sap->sin_addr.s_addr != rp->rb_sa.sin_addr.s_addr)) {
    warn ("right XID, wrong IP address!\n");
    return;
  }

  xdrmem_create (&x, out, len, XDR_DECODE);
  m.acpted_rply.ar_verf = _null_auth; 
  m.acpted_rply.ar_results.where = rp->rb_out;
  m.acpted_rply.ar_results.proc = rp->rb_outproc;

#ifdef DMALLOC_EXTRA_DISCIPLINE
  dmalloc_verify (NULL);
#endif /* DMALLOC */
  res = xdr_replymsg (&x, &m);
#ifdef DMALLOC_EXTRA_DISCIPLINE
  dmalloc_verify (NULL);
#endif /* DMALLOC */
  xdr_destroy (&x);

  if (!res)
    re.re_status = RPC_CANTDECODERES;
  else
#ifdef HAVE___SETERR_REPLY
    __seterr_reply (&m, &re);
#else
    _seterr_reply (&m, &re);
#endif

  /* XXX yuck */
  cb = rp->rb_callback;
  arg = rp->rb_arg;
  outproc = rp->rb_outproc;
  outmem = rp->rb_out;
  arpc_free (rp);
  if (re.re_status == RPC_SUCCESS) {
    switch (cb (arg, re.re_status, outmem)) {
    case ACFREE_XDR:
      xdr_free (outproc, outmem);
      break;
    case ACFREE_ALL:
      xdr_free (outproc, outmem);
      xfree (outmem);
    }
  }
  else
    cb (arg, re.re_status, outmem);
}

struct pmcb {
  u_short port;
  void (*cb) (void *, u_short);
  void *arg;
};
static int
apmap_gotport (void *_pc, enum clnt_stat cs, void *_portp)
{
  struct pmcb * const pc = _pc;
  pc->cb (pc->arg, pc->port);
  xfree (pc);
  return (0);
}
void
apmap_getport (struct in_addr ipaddr, u_int32_t program, u_int32_t version,
	       u_int protocol, void (*cb) (void *, u_short), void *arg)
{
  struct sockaddr_in sa;
  struct pmcb *pc;
  struct pmap pm;

  pc = xmalloc (sizeof (*pc));
  pc->port = 0;
  pc->cb = cb;
  pc->arg = arg;

  bzero (&sa, sizeof (sa));
  sa.sin_family = AF_INET;
  sa.sin_port = sunrpc_port;
  sa.sin_addr = ipaddr;

  pm.pm_prog = program;
  pm.pm_vers = version;
  pm.pm_prot = protocol;
  pm.pm_port = 0;

  arpc_call (NULL, &sa, NULL,
	     PMAPPROG, PMAPVERS, PMAPPROC_GETPORT,
	     (xdrproc_t) xdr_pmap, &pm, 
             (xdrproc_t) xdr_u_short, &pc->port,
	     apmap_gotport, pc);
}

struct pmas {
  struct sockaddr_in sin;
  u_int32_t prog;
  u_int32_t vers;
  u_int prot;
  void (*cb) (void *, enum clnt_stat, struct sockaddr_in *);
  void *arg;
};
static void
apmap_getaddr_port (void *_st, u_short port)
{
  struct pmas *st = _st;

  if (port) {
    st->sin.sin_port = htons (port);
    st->cb (st->arg, RPC_SUCCESS, &st->sin);
  }
  else
    st->cb (st->arg, RPC_PROGNOTREGISTERED, NULL);
  xfree (st);
  return;

}
static void
apmap_getaddr_inaddr (void *_st, struct hostent *h)
{
  struct pmas *st = _st;

  if (!h) {
    st->cb (st->arg, RPC_UNKNOWNHOST, NULL);
    xfree (st);
    return;
  }

  st->sin.sin_addr = *(struct in_addr *) h->h_addr;

  if (ntohs (st->sin.sin_port)) {
    st->cb (st->arg, RPC_SUCCESS, &st->sin);
    xfree (st);
    return;
  }

  apmap_getport (st->sin.sin_addr, st->prog, st->vers, st->prot,
		 apmap_getaddr_port, st);
}
static void
apmap_getaddr (char *host, u_short port,
	       u_int32_t prog, u_int32_t vers, u_int prot,
	       void (*cb) (void *, enum clnt_stat, struct sockaddr_in *),
	       void *arg)
{
  struct pmas *st;

  st = xmalloc (sizeof (*st));
  bzero (&st->sin, sizeof (st->sin));
  st->sin.sin_family = AF_INET;
  st->sin.sin_port = htons (port);
  st->prog = prog;
  st->vers = vers;
  st->prot = prot;
  st->cb = cb;
  st->arg = arg;

  ar_hostbyname (host, apmap_getaddr_inaddr, st); 
}

/* Free an ARPC client.  If there are outstanding RPCs, they will all
 * return with failure before aclnt_free() returns, so that an eof
 * callback can enusre there are no outstanding RPCs before freeing
 * memory associated with a connection.  */
void
aclnt_free (struct aclnt *c)
{
  if (c->ac_type == AC_XPRT) {
    xh_free (c->ac_xh);
    c->ac_xh = NULL;
  }
  aclnt_purge (c);
  xfree (c);
}
/* Like free, but save and return transport handle for reuse. */
struct axh *
aclnt_reclaim (struct aclnt *c)
{
  struct axh *xh;
  assert (c->ac_type == AC_XPRT);
  xh = c->ac_xh;
  c->ac_xh = NULL;
  aclnt_purge (c);
  xfree (c);
  return xh;
}

void
aclnt_call_xdrproc (struct aclnt *c, AUTH *auth, u_int32_t procnum,
		    xdrproc_t inproc, void *in, xdrproc_t outproc, void *out,
		    clnt_cb cb, void *arg)
{
  if (c->ac_type == AC_XPRT && c->ac_eof)
    cb (arg, RPC_CANTSEND, NULL);
  else
    arpc_call (c, NULL, auth, c->ac_prog, c->ac_vers, procnum,
	       inproc, in, outproc, out, cb, arg);
}

void
aclnt_call (struct aclnt *c, AUTH *auth, u_int32_t procnum,
	    void *in, void *out, clnt_cb cb, void *arg)
{
  assert (procnum < c->ac_nproc);
  if (c->ac_type == AC_XPRT && c->ac_eof)
    cb (arg, RPC_CANTSEND, NULL);
  else
    arpc_call (c, NULL, auth, c->ac_prog, c->ac_vers, procnum,
	       c->ac_tbl[procnum].xdr_arg, in,
	       c->ac_tbl[procnum].xdr_res, out,
	       cb, arg);
}

struct aclnt *
aclnt_rawudp_create (struct sockaddr_in *sinp,
		     u_int32_t prog, u_int32_t vers,
		     struct rpcgen_table *xdr_tbl, u_int nproc)

{
  struct aclnt *c = xmalloc (sizeof (*c));

  LIST_INIT (&c->ac_rpccbs);
  c->ac_prog = prog;
  c->ac_vers = vers;
  c->ac_type = AC_UDP;
  c->ac_addr = *sinp;
  c->ac_tbl = xdr_tbl;
  c->ac_nproc = nproc;
  return c;
}

struct aclnt *
aclnt_axh_create (struct axh *xh, u_int32_t prog, u_int32_t vers,
		  struct rpcgen_table *xdr_tbl, u_int nproc,
		  void (*eofcb) (void *), void *eofarg)
{
  struct aclnt *c = xmalloc (sizeof (*c));

  LIST_INIT (&c->ac_rpccbs);
  c->ac_prog = prog;
  c->ac_vers = vers;
  c->ac_type = AC_XPRT;
  c->ac_eof = 0;
  c->ac_un.x.acx_eofcb = eofcb;
  c->ac_un.x.acx_eofarg = eofarg;
  c->ac_un.x.acx_xh = xh;
  c->ac_un.x.acx_reliable = xh_isreliable (xh);
  c->ac_tbl = xdr_tbl;
  c->ac_nproc = nproc;
  xh_setcb (xh, aclnt_reap, c);
  return c;
}

void
aclnt_seteof (struct aclnt *c, void (*cb) (void *), void *arg)
{
  assert (c->ac_type == AC_XPRT);
  c->ac_un.x.acx_eofcb = cb;
  c->ac_un.x.acx_eofarg = arg;
}

void
aclnt_encrypt (struct aclnt *c, char *key, int keylen)
{
  xh_encrypt (c->ac_xh, 0, key, keylen);
}

struct aclntudpstat {
  struct aclnt *c;
  void (*cb) (void *, struct aclnt *);
  void *arg;
};
static void
aclntudp_gotaddr (void *_st, enum clnt_stat err, struct sockaddr_in *sinp)
{
  struct aclntudpstat *st = _st;
  struct aclnt *c = st->c;

  if (err) {
    st->cb (st->arg, NULL);
    xfree (st->c);
    xfree (st);
    return;
  }
  c->ac_addr = *sinp;
  st->cb (st->arg, c);
  xfree (st);
}
void
aclntudp_create (char *host, u_short port,
		 u_int32_t prog, u_int32_t vers,
		 struct rpcgen_table *xdr_tbl, u_int nproc,
		 void (*cb) (), void *arg)
{
  struct aclntudpstat *st;

  st = xmalloc (sizeof (*st));
  st->c = xmalloc (sizeof (*st->c));
  LIST_INIT (&st->c->ac_rpccbs);
  st->c->ac_prog = prog;
  st->c->ac_vers = vers;
  st->c->ac_type = AC_UDP;
  st->c->ac_tbl = xdr_tbl;
  st->c->ac_nproc = nproc;
  st->cb = cb;
  st->arg = arg;
  apmap_getaddr (host, port, st->c->ac_prog, st->c->ac_vers, IPPROTO_UDP,
		 aclntudp_gotaddr, st);
}

struct aclnttcpstat {
  struct aclnt *c;
  void (*cb) (void *, struct aclnt *);
  void *arg;
  int s;
};
static void
aclnttcp_connected (void *_st)
{
  struct aclnttcpstat *st = _st;
  struct sockaddr_in sin;
  int len = sizeof (sin);

  cb_free (st->s, 1);
  /* Our socket has selected for writing, which means that the connect
   * has completed (either successfully or not).  We call getpeername
   * to find out if it the connect actually succeeded.  It sounds
   * strange, but do you have any better ideas? */
  if (getpeername (st->s, (struct sockaddr *) &sin, &len) < 0) {
    close (st->s);
    xfree (st->c);
    st->cb (st->arg, NULL);
    xfree (st);
    return;
  }
  st->c->ac_xh = ps_alloc (st->s);
  xh_setcb (st->c->ac_xh, aclnt_reap, st->c);
  st->cb (st->arg, st->c);
  xfree (st);
}
static void
aclnttcp_gotaddr (void *_st, enum clnt_stat err, struct sockaddr_in *sinp)
{
  struct aclnttcpstat *st = _st;

  if (err) {
    xfree (st->c);
    st->cb (st->arg, NULL);
    xfree (st);
    return;
  }
  st->s = socket (AF_INET, SOCK_STREAM, 0);
  if (st->s < 0) {
    warn ("socket: %s\n", strerror (errno));
    xfree (st->c);
    st->cb (st->arg, NULL);
    xfree (st);
    return;
  }
  make_async (st->s);
  if (connect (st->s, (struct sockaddr *) sinp, sizeof (*sinp)) < 0
      && errno != EINPROGRESS) {
    warn ("connect: %s\n", strerror (errno));
    close (st->s);
    xfree (st->c);
    st->cb (st->arg, NULL);
    xfree (st);
    return;
  }
  cb_add (st->s, 1, aclnttcp_connected, st);
}
void
aclnttcp_create (char *host, u_short port,
		 u_int32_t prog, u_int32_t vers,
		 struct rpcgen_table *xdr_tbl, u_int nproc,
		 void (*eofcb) (void *), void *eofarg,
		 void (*cb) (void *, struct aclnt *), void *arg)
{
  struct aclnttcpstat *st;

  st = xmalloc (sizeof (*st));
  st->c = xmalloc (sizeof (*st->c));
  LIST_INIT (&st->c->ac_rpccbs);
  st->c->ac_prog = prog;
  st->c->ac_vers = vers;
  st->c->ac_type = AC_XPRT;
  st->c->ac_tbl = xdr_tbl;
  st->c->ac_nproc = nproc;
  st->c->ac_un.x.acx_eofcb = eofcb;
  st->c->ac_un.x.acx_eofarg = eofarg;
  st->c->ac_un.x.acx_reliable = 1;
  st->c->ac_eof = 0;
  st->cb = cb;
  st->arg = arg;
  apmap_getaddr (host, port, st->c->ac_prog, st->c->ac_vers, IPPROTO_TCP,
		 aclnttcp_gotaddr, st);
}


static void
acallrpc_2 (void *_rp, enum clnt_stat err, struct sockaddr_in *sinp)
{
  struct rpccb *rp = _rp;
  if (err) {
    rp->rb_callback (rp->rb_arg, err, rp->rb_out);
    arpc_free (rp);
  }
  else {
    rp->rb_sa = *sinp;
    arpc_start (rp);
  }
}
void
acallrpc (char *host, AUTH *auth,
	  u_int32_t prognum, u_int32_t versnum, u_int32_t procnum,
	  xdrproc_t inproc, void *in, xdrproc_t outproc, void *out,
	  int (*cb) (void *, enum clnt_stat, void *), void *arg)
{
  struct rpccb *rp;

  rp = arpc_alloc (NULL, NULL, auth, prognum, versnum, procnum,
		   inproc, in, outproc, out, cb, arg);
  apmap_getaddr (host, 0, prognum, versnum, IPPROTO_UDP,
		 acallrpc_2, rp);
}
