/* $Id: arpcserver.c,v 1.1 1998/12/07 20:42:36 fubobb Exp $ */

#include "sysconf.h"
#include "xdr_suio.h"
#include "misc.h"
#include "arpc.h"
#include "queue.h"

#define RC_MAXOLD 16
#define RC_NCHAIN 29

LIST_HEAD (svccb_list, svccb);
TAILQ_HEAD (svccb_tailq, svccb);

#define svccb_hash(sbp)					\
  (((sbp)->sb_xid | ((sbp)->sb_sin.sin_port << 16)	\
    | (sbp)->sb_sin.sin_addr.s_addr) % RC_NCHAIN)

struct rcache {
  u_int rc_nold;                    /* Number of old requests replay cache */
  struct svccb_tailq rc_lru;        /* LRU cache of old replies */
  struct svccb_list rc_hash[RC_NCHAIN];  /* Hash chains */
};

struct asvc;
struct asvcprog {
  LIST_ENTRY(asvcprog) ap_link;     /* All programs on this descriptor */
  struct asvc *ap_svc;              /* Head of list */
  u_int32_t ap_prog;                /* RPC Program number */
  u_int32_t ap_vers;                /* RPC Version number */
  struct rpcgen_table *ap_tbl;      /* Dispatch table */
  u_int ap_nproc;                   /* Dispatch table size */
};
struct asvc {
  LIST_HEAD (, asvcprog) as_progs;  /* Links to all programs */
  void (*as_disp_fn) (void *, struct svccb *);   /* Dispatch function */
  void *as_disp_arg;                /* Argument to dispatch function */
  struct axh *as_xh;                /* Transport handle */
  struct rcache *as_rcache;         /* Reply cache for unreliable transports */
  struct aclnt *as_clnt;            /* Client for bidirectional connections */
  u_int as_count;                   /* Number of outstanding calls */
};
#define sb_svc sb_svcprog->ap_svc
#define sb_tbl sb_svcprog->ap_tbl

static void asvc_dispatch (void *, char *, int len, void *, int);

static void
asvc_freesvccb (struct svccb *sbp)
{
  struct rpcgen_table *rtp = &sbp->sb_tbl[sbp->sb_procno];

  sbp->sb_svc->as_count--;
  if (sbp->sb_svc->as_rcache) {
    LIST_REMOVE (sbp, sb_hlink);
    if (sbp->sb_rdat)
      TAILQ_REMOVE (&sbp->sb_svc->as_rcache->rc_lru, sbp, sb_tlink);
  }
  xdr_free ((xdrproc_t) xdr_callmsg, (void *) sbp->sb_msg);
  xfree (sbp->sb_msg);
  xdr_free (rtp->xdr_arg, sbp->sb_arg);
  xfree (sbp->sb_arg);
  if (sbp->sb_aup) {
    xdr_free ((xdrproc_t) xdr_authunix_parms, (void *) sbp->sb_aup);
    xfree (sbp->sb_aup);
  }
  xfree (sbp->sb_rdat);
  xfree (sbp);
}

struct asvc *
asvc_create (struct axh *xh,
	     void (*cb) (void *, struct svccb *), void *arg)
{
  struct asvc *ap = xmalloc (sizeof (*ap));

  LIST_INIT (&ap->as_progs);
  ap->as_disp_fn = cb;
  ap->as_disp_arg = arg;
  ap->as_xh = xh;
  if (xh_isreliable (ap->as_xh))
    ap->as_rcache = NULL;
  else {
    ap->as_rcache = xmalloc (sizeof (*ap->as_rcache));
    bzero (ap->as_rcache, sizeof (*ap->as_rcache));
    TAILQ_INIT (&ap->as_rcache->rc_lru);
  }
  ap->as_clnt = NULL;
  ap->as_count = 0;
  return ap;
}

static struct asvcprog *
asvc_lookup (struct asvc *ap, u_int32_t prog, u_int32_t vers)
{
  struct asvcprog *pp;
  for (pp = ap->as_progs.lh_first;
       pp && pp->ap_prog != prog && pp->ap_vers != vers;
       pp = pp->ap_link.le_next)
    ;
  return pp;
}

void
asvc_serve (struct asvc *ap,
	    u_int32_t prog, u_int32_t vers,
	    struct rpcgen_table *xdr_tbl, u_int nproc)
{
  struct asvcprog *pp;
  pp = asvc_lookup (ap, prog, vers);
  if (!pp) {
    pp = xmalloc (sizeof (*pp));
    LIST_INSERT_HEAD (&ap->as_progs, pp, ap_link);
    pp->ap_svc = ap;
    pp->ap_prog = prog;
    pp->ap_vers = vers;
  }
  pp->ap_tbl = xdr_tbl;
  pp->ap_nproc = nproc;
  xh_setcb (ap->as_xh, asvc_dispatch, ap);
}

struct aclnt *
asvc_mkclnt (struct asvc *ap, u_int32_t prog, u_int32_t vers,
	     struct rpcgen_table *xdr_tbl, u_int nproc)
{
  if (ap->as_clnt)
    aclnt_reclaim (ap->as_clnt);
  ap->as_clnt = aclnt_axh_create (ap->as_xh, prog, vers,
				  xdr_tbl, nproc, NULL, NULL);
  /* Don't let the client code steal the input callback. */
  xh_setcb (ap->as_xh, asvc_dispatch, ap);
  return ap->as_clnt;
}

void
asvc_free (struct asvc *ap)
{
  if (ap->as_clnt) {
    aclnt_free (ap->as_clnt);
    ap->as_clnt = NULL;
  }
  else if (ap->as_xh)
    xh_free (ap->as_xh);
  ap->as_xh = NULL;
  if (ap->as_rcache)
    while (ap->as_rcache->rc_lru.tqh_first)
      asvc_freesvccb (ap->as_rcache->rc_lru.tqh_first);
  if (ap->as_count == 0) {
    struct asvcprog *pp, *npp;
    if (ap->as_rcache)
      xfree (ap->as_rcache);
    for (pp = ap->as_progs.lh_first; pp; pp = npp) {
      npp = pp->ap_link.le_next;
      xfree (pp);
    }
    xfree (ap);
  }
}

void
asvc_encrypt (struct asvc *ap, char *key, int keylen)
{
  xh_encrypt (ap->as_xh, 1, key, keylen);
}

/* For the "client" end of a bidirectional connection. */
void
asvc_encrypt_client (struct asvc *ap, char *key, int keylen)
{
  xh_encrypt (ap->as_xh, 0, key, keylen);
}

static inline int
asvc_replaycheck (struct svccb *sbp)
{
  struct svccb *s;

  if (sbp->sb_svc->as_rcache) {
    for (s = sbp->sb_svc->as_rcache->rc_hash[svccb_hash (sbp)].lh_first;
	 s; s = s->sb_hlink.le_next)
      if (s->sb_xid == sbp->sb_xid
	  && s->sb_sin.sin_port == sbp->sb_sin.sin_port
	  && s->sb_sin.sin_addr.s_addr == sbp->sb_sin.sin_addr.s_addr) {
	warn ("Replay of XID 0x%x proc %ld vers %ld prog %ld\n", 
	      (u_int) s->sb_xid, (long) sbp->sb_msg->rm_call.cb_proc,
	      (long) sbp->sb_msg->rm_call.cb_vers, 
	      (long) sbp->sb_msg->rm_call.cb_prog);
	if (s->sb_rdat)
	  xh_send (s->sb_svc->as_xh, s->sb_rdat, s->sb_rlen, &s->sb_sin);
	return 1;
      }
    LIST_INSERT_HEAD (&sbp->sb_svc->as_rcache->rc_hash[svccb_hash (sbp)],
		      sbp, sb_hlink);
  }
  return 0;
}

static void
asvc_dispatch (void *_ap, char *buf, int len, void *sa, int salen)
{
  struct asvc *const ap = _ap;
  struct asvcprog *pp;
  struct svccb *sbp;
  struct rpc_msg *m;
  struct rpcgen_table *rtp;
  XDR x;

  if (len <= 0)
    goto error;

  /* Check for a call if this is a bidirectional asvc and we got a reply */
  if (ap->as_clnt && len >= 8
      && ((u_int32_t *) buf)[1] == htonl (REPLY)) {
    if (ap->as_clnt)
      aclnt_reap (ap->as_clnt, buf, len, sa, salen);
    else
      warn ("asvc_dispatch: reply on non-client transport\n");
    return;
  }

  sbp = xmalloc (sizeof *sbp);
  m = xmalloc (sizeof *m);

  xdrmem_create (&x, buf, len, XDR_DECODE);
  bzero (m, sizeof (*m));
  if (! xdr_callmsg (&x, (void *) m)) {
    warn ("asvc_dispatch: bad RPC message\n");
    goto freexdr;
  }

  if (m->rm_call.cb_rpcvers != RPC_MSG_VERSION) {
    warn ("asvc_dispatch: bad RPC_MSG_VERSION\n");
    goto freemsg;
  }
  pp = asvc_lookup (ap, m->rm_call.cb_prog, m->rm_call.cb_vers);
  if (!pp) {
    warn ("asvc_dispatch: bad prog/vers (%ld/%ld)\n",
	  (long) m->rm_call.cb_prog, (long) m->rm_call.cb_vers);
    goto freemsg;
  }
  if (m->rm_call.cb_proc >= pp->ap_nproc) {
    warn ("asvc_dispatch: bad proc (%ld)\n", (long) m->rm_call.cb_proc);
    goto freemsg;
  }

  sbp->sb_svcprog = pp;
  sbp->sb_msg = m;
  rtp = &sbp->sb_tbl[sbp->sb_procno];
  if (salen > 0) {
    if (salen > sizeof (sbp->sb_sin))
      fatal ("asvc_dispatch: socket name too large\n");
    memcpy (&sbp->sb_sin, sa, salen);
  }
  sbp->sb_rdat = NULL;
  if (asvc_replaycheck (sbp))
    goto freemsg;

  sbp->sb_arg = xmalloc (rtp->len_arg);

  bzero (sbp->sb_arg, rtp->len_arg);
  if (! rtp->xdr_arg (&x, sbp->sb_arg)) {
    warn ("asvc_dispatch: bad message prog %d vers %d proc %d\n",
	  (u_int) m->rm_call.cb_prog, (u_int) m->rm_call.cb_vers,
	  (u_int) m->rm_call.cb_proc);
    goto freearg;
  }

  xdr_destroy (&x);
  if (m->rm_call.cb_cred.oa_flavor == AUTH_UNIX) {
    xdrmem_create (&x, m->rm_call.cb_cred.oa_base,
		   m->rm_call.cb_cred.oa_length, XDR_DECODE);
    sbp->sb_aup = xmalloc (sizeof (*sbp->sb_aup));
    bzero (sbp->sb_aup, sizeof (*sbp->sb_aup));
    if (! xdr_authunix_parms (&x, sbp->sb_aup))
      goto freeauth;
    xdr_destroy (&x);
  }
  else
    sbp->sb_aup = NULL;

  ap->as_count++;
  ap->as_disp_fn (ap->as_disp_arg, sbp);
  return;

freeauth:
  xfree (sbp->sb_aup);
freearg:
  xfree (sbp->sb_arg);
freemsg:
  xdr_free ((xdrproc_t) xdr_callmsg, (void *) &m);
freexdr:
  xdr_destroy (&x);
  xfree (sbp);
  xfree (m);
error:
  if (xh_isconnected (ap->as_xh))
    ap->as_disp_fn (ap->as_disp_arg, NULL);
}

void
asvc_noreply (struct svccb *sbp)
{
  struct asvc *ap = sbp->sb_svc;
  struct rcache *rc = ap->as_rcache;

  if (! ap->as_xh) {
    asvc_freesvccb (sbp);
    if (ap->as_count == 0) {
      if (ap->as_rcache)
	xfree (ap->as_rcache);
      xfree (ap);
    }
    return;
  }

  if (rc && sbp->sb_rdat) {
    if (rc->rc_nold == RC_MAXOLD)
      asvc_freesvccb (rc->rc_lru.tqh_first);
    else
      rc->rc_nold++;
    TAILQ_INSERT_TAIL (&rc->rc_lru, sbp, sb_tlink);
  }
  else
    asvc_freesvccb (sbp);
}

void
asvc_reply_raw (struct svccb *sbp, struct iovec *iov, int iovcnt)
{
   if (sbp->sb_svc->as_xh) {
      xh_sendv (sbp->sb_svc->as_xh, iov, iovcnt, &sbp->sb_sin);
   }
   asvc_noreply (sbp);
}
void
asvc_reply_suio (struct svccb *sbp, struct suio *uio)
{
  XDR x;
  u_int32_t *dp;

  assert (sbp->sb_msg);
  if (sbp->sb_svc->as_xh) {
    xdrsuio_create (&x, XDR_ENCODE);
    dp = xdrsuio_inline (&x, 6 * 4);
    *dp++ = htonl (sbp->sb_xid);
    *dp++ = htonl (REPLY);
    *dp++ = htonl (MSG_ACCEPTED);
    *dp++ = htonl (AUTH_NONE);
    *dp++ = htonl (0);
    *dp++ = htonl (SUCCESS);
    suio_cat (xsuio (&x), uio);

    sbp->sb_rlen = xsuio(&x)->uio_resid;
    sbp->sb_rdat = suio_flatten (xsuio (&x));
    xdr_destroy (&x);

    xh_send (sbp->sb_svc->as_xh, sbp->sb_rdat, sbp->sb_rlen, &sbp->sb_sin);
  }

  asvc_noreply (sbp);
}

inline void
asvc_reply_xdrproc (struct svccb *sbp, xdrproc_t resproc, void *resp)
{
  struct rpc_msg rm;
  XDR x;

  assert (sbp->sb_msg);
  if (sbp->sb_svc->as_xh) {
    rm.rm_xid = sbp->sb_xid;
    rm.rm_direction = REPLY;
    rm.rm_reply.rp_stat = MSG_ACCEPTED;
    rm.acpted_rply.ar_verf = _null_auth;
    rm.acpted_rply.ar_stat = SUCCESS;
    rm.acpted_rply.ar_results.where = resp;
    rm.acpted_rply.ar_results.proc = resproc;

    xdrsuio_create (&x, XDR_ENCODE);
    if (! xdr_replymsg (&x, &rm))
      fatal ("asvc_reply: xdr_replymsg failed\n");
    sbp->sb_rlen = xsuio(&x)->uio_resid;
    sbp->sb_rdat = suio_flatten (xsuio (&x));
    xdr_destroy (&x);

    xh_send (sbp->sb_svc->as_xh, sbp->sb_rdat, sbp->sb_rlen, &sbp->sb_sin);
  }

  asvc_noreply (sbp);
}

void
asvc_reply (struct svccb *sbp, void *resp)
{
  struct rpcgen_table *rtp = &sbp->sb_tbl[sbp->sb_procno];

  asvc_reply_xdrproc (sbp, rtp->xdr_res, resp);
}

/* Clone a server by passing its associated file descriptor down the
 * Unix domain socket associated with child.  Note:  ap must be
 * associated with a ps transport, and child must be a ux
 * transport.  Otherwise, core dump. */
void
asvc_clone (struct asvc *ap, struct svccb *sbp, struct axh *child)
{
#ifdef HAVE_FDPASS
  XDR x;
  struct rpcgen_table *rtp = &sbp->sb_tbl[sbp->sb_procno];

  xdrsuio_create (&x, XDR_ENCODE);
  if (!xdr_callmsg (&x,  sbp->sb_msg) || !rtp->xdr_arg (&x, sbp->sb_arg))
    warn ("asvc_clone: bad RPC message\n");
  else {
    us_sendfd (child, ps_getfd (ap->as_xh));
    xh_sendv (child, xsuio(&x)->uio_iov, xsuio(&x)->uio_iovcnt, NULL);
  }

  xdr_destroy (&x);
  asvc_noreply (sbp);
  ps_ncfree (ap->as_xh);
  ap->as_xh = NULL;
  asvc_free (ap);
#else
  fatal ("Don't support cloning with fd passing\n");
#endif
}
