/* $Id: axprt_ps.c,v 1.1 1998/12/07 20:42:41 fubobb Exp $ */

#include "sysconf.h"

#include "arpc.h"
#include "suio.h"
#if 0
#include "arc4.h"
#include "sha1.h"
#endif

#define MAX_PACKET 0x10100
#define BLOCK_SIZE 8

/* Packet stream structure */
struct ps {
  struct axh ps_axh;                        /* Common transport handle */
  int ps_fd;                                /* The socket */
  char ps_cryptsend;                        /* Non-zero when encrypting */
  char ps_cryptrecv;                        /* Non-zero when decrypting */
  struct sockaddr_in ps_peer;               /* Name of the peer */

  void (*ps_recv_fn) (void *, char *, int,
                      void *, int);         /* Dispatch for received packets */
  void *ps_recv_arg;                        /* With argument */
  enum {RMAC, RLEN, RDAT} ps_rstat;         /* Read status (length/data) */
  int ps_nbytes;                            /* # Valid bytes in buffer */
  int ps_buf_size;                          /* Size of dynamic buffer */
  char ps_rmacdat[2*BLOCK_SIZE];            /* Small buffer for reading LEN */
  u_int32_t ps_rlen;                        /* Place to read length */
  char *ps_buf_packet;                      /* Dynamic buffer for packet */
#if 0
  struct arc4_stream ps_ras;                /* Receiving encryption state */
#endif
  struct suio *ps_out;                      /* IOV of output packets */
#if 0
  struct arc4_stream ps_sas;                /* Sending encryption state */
#endif
  void *ps_savedpkt;                        /* Saved packet to be read */
  int ps_savedlen;                          /* Length of saved packet */
};

static void ps_input (void *);
static void ps_output (void *);

static void ps_sendv (struct axh *, struct iovec *, int, void *);
static void ps_free (struct axh *);
static void ps_encrypt (struct axh *, int, char *, int);
static void ps_setcb (struct axh *,
		      void (*) (void *, char *, int, void *, int), void *);

const struct axprt axprt_ps = {
  1, ps_sendv, ps_free, ps_setcb, ps_encrypt
};

struct axh *
ps_alloc (int fd)
{
  struct ps *psp;
  int socklen;

  psp = xmalloc (sizeof (*psp));

  psp->ps_axh.x = &axprt_ps;
  psp->ps_axh.connected = 1;
  psp->ps_fd = fd;
  psp->ps_cryptsend = psp->ps_cryptrecv = 0;
  psp->ps_recv_fn = NULL;
  psp->ps_recv_arg = NULL;

  psp->ps_rstat = RMAC;
  psp->ps_nbytes = 0;
  psp->ps_buf_packet = NULL;

  psp->ps_out = suio_alloc ();

  socklen = sizeof (psp->ps_peer);
  bzero (&psp->ps_peer, socklen);
  getpeername (psp->ps_fd, (struct sockaddr *) &psp->ps_peer, &socklen);

  psp->ps_savedpkt = NULL;

#if 0
  cb_add (fd, 0, ps_input, psp);
  cb_enable (fd, 0, 0);
#endif

  return &psp->ps_axh;
}

/* "Unreceive" one packet (e.g. the first packet when receiving
 * granted file descriptors). */
void
ps_unrecv (struct axh *_psp, void *pkt, int len)
{
  struct ps *psp = (struct ps *) _psp;
  assert (psp->ps_axh.x = &axprt_ps);
  assert (!psp->ps_savedpkt);
  psp->ps_savedpkt = pkt;
  psp->ps_savedlen = len;
}

static void
ps_encrypt (struct axh *_psp, int invert, char *key, int keylen)
{
  fatal ("ps_encrypt: encryption not supported by transport layer.\n");

#if 0
  struct ps *psp = (struct ps *) _psp;
  u_int32_t zero;

  keylen >>= 1;
  if (invert) {
    arc4_init (&psp->ps_ras, key + keylen, keylen);
    arc4_init (&psp->ps_sas, key, keylen);
  }
  else {
    arc4_init (&psp->ps_ras, key, keylen);
    arc4_init (&psp->ps_sas, key + keylen, keylen);
  }
  psp->ps_cryptsend = 1;
  zero = htonl (0);
  suio_copy (psp->ps_out, (char *) &zero, sizeof (zero));
  ps_output (psp);
#endif
}

/* Free without closing */
inline void
ps_ncfree (struct axh *_psp)
{
  struct ps *psp = (struct ps *) _psp;

  assert (psp->ps_axh.x = &axprt_ps);

  cb_free (psp->ps_fd, 0);
  cb_free (psp->ps_fd, 1);
  xfree (psp->ps_buf_packet);
  suio_free (psp->ps_out);
  xfree (psp->ps_savedpkt);
  xfree (psp);
}

static void
ps_free (struct axh *_psp)
{
  struct ps *psp = (struct ps *) _psp;
  int fd = psp->ps_fd;

  ps_ncfree (&psp->ps_axh);
  close (fd);
}

static void
ps_sendv (struct axh *_psp, struct iovec *iov, int cnt, void *_null)
{
  struct ps *psp = (struct ps *) _psp;
  int len = iovsize (iov, cnt);
  int i;
  char *pkt;
  int psize;
  char *d;

  if (len <= 0 || len > MAX_PACKET)
    fatal ("ps_sendpacket: invalid len 0x%x\n", len);

#if 0
  if (!psp->ps_cryptsend) {
#endif
    psize = len + 4;
    pkt = xmalloc (psize);
    ((u_int32_t *) pkt)[0] = htonl (0x80000000 | len);
    d = pkt + 4;
    for (i = 0; i < cnt; i++) {
      memcpy (d, iov[i].iov_base, iov[i].iov_len);
      d += iov[i].iov_len;
    }

#if 0
  }
  else {
    char macdat[2 * BLOCK_SIZE];
    sha1_ctx c;

    psize = len + 4 + SHA1_SIZE;
    pkt = xmalloc (psize);

    ((u_int32_t *) pkt)[0] = htonl (0x80000000 | len);
    d = pkt + 4;
    for (i = 0; i < cnt; i++) {
      memcpy (d, iov[i].iov_base, iov[i].iov_len);
      d += iov[i].iov_len;
    }

    arc4_fill (&psp->ps_sas, macdat, sizeof (macdat));
    sha1_init (&c);
    sha1_update (&c, macdat, BLOCK_SIZE);
    sha1_update (&c, pkt, len + 4);
    sha1_update (&c, macdat + BLOCK_SIZE, BLOCK_SIZE);
    sha1_final (pkt + len + 4, &c);
    arc4_transform (&psp->ps_sas, pkt, psize);
  }
#endif

  suio_print (psp->ps_out, pkt, psize);
  suio_callfree (psp->ps_out, pkt);
  ps_output (psp);
}

static void
ps_input (void *_psp)
{
  struct ps *psp = _psp;
  int n;

  /* If ps_setcb was called with fn=NULL from within cb_check,
   * ps_input might still get invoked once if ps_fd has already
   * selected true.  We will just ignore the pending data and reap it
   * once a valid callback is set. */
  if (!psp->ps_recv_fn)
    return;

  /* First, pull secret data out of arc4 stream for MAC */
  if (psp->ps_rstat == RMAC) {
#if 0
    if (psp->ps_cryptrecv)
      arc4_fill (&psp->ps_ras, psp->ps_rmacdat, sizeof (psp->ps_rmacdat));
#endif
    psp->ps_rstat = RLEN;
    psp->ps_nbytes = 0;
  }

  /* Get the length of this packet if we haven't already */
  if (psp->ps_rstat == RLEN) {
    /* Get a full block of text */
    n = read (psp->ps_fd, (char *) &psp->ps_rlen + psp->ps_nbytes, 
	      sizeof (psp->ps_rlen) - psp->ps_nbytes);
    if (n <= 0) {
      if (n == 0 || errno != EAGAIN)
         psp->ps_recv_fn (psp->ps_recv_arg, NULL, n, NULL, 0);
      return;
    }
    psp->ps_nbytes += n;
    if (psp->ps_nbytes < sizeof (psp->ps_rlen))
      return;

#if 0
    if (psp->ps_cryptsend && !psp->ps_cryptrecv
	&& psp->ps_rlen == htonl (0)) {
      psp->ps_rstat = RMAC;
      psp->ps_cryptrecv = 1;
      return;
    }

    /* We have a full block, decode it */
    if (psp->ps_cryptrecv) {
      arc4_transform (&psp->ps_ras, (char *) &psp->ps_rlen,
		      sizeof (psp->ps_rlen));
      psp->ps_buf_size = ntohl (psp->ps_rlen) + SHA1_SIZE;
    }
    else
#endif
      psp->ps_buf_size = ntohl (psp->ps_rlen);

    /* Check packet length */
    if (! (psp->ps_buf_size & 0x80000000)
	|| !(psp->ps_buf_size &= 0x7fffffff, psp->ps_buf_size)
	|| psp->ps_buf_size > MAX_PACKET) {
      warn ("Invalid packet size in ps_input!\n");
      psp->ps_recv_fn (psp->ps_recv_arg, NULL, -1, NULL, 0);
      return;
    }
    psp->ps_buf_packet = xmalloc (psp->ps_buf_size);
    psp->ps_nbytes = 0;
    psp->ps_rstat = RDAT;
  }

  /* Read the rest of the current packet */
  n = read (psp->ps_fd, psp->ps_buf_packet + psp->ps_nbytes,
	    psp->ps_buf_size - psp->ps_nbytes);
  if (n <= 0) {
    if (n == 0 || errno != EAGAIN)
      psp->ps_recv_fn (psp->ps_recv_arg, NULL, n, NULL, 0);
    return;
  }
#if 0
  if (psp->ps_cryptrecv)
    arc4_transform (&psp->ps_ras, psp->ps_buf_packet + psp->ps_nbytes, n);
#endif
  psp->ps_nbytes += n;

  if (psp->ps_nbytes > psp->ps_buf_size)
    fatal ("Packet overflow!!!\n");
  if (psp->ps_nbytes == psp->ps_buf_size) {
    char *buf;
#if 0
    if (psp->ps_cryptrecv) {
      sha1_ctx c;
      char mac[SHA1_SIZE];

      sha1_init (&c);
      sha1_update (&c, psp->ps_rmacdat, BLOCK_SIZE);
      sha1_update (&c, (char *) &psp->ps_rlen, sizeof (psp->ps_rlen));
      sha1_update (&c, psp->ps_buf_packet, psp->ps_buf_size - SHA1_SIZE);
      sha1_update (&c, psp->ps_rmacdat + BLOCK_SIZE, BLOCK_SIZE);
      sha1_final (mac, &c);
      if (memcmp (mac, psp->ps_buf_packet + psp->ps_buf_size - SHA1_SIZE,
		  SHA1_SIZE)) {
	warn ("MAC failure!\n");
	psp->ps_recv_fn (psp->ps_recv_arg, NULL, -1, NULL, 0);
	return;
      }
    }
#endif

    /* Note psp may be freed from the callback, so we have to save
     * ps_buf_packet somewhere, first. */
    buf = psp->ps_buf_packet;
    psp->ps_buf_packet = NULL;
    psp->ps_rstat = RMAC;
    psp->ps_recv_fn (psp->ps_recv_arg, buf,
		     psp->ps_buf_size,
#if 0
		     psp->ps_buf_size - (psp->ps_cryptrecv ? SHA1_SIZE : 0),
#endif
		     &psp->ps_peer, sizeof (psp->ps_peer));
    xfree (buf);
  }
}

static void
ps_output (void *_psp)
{
  struct ps *psp = _psp;
  int n;

  n = writev (psp->ps_fd, psp->ps_out->uio_iov,
	      min (UIO_MAXIOV, psp->ps_out->uio_iovcnt));
  if (n == 0 || (n < 0 && errno != EAGAIN)) {
    cb_free (psp->ps_fd, 1);
    if (n < 0)
      warn ("packet stream write error (%s)\n", strerror (errno));
    psp->ps_recv_fn (psp->ps_recv_arg, NULL, -1, NULL, 0);
    return;
  }
  if (n > 0)
    suio_rembytes (psp->ps_out, n);
  if (psp->ps_out->uio_resid == 0)
    cb_free (psp->ps_fd, 1);
  else
    cb_add (psp->ps_fd, 1, ps_output, psp);
}

static void
ps_setcb (struct axh *_psp,
	  void (*fn) (void *, char *, int, void *, int), void *arg)
{
  struct ps *psp = (struct ps *) _psp;
  psp->ps_recv_fn = fn;
  psp->ps_recv_arg = arg;
  if (fn) {
    if (psp->ps_savedpkt) {
      fn (arg, psp->ps_savedpkt, psp->ps_savedlen,
	  &psp->ps_peer, sizeof (psp->ps_peer));
      xfree (psp->ps_savedpkt);
      psp->ps_savedpkt = NULL;
    }
    cb_add (psp->ps_fd, 0, ps_input, psp);
  }
  else
    cb_free (psp->ps_fd, 0);
}

int
ps_getfd (struct axh *_psp)
{
  struct ps *psp = (struct ps *) _psp;
  assert (psp->ps_axh.x = &axprt_ps);
  return (psp->ps_fd);
}
