/* $Id: axprt_us.c,v 1.1 1998/12/07 20:42:43 fubobb Exp $ */

#include "sysconf.h"

#include "arpc.h"
#include "suio.h"
#if 0
#include "arc4.h"
#include "sha1.h"
#endif

#define MAX_PACKET 0x10100

struct outfd {
  u_int of_niov;
  int of_fd;
  TAILQ_ENTRY (outfd) of_link;
};

/* Unix-domain stream structure */
struct us {
  struct axh us_axh;                        /* Common transport handle */
  int us_fd;                                /* The socket */

  void (*us_recv_fn) (void *, char *, int,
                      void *, int);         /* Dispatch for received packets */
  void *us_recv_arg;                        /* With argument */

  enum {RLEN, RDAT} us_rstat;               /* Read status (length/data) */
  int us_nbytes;                            /* # Valid bytes in buffer */
  int us_buf_size;                          /* Size of dynamic buffer */
  u_int32_t us_rlen;                        /* Place to read length */
  char *us_buf_packet;                      /* Dynamic buffer for packet */
  int us_infd;                              /* Received file descriptor */

  struct suio *us_out;                      /* IOV of output packets */
  TAILQ_HEAD (, outfd) us_outfds;           /* Granted file descriptors */

  void *us_savedpkt;                        /* Saved packet to be read */
  int us_savedlen;                          /* Length of saved packet */
};

static void us_input (void *);
static void us_output (void *);

static void us_sendv (struct axh *, struct iovec *, int, void *);
static void us_free (struct axh *);
static void us_encrypt (struct axh *, int, char *, int);
static void us_setcb (struct axh *,
		      void (*) (void *, char *, int, void *, int), void *);

const struct axprt axprt_us = {
  1, us_sendv, us_free, us_setcb, us_encrypt
};

struct axh *
us_alloc (int fd)
{
  struct us *usp;

  usp = xmalloc (sizeof (*usp));

  usp->us_axh.x = &axprt_us;
  usp->us_axh.connected = 1;
  usp->us_fd = fd;
  usp->us_recv_fn = NULL;
  usp->us_recv_arg = NULL;

  usp->us_rstat = RLEN;
  usp->us_nbytes = 0;
  usp->us_buf_packet = NULL;
  usp->us_infd = -1;

  usp->us_out = suio_alloc ();
  TAILQ_INIT (&usp->us_outfds);

  usp->us_savedpkt = NULL;

  return &usp->us_axh;
}

/* "Unreceive" one packet (e.g. the first packet when receiving
 * granted file descriptors). */
void
us_unrecv (struct axh *_usp, void *pkt, int len)
{
  struct us *usp = (struct us *) _usp;
  assert (usp->us_axh.x = &axprt_us);
  assert (!usp->us_savedpkt);
  usp->us_savedpkt = pkt;
  usp->us_savedlen = len;
}

static void
us_encrypt (struct axh *_usp, int invert, char *key, int keylen)
{
  fatal ("ux_encrypt: encryption not supported by transport layer.\n");
}

static void
us_free (struct axh *_usp)
{
  struct us *usp = (struct us *) _usp;
  struct outfd *ofp;

  cb_free (usp->us_fd, 0);
  cb_free (usp->us_fd, 1);
  close (usp->us_fd);
  xfree (usp->us_buf_packet);
  suio_free (usp->us_out);
  while ((ofp = usp->us_outfds.tqh_first)) {
    close (ofp->of_fd);
    TAILQ_REMOVE (&usp->us_outfds, ofp, of_link);
    xfree (ofp);
  }
  if (usp->us_infd != -1)
    close (usp->us_infd);
  xfree (usp->us_savedpkt);
  xfree (usp);
}

static void
us_input (void *_usp)
{
  struct us *usp = _usp;
  int n;

  /* If us_setcb was called with fn=NULL from within cb_check,
   * us_input might still get invoked once if us_fd has already
   * selected true.  We will just ignore the pending data and reap it
   * once a valid callback is set. */
  if (!usp->us_recv_fn)
    return;

  /* Get the length of this packet if we haven't already */
  if (usp->us_rstat == RLEN) {
    /* Get a full block of text */
    n = read (usp->us_fd, (char *) &usp->us_rlen + usp->us_nbytes, 
	      sizeof (usp->us_rlen) - usp->us_nbytes);
    if (n <= 0) {
      if (n == 0 || errno != EAGAIN)
         usp->us_recv_fn (usp->us_recv_arg, NULL, n, NULL, 0);
      return;
    }
    usp->us_nbytes += n;
    if (usp->us_nbytes < sizeof (usp->us_rlen))
      return;

    /* We have a full block, decode it */
    usp->us_buf_size = ntohl (usp->us_rlen);

#ifdef HAVE_FDPASS
    /* 0 means we have been sent a file descriptor */
    if (!usp->us_buf_size) {
      if (usp->us_infd != -1) {
	warn ("us_input: received two file descriptors in a row!\n");
	usp->us_recv_fn (usp->us_recv_arg, NULL, -1, NULL, 0);
	return;
      }
      usp->us_infd = recvfd (usp->us_fd);
      usp->us_rstat = RLEN;
      usp->us_nbytes = 0;
      return;
    }
#endif /* HAVE_FDPASS */

    /* Check packet length */
    if (!(usp->us_buf_size & 0x80000000)
	|| !(usp->us_buf_size &= 0x7fffffff, usp->us_buf_size)
	|| usp->us_buf_size > MAX_PACKET) {
      warn ("Invalid packet size in us_input!\n");
      usp->us_recv_fn (usp->us_recv_arg, NULL, -1, NULL, 0);
      return;
    }
    usp->us_buf_packet = xmalloc (usp->us_buf_size);
    usp->us_nbytes = 0;
    usp->us_rstat = RDAT;
  }

  /* Read the rest of the current packet */
  n = read (usp->us_fd, usp->us_buf_packet + usp->us_nbytes,
	    usp->us_buf_size - usp->us_nbytes);
  if (n <= 0) {
    if (n == 0 || errno != EAGAIN)
      usp->us_recv_fn (usp->us_recv_arg, NULL, n, NULL, 0);
    return;
  }
  usp->us_nbytes += n;

  assert (usp->us_nbytes <= usp->us_buf_size);
  if (usp->us_nbytes == usp->us_buf_size) {
    /* Note usp may be freed from the callback, so we have to save
     * us_buf_packet somewhere, first. */
    char *buf;

    buf = usp->us_buf_packet;
    usp->us_buf_packet = NULL;
    usp->us_rstat = RLEN;
    usp->us_nbytes = 0;
    usp->us_recv_fn (usp->us_recv_arg, buf,
		     usp->us_buf_size, NULL, 0);
    xfree (buf);
  }
}

/* Queue up the packet, but try writing some immediately if possible,
 * so as to avoid some copying overhead. */
static void
us_sendv (struct axh *_usp, struct iovec *iov, int cnt, void *_null)
{
  struct us *usp = (struct us *) _usp;
  int len;
  u_int32_t xlen;
  int b, n, m;
  int psize;
  char *pkt, *d;

  len = iovsize (iov, cnt);
  if (len <= 0 || len > MAX_PACKET)
    fatal ("ps_sendpacket: invalid len 0x%x\n", len);
  xlen = htonl (0x80000000 | len);
  n = 0;
  m = 0;

  if (!usp->us_out->uio_resid && !usp->us_outfds.tqh_first) {
    n = write (usp->us_fd, &xlen, sizeof (xlen));
    if (n == 0 || (n < 0 && errno != EAGAIN)) {
      cb_free (usp->us_fd, 1);
      if (n < 0)
	warn ("us_sendv: write error (%s)\n", strerror (errno));
      usp->us_recv_fn (usp->us_recv_arg, NULL, -1, NULL, 0);
      return;
    }
    if (n < 0)
      n = 0;
  }

  if (n == sizeof (xlen)) {
    m = writev (usp->us_fd, iov, min (cnt, UIO_MAXIOV));
    if (m == 0 || (m < 0 && errno != EAGAIN)) {
      cb_free (usp->us_fd, 1);
      if (m < 0)
	warn ("us_sendv: writev error (%s)\n", strerror (errno));
      usp->us_recv_fn (usp->us_recv_arg, NULL, -1, NULL, 0);
      return;
    }
    if (m < 0)
      m = 0;
  }

  psize = 4 - n + len - m;
  if (!psize)
    return;

  d = pkt = xmalloc (psize);
  memcpy (d, (char *) &xlen + n, 4 - n);
  d += 4 - n;

  b = 0;
  while (cnt > 0 && b < m) {
    if (b + iov->iov_len > m) {
      memcpy (d, iov->iov_base + m - b, b + iov->iov_len - m);
      d += iov->iov_len - m;
    }
    b += iov->iov_len;
    iov++;
    cnt--;
  }
  while (cnt > 0) {
    memcpy (d, iov->iov_base, iov->iov_len);
    d += iov->iov_len;
    iov++;
    cnt--;
  }
  assert (d = pkt + psize);
  suio_print (usp->us_out, pkt, psize);
  suio_callfree (usp->us_out, pkt);
  us_output (usp);
}

static void
us_output (void *_usp)
{
  struct outfd *ofp;
  struct us *usp = _usp;
  int cnt, n;

  while ((ofp = usp->us_outfds.tqh_first) || usp->us_out->uio_resid) {
#ifdef HAVE_FDPASS
    if (ofp && usp->us_out->uio_nremiov >= ofp->of_niov) {
      assert (usp->us_out->uio_nremiov == ofp->of_niov);
      if (!readytowrite (usp->us_fd))
	break;
      if (sendfd (usp->us_fd, ofp->of_fd) < 0) {
	cb_free (usp->us_fd, 1);
	warn ("us_output: sendfd: %s\n", strerror (errno));
	usp->us_recv_fn (usp->us_recv_arg, NULL, -1, NULL, 0);
	return;
      }
      close (ofp->of_fd);
      TAILQ_REMOVE (&usp->us_outfds, ofp, of_link);
      xfree (ofp);
      continue;
    }

    if (ofp) {
      cnt = ofp->of_niov - usp->us_out->uio_nremiov;
      assert (cnt > 0);
      assert (cnt <= usp->us_out->uio_iovcnt);
    }
    else
#endif /* HAVE_FDPASS */
      cnt = usp->us_out->uio_iovcnt;

    n = writev (usp->us_fd, usp->us_out->uio_iov, min (UIO_MAXIOV, cnt));

    if (n == 0 || (n < 0 && errno != EAGAIN)) {
      cb_free (usp->us_fd, 1);
      if (n < 0)
	warn ("us_output: write error (%s)\n", strerror (errno));
      usp->us_recv_fn (usp->us_recv_arg, NULL, -1, NULL, 0);
      return;
    }

    if (n < 0)
      break;
    suio_rembytes (usp->us_out, n);
  }

  if (usp->us_out->uio_resid || usp->us_outfds.tqh_first)
    cb_add (usp->us_fd, 1, us_output, usp);
  else
    cb_free (usp->us_fd, 1);
}

static void
us_setcb (struct axh *_usp,
	  void (*fn) (void *, char *, int, void *, int), void *arg)
{
  struct us *usp = (struct us *) _usp;
  usp->us_recv_fn = fn;
  usp->us_recv_arg = arg;
  if (fn) {
    if (usp->us_savedpkt) {
      fn (arg, usp->us_savedpkt, usp->us_savedlen, NULL, 0);
      xfree (usp->us_savedpkt);
      usp->us_savedpkt = NULL;
    }
    cb_add (usp->us_fd, 0, us_input, usp);
  }
  else
    cb_free (usp->us_fd, 0);
}

#ifdef HAVE_FDPASS

int
us_recvfd (struct axh *_usp)
{
  struct us *usp = (struct us *) _usp;
  int fd;

  assert (usp->us_axh.x = &axprt_us);

  fd = usp->us_infd;
  usp->us_infd = -1;
  return fd;
}

void
us_sendfd (struct axh *_usp, int fd)
{
  struct us *usp = (struct us *) _usp;
  struct outfd *ofp;
  u_int32_t *zerop;

  assert (usp->us_axh.x = &axprt_us);

  ofp = xmalloc (sizeof (*ofp));
  zerop = xmalloc (sizeof (*zerop));
  *zerop = htonl (0);
  /* The outfd structures specify the iovec after which a file
   * descriptor must be sent, not the byte.  Thus, we ensure the zero
   * word occupies its own iovec structure rather than simply calling
   * suio_print which might copy the data anyway. */
  __suio_addiov (usp->us_out, (char *) zerop, sizeof (*zerop));
  suio_callfree (usp->us_out, zerop);
  ofp->of_niov = usp->us_out->uio_nremiov + usp->us_out->uio_iovcnt;
  ofp->of_fd = fd;
  TAILQ_INSERT_TAIL (&usp->us_outfds, ofp, of_link);
}

#endif /* !HAVE_FDPASS */
