/* $Id: dns.c,v 1.1 1998/12/07 20:42:46 fubobb Exp $ */
/* Asynchronous resolver.  This code is easy to understand if and only
 * if you also have a copy of RFC 1035.  You can get this by anonymous
 * ftp from, among other places, ftp.internic.net:/rfc/rfc1035.txt.
 */

#include "sysconf.h"
#include "resolv.h"

#include "queue.h"
#include "dns.h"
#include "misc.h"

extern int __dn_skipname (const u_char *, const u_char *);

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif /* !INADDR_NONE */

static void ar_reap (void *);
static void ar_timeout (void *);

int ar_sock;                    /* UDP socket connected to nameserver */
int ar_errno;                   /* Error code */

static u_short ar_lastid;       /* Counter for DNS query id's */

#define QBSIZE 8192
#define NSPORT NAMESERVER_PORT

typedef union {
  HEADER hdr;
  u_char buf[QBSIZE];
} querybuf;

/* Size of the hash table for outstanding requests. */
#define LOG2ARHSZ 3
#define ARHSZ (1<<LOG2ARHSZ)
#define arhash(id) ((id) & (ARHSZ - 1))

#define MINTO 2                 /* Time to wait before first retransmit */
#define MAXSEND 5               /* Maximum number of times to ask nameserver */

struct ar_r {
  u_short arr_id;               /* DNS query ID */
  u_short arr_type;             /* Type of query (T_A or T_PTR) */
  u_char arr_toqn;              /* Number of this request's timeout queue */
  time_t arr_resend;            /* Time at which to retransmit or give up */
  char *arr_name;               /* Name for which to query */
  struct in_addr arr_addr;      /* Adress of inverse queries (for checking) */
  void (*arr_cb) ();            /* Function to call with result */
  void *arr_arg;                /* First argument to callback function */
  TAILQ_ENTRY(ar_r) arr_link;   /* Link for query ID hash table */
  TAILQ_ENTRY(ar_r) arr_to;     /* Link on retransmit queue */
};
TAILQ_HEAD (arr_tq, ar_r);

struct arr_tq arh[ARHSZ];       /* Hash table indexed by arhash(id) */
struct arr_tq artoq[MAXSEND];   /* Time out lists for retransmits */
char timeout_pending[MAXSEND];  /* Non-zero if a timeout is registered for
				   the corresponding timeout queue */

static inline struct ar_r *
arh_lookup (u_short id)
{
  struct ar_r *r;

  for (r = arh[arhash (id)].tqh_first;
       r && r->arr_id != id;
       r = r->arr_link.tqe_next)
    ;
  return (r);
}

static inline void
arr_free (struct ar_r *r)
{
  xfree (r->arr_name);
  TAILQ_REMOVE (&arh[arhash(r->arr_id)], r, arr_link);
  TAILQ_REMOVE (&artoq[r->arr_toqn], r, arr_to);
  xfree (r);
}

/* Schedule a timeout if we need one */
static inline void
sched_timeout (void)
{
  int n;

  for (n = 0; n < MAXSEND; n++) {
    struct ar_r *r;
    if (timeout_pending[n])
      break;
    r = artoq[n].tqh_first;
    if (r) {
      timeout_pending[n] = 1;
      tmo_add (r->arr_resend, ar_timeout, &timeout_pending[n]);
    }
  }
}

/* Get a pointer to the beginning of the answer section (skip the
 * header and queries). */
static const u_char *
getanp (const u_char *qbraw, const u_char *eom)
{
  const querybuf *qb = (querybuf *) qbraw;
  int i, l, n;
  const u_char *cp = &qb->buf[sizeof (HEADER)];

  if (cp > eom || qb->hdr.tc || !qb->hdr.qr || qb->hdr.rcode) {
    if (qb->hdr.rcode)
      ar_errno = qb->hdr.rcode;
    else
      ar_errno = ARERR_BADRESP;
    return (NULL);
  }
  l = ntohs (qb->hdr.qdcount);
  for (i = 0; i < l; i++) {
    n = __dn_skipname (cp, eom);
    cp += n + 4;
    if (n < 0 || cp > eom)
      return (NULL);
  }
  return (cp);
}

/* This is really part of ar_dns2h, but was broken out for readability.
 * 
 *     qb       - Query buffer (response from name server)
 *     eom      - End of message (qb + size of resonse)
 *     rr_data  - Resource records (this should point to the answer section
 *                   of the response)
 *     rr_count - Number of resource records (ancount from response header)
 *     hostname - Either the hostname one is interested in, or an empty string
 *                   of length MAXDNAME which will get the hostname on return
 *     *naddr   - On return, the number of A records for hostname
 *     *nalias  - On return, the number of CNAMES with value hostname
 *     *nchars  - On return, the number of bytes of space occupied by all the
 *                names of aliases (including the null bytes terminating
 *                strings)
 *
 * Return value is 0 on success, or -1 for an invalid or truncated qb.
 */
static int
count_rr_space (const char * const qb, const char * const eom,
		const char *rr_data, const int rr_count,
		u_char * const hostname,
		int * const naddr, int * const nalias, int * const nchars)
{
  int i;

  *naddr = *nalias = *nchars = 0;

  for (i = 0; i < rr_count; i++) {
    int n, l;
    u_short type, class, rdlen;
    u_char name[MAXDNAME];

    if (rr_data > eom)
      return (-1);
    n = dn_expand (qb, eom, rr_data, name, sizeof (name));
    rr_data += n;
    if (n < 0 || rr_data + 10 > eom) {
      fprintf (stderr, "count_rr_space: response overflow!\n");
      return (-1);
    }
    GETSHORT (type, rr_data);
    GETSHORT (class, rr_data);
    rr_data += 4;
    GETSHORT (rdlen, rr_data);
    if (rr_data + rdlen > eom) {
      fprintf (stderr, "count_rr_space: response overflow!\n");
      return (-1);
    }

    if (class == C_IN) {
      switch (type) {
      case T_A:
	if (! *hostname)
	  strcpy (hostname, name);
	if (! strcmp (hostname, name))
	  (*naddr)++;
	break;
      case T_CNAME:
	l = strlen (name);
	n = dn_expand (qb, eom, rr_data, name, sizeof (name));
	if (n < 0) {
	  fprintf (stderr, "count_rr_space: response overflow!\n");
	  return (-1);
	}
	if (! *hostname)
	  strcpy (hostname, name);
	if (! strcmp (hostname, name)) {
	  (*nalias)++;
	  (*nchars) += l + 1;
	}
	break;
      }
    }

    rr_data += rdlen;
  }

  return (*hostname ? 0 : -1);
}

/* Takes the response from an A query, and turns it into a dynamically
 * allocated hostent structure (defined in netdb.h).  All data for the
 * hostent structure is allocated contiguously, so that it can all be
 * freed by simply freeing the address of the hostent structure returned.
 */
struct hostent *
ar_dns2h (const u_char *qbraw, int len)
{
  const querybuf * const qb = (querybuf *) qbraw;
  const char * const eom = qb->buf + len;
  const int ancount = ntohs (qb->hdr.ancount);
  const u_char *cp = getanp (qb->buf, eom);
  char hostname[MAXDNAME] = "";
  int l, i;
  int n_addr, n_alias, n_chars;
  int c_addr = 0, c_alias = 0, c_chars = 0;
  struct in_addr **addrp_buf, *addr_buf;
  char **namep_buf, *name_buf;
  struct hostent *h;
  int h_size;

  if (! cp || count_rr_space (qb->buf, eom, cp, ancount, hostname,
				 &n_addr, &n_alias, &n_chars) < 0)
    return (NULL);

  l = strlen (hostname);
  h_size = sizeof (*h)
    + (n_addr * (sizeof (addrp_buf[0]) + sizeof (addr_buf[0]))
       + sizeof (addrp_buf[0]))
    + (n_alias) * (1 + sizeof (namep_buf[0])) + sizeof (namep_buf[0])
    + n_chars + l + 1;

  h = xmalloc (h_size);
  namep_buf = (char **) &h[1];
  h->h_aliases = namep_buf;
  h->h_addrtype = AF_INET;
  h->h_length = sizeof (struct in_addr);
  h->h_addr_list = &namep_buf[n_alias+1];
  addrp_buf = (struct in_addr **) h->h_addr_list;
  addr_buf = (struct in_addr *) &h->h_addr_list[n_addr+1];
  h->h_name = (char *) &addr_buf[n_addr];
  name_buf = (char *) h->h_name + l + 1;

  strcpy ((char *) h->h_name, hostname);
  namep_buf[n_alias] = NULL;
  addrp_buf[n_addr] = NULL;

  /* Now extract the actual data */
  for (i = 0; i < ancount; i++) {
    u_short type, class, rdlen;
    u_char name[MAXDNAME];
    int n, m;

    n = dn_expand (qb->buf, eom, cp, name, sizeof (name));
    cp += n;
    GETSHORT (type, cp);
    GETSHORT (class, cp);
    cp += 4;
    GETSHORT (rdlen, cp);
    if (class == C_IN) {
      switch (type) {
      case T_A:
	if (! strcmp (hostname, name)) {
	  addr_buf[c_addr] = *(struct in_addr *) cp;
	  addrp_buf[c_addr] = &addr_buf[c_addr];
	  c_addr++;
	}
	break;
      case T_CNAME:
	m = strlen (name);
	if (c_chars + m < n_chars) {
	  strcpy (name_buf + c_chars, name);
	  dn_expand (qb->buf, eom, cp, name, sizeof (name));
	  if (! strcmp (hostname, name)) {
	    namep_buf[c_alias++] = name_buf + c_chars;
	    c_chars += m;
	  }
	}
	break;
      }
    }
    cp += rdlen;
  }

  assert (name_buf <= (char *) h + h_size);
  return (h);
}

struct mxrec {
  u_short pref;
  char *name;
};

int
mxrec_cmp (const void *_a, const void *_b)
{
  const struct mxrec *a = _a, *b = _b;
  int d = (int) a->pref - (int) b->pref;
  if (d)
    return d;
  return strcasecmp (a->name, b->name);
}

static char **
ar_dns2mxlist (const char *qbraw, int len)
{
  const querybuf * const qb = (querybuf *) qbraw;
  const char * const eom = qb->buf + len;
  const int ancount = ntohs (qb->hdr.ancount);
  const char *cp = getanp (qb->buf, eom);
  u_char cname[MAXDNAME] = "";  /* Cname, if we get one for rname */
  u_char rname[MAXDNAME] = "";  /* Reply name (domain name of MX records) */
  char *np;
  char **mxlist = NULL;
  int n_chars = 0;
  struct mxrec *mxes = xmalloc ((1 + ancount) * sizeof (*mxes));
  int nmx = 0;
  int i;

  if (!cp)
    return NULL;

  for (i = 0; i < ancount; i++) {
    u_short type, class, rdlen;
    u_char name[MAXDNAME];
    int n;

    n = dn_expand (qb->buf, eom, cp, name, sizeof (name));
    cp += n;
    if (n < 0 || cp + 10 > eom) {
      fprintf (stderr, "get_mx_records: response overflow\n");
      goto fail;
    }
    GETSHORT (type, cp);
    GETSHORT (class, cp);
    cp += 4;
    GETSHORT (rdlen, cp);

    if (class == C_IN && type == T_MX) {
      u_char mname[MAXDNAME];
      u_short pr;
      int mxi;

      GETSHORT (pr, cp);
      n = dn_expand (qb->buf, eom, cp, mname, sizeof (mname));
      cp += n;
      if (n < 0 || cp > eom || nmx >= ancount) {
	fprintf (stderr, "get_mx_records: response overflow\n");
	goto fail;
      }

      if (! *rname)
	strcpy (rname, name);
      else if (strcmp (rname, name))
	continue;

      for (mxi = 0; mxi < nmx && strcasecmp (mname, mxes[mxi].name); mxi++)
	;
      if (mxi < nmx) {
	if (pr < mxes[mxi].pref)
	  mxes[mxi].pref = pr;
      }
      else {
	n_chars += strlen (mname) + 1;
	mxes[nmx].pref = pr;
	mxes[nmx].name = xstrdup (mname);
	nmx++;
      }
      continue;
    }
    else if (class == C_IN && type == T_CNAME) {
      u_char vname[MAXDNAME];
      n = dn_expand (qb->buf, eom, cp, vname, sizeof (vname));
      cp += n;
      if (n < 0 || cp > eom || nmx >= ancount) {
	fprintf (stderr, "get_mx_records: response overflow\n");
	goto fail;
      }
      if (!*rname) {
	strcpy (cname, name);
	strcpy (rname, vname);
      }
      else if (!*cname && !strcmp (rname, vname))
	strcpy (rname, vname);
      continue;
    }
    cp += rdlen;
  }
  if (nmx > 1)
    qsort (mxes, nmx, sizeof (*mxes), mxrec_cmp);

  n_chars += strlen (rname) + 1;
  mxlist = xmalloc ((2 + nmx) * sizeof (char *) + n_chars);
  np = (char *) mxlist + (2 + nmx) * sizeof (char *);

  mxlist[0] = np;
  strcpy (np, rname);
  np += strlen (np) + 1;

  for (i = 0; i < nmx; i++) {
    mxlist[i+1] = np;
    strcpy (np, mxes[i].name);
    np += strlen (np) + 1;
  }
  mxlist[i+1] = NULL;

fail:
  for (i = 0; i < nmx; i++)
    xfree (mxes[i].name);
  xfree (mxes);
  return mxlist;
}

/* Send a query to the nameserver */
static int
send_query (const char *dname, u_short type, u_short id)
{
  u_char qb[sizeof (HEADER) + MAXDNAME + 4];
  HEADER * const h = (HEADER *) qb;
  int n;

  n = res_mkquery (QUERY, dname, C_IN, type, NULL, 0, NULL, qb, sizeof (qb));
  if (n < 0)
    return (-1);
  h->id = htons (id);  /* XXX - Why bother with htons? */
  if (send (ar_sock, qb, n, 0) < 0) {
    fprintf (stderr, "ar_aquery: %s\n", strerror (errno));
    return (1);
  }
  return (0);
}

static struct ar_r *
new_request (const char *dname, u_short type,
	     void (*cb) (), void *arg)
{
  struct ar_r *r;

  if (send_query (dname, type, ++ar_lastid) < 0) {
    ar_errno = ARERR_CANTSEND;
    cb (arg, NULL);
    return (NULL);
  }

  r = xmalloc (sizeof (struct ar_r));
  r->arr_id = ar_lastid;
  r->arr_type = type;
  r->arr_toqn = 0;
  r->arr_resend = time (NULL) + MINTO;
  r->arr_name = xstrdup (dname);
  r->arr_cb = cb;
  r->arr_arg = arg;
  TAILQ_INSERT_TAIL (&arh[arhash (r->arr_id)], r, arr_link);
  TAILQ_INSERT_TAIL (&artoq[0], r, arr_to);
  sched_timeout ();
  return (r);
}

void
ar_hostbyname (const char *dname, void (*cb) (void *, struct hostent *),
	       void *arg)
{
  struct ar_r *r;

  /* Handle doted quads. */
  if (isdigit(dname[0])) {
    const char *cp;
    for (cp = dname;; ++cp) {
      if (!*cp) {
	struct in_addr addr;
	if (*--cp == '.')
	  break;
	/* If this is a doted quad, fake a hostent */
	if (!inet_aton(dname, &addr)) {
	  ar_errno = ARERR_REQINVAL;
	  cb (arg, NULL);
	}
	else {
	  struct hostent *h;
	  h = xmalloc (sizeof (*h) + 3 * sizeof (void *) + sizeof (addr)
		       + strlen (dname) + 1);

	  h->h_aliases = (char **) &h[1];
	  h->h_addrtype = AF_INET;
	  h->h_length = sizeof (addr);
	  h->h_addr_list = &h->h_aliases[1];

	  h->h_aliases[0] = NULL;
	  h->h_addr_list[0] = (void *) &h->h_addr_list[2];
	  h->h_addr_list[1] = NULL;

	  *(struct in_addr *) h->h_addr_list[0] = addr;
	  h->h_name = (void *) h->h_addr_list[0] + sizeof (addr);
	  strcpy ((char *) h->h_name, dname);

	  cb (arg, h);
	  xfree (h);
	}
	return;
      }
      if (!isdigit(*cp) && *cp != '.') 
	break;
    }
  }

  r = new_request (dname, T_A, cb, arg);
  if (r)
    r->arr_addr.s_addr = INADDR_NONE;
}

void
ar_hostbyaddr (const u_long addr, void (*cb) (void *, struct hostent *),
	       void *arg)
{
  static const char iaa[] = "in-addr.arpa";
  char dname[16 + sizeof (iaa)];
  u_char *a = (u_char *) &addr;
  struct ar_r *r;

  sprintf (dname, "%d.%d.%d.%d.%s", a[3], a[2], a[1], a[0], iaa);
  if ((r = new_request (dname, T_PTR, cb, arg)))
    r->arr_addr.s_addr = addr;
}

void
ar_mxbyname (const char *dname, void (*cb) (void *, char **), void *arg)
{
  new_request (dname, T_MX, cb, arg);
}

static void
ar_timeout (void *pending_flag)
{
  time_t now;
  struct ar_r *r, *next_r;
  int n;

  *(char *) pending_flag = 0;

  time (&now);
  ar_reap (NULL);

  /* Resend/expire any requests that have come due */
  ar_errno = ARERR_TIMEOUT;
  for (n = MAXSEND - 1; n >= 0; n--)
    for (r = artoq[n].tqh_first; r && r->arr_resend <= now; r = next_r) {
      next_r = r->arr_to.tqe_next;
      if (n + 1 == MAXSEND) {
	/* fprintf (stderr, "DNS lookup of `%s' timed out.\n", r->arr_name); */
	r->arr_cb (r->arr_arg, NULL);
	arr_free (r);
      } else {
	send_query (r->arr_name, r->arr_type, r->arr_id);
	r->arr_resend = now + (MINTO << (n + 1));
	TAILQ_REMOVE (&artoq[n], r, arr_to);
	TAILQ_INSERT_TAIL (&artoq[n + 1], r, arr_to);
	r->arr_toqn = n + 1;
      }
    }

  sched_timeout ();
}

static inline void
reap_a (struct ar_r *r, querybuf *qb, int len)
{
  struct hostent *h;

  ar_errno = 0;
  h = ar_dns2h (qb->buf, len);
  /* If we have an address (from an inverse query), we must verify it. */
  if (h && r->arr_addr.s_addr != INADDR_NONE) {
    struct in_addr **ap;
    for (ap = (struct in_addr **) h->h_addr_list; *ap; ap++)
      if ((*ap)->s_addr == r->arr_addr.s_addr)
	break;
    if (! *ap) {
      fprintf (stderr, "ar_reap: bogus PTR `%s' for %s\n",
	       r->arr_name, inet_ntoa (r->arr_addr));
      ar_errno = ARERR_PTRSPOOF;
      xfree (h);
      h = NULL;
    }
  }
  r->arr_cb (r->arr_arg, h);
  arr_free (r);
  xfree (h);
}

static inline void
reap_ptr (struct ar_r *r, querybuf *qb, int len)
{
  const u_char * const eom = qb->buf + len;
  const u_char *cp = getanp (qb->buf, eom);
  const int ancount = ntohs (qb->hdr.ancount);
  struct ar_r *nr;
  u_char name[MAXDNAME];
  int i;

  if (! cp) {
    r->arr_cb (r->arr_arg, NULL);
    arr_free (r);
    return;
  }

  for (i = 0; i < ancount; i++) {
    u_short class, type, rdlen;
    int n;
    if (cp >= eom) {
      fprintf (stderr, "reap_ptr: garbled DNS reply for `%s'\n", r->arr_name);
      return;
    }
    n = dn_expand (qb->buf, eom, cp, name, sizeof (name));
    cp += n;
    if (n < 0 || cp + 10 > eom) {
      fprintf (stderr, "reap_ptr: garbled DNS reply for `%s'\n", r->arr_name);
      return;
    }
    GETSHORT (type, cp);
    GETSHORT (class, cp);
    cp += 4;
    GETSHORT (rdlen, cp);
    if (class == C_IN && type == T_PTR && !strcmp (r->arr_name, name)) {
      if (dn_expand (qb->buf, eom, cp, name, sizeof (name)) != rdlen) {
	fprintf (stderr, "reap_ptr: garbled PTR for `%s'\n", r->arr_name);
	return;
      }
      nr = new_request (name, T_A, r->arr_cb, r->arr_arg);
      nr->arr_addr = r->arr_addr;
      arr_free (r);
      return;
    }
    cp += rdlen;
  }
}

static inline void
reap_mx (struct ar_r *r, querybuf *qb, int len)
{
  char **mxlist;

  ar_errno = 0;
  mxlist = ar_dns2mxlist (qb->buf, len);
  r->arr_cb (r->arr_arg, mxlist);
  arr_free (r);
  xfree (mxlist);
}

/* Process any replies we've gotten from the name server.  Should be
 * called whenever ar_sock selects readable. */
static void
ar_reap (void *null)
{
  querybuf qb;
  struct ar_r *r;
  int n;

  while ((n = recv (ar_sock, qb.buf, sizeof (qb), 0)) >= 0) {
    if (n < sizeof (HEADER))
      continue;
    r = arh_lookup (ntohs (qb.hdr.id));
    if (!r)
      continue;
    switch (r->arr_type) {
    case T_A:
      reap_a (r, &qb, n);
      break;
    case T_PTR:
      reap_ptr (r, &qb, n);
      break;
    case T_MX:
      reap_mx (r, &qb, n);
      break;
    default:
      fatal ("ar_reap: unknown request type!\n");
      break;
    }
  }
}

void
ar_init (void)
{
  int i;

  if (!(_res.options & RES_INIT))
    res_init ();

  /* We use a reserved port here to prevent port-capture problems.
   * (An unprivileged user can steal packets destined to an INADDR_ANY
   * socket by binding to the same port on the more-specific interface
   * address. */
  ar_sock = inetsocket_resvport (SOCK_DGRAM, INADDR_ANY);
  if (ar_sock < 0)
    fatal ("inetsocket: %s\n", strerror (errno));
  close_on_exec (ar_sock);
  if (connect (ar_sock, (struct sockaddr *) &_res.nsaddr,
	       sizeof (_res.nsaddr)) < 0)
    fatal ("connect: %s\n", strerror (errno));
  make_async (ar_sock);

  for (i = 0; i < ARHSZ; i++)
    TAILQ_INIT (&arh[i]);
  for (i = 0; i < MAXSEND; i++)
    TAILQ_INIT (&artoq[i]);
  cb_add (ar_sock, 0, ar_reap, NULL);
}
