/* $Id: parseopt.c,v 1.1 1998/12/07 20:42:55 fubobb Exp $ */

#include <stdlib.h>
#include <stdarg.h>

#include "arpc/misc.h"
#include "parseopt.h"

char *
getinput (int fd)
{
  int used = 0;
  int size = 2;
  char *data = xmalloc (size);

  for (;;) {
    int n = read (fd, data + used, size - used);
    if (n < 0)
      fatal ("getinput: %s\n", strerror (errno));
    else if (!n) {
      data[used] = '\0';
      return data;
    }
    used += n;
    if (used >= size) {
      size <<= 1;
      data = xrealloc (data, size);
    }
  }
}

char *
getfile (const char *path)
{
  int fd;
  struct stat sb;
  char *data;

  fd = open (path, O_RDONLY);
  if (fd < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    return NULL;
  }
  if (fstat (fd, &sb) < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    return NULL;
  }
  data = xmalloc (sb.st_size + 1);
  data[sb.st_size] = '\0';
  if (read (fd, data, sb.st_size) != sb.st_size) {
    warn ("%s: could not read file\n", path);
    xfree (data);
    return NULL;
  }
  close (fd);
  return data;
}

/* These are obviously less efficient than the ctype.h macros, but
 * isspace actually contains other characters which would make it
 * slightly incompatible with things like strsep (&s, " \t\n").  These
 * at least are consistent.  */
static inline int
isspc (char c)
{
  return c == ' ' || c == '\t' || c == '\n';
}
static inline int
isspcnnl (char c)
{
  return c == ' ' || c == '\t';
}

static inline int
argcpy_i (char *d, char **sp)
{
  char *s = *sp;
  int n = 0;

  while (isspcnnl (*s))
    s++;
  if (!*s || *s == '\n')
    return 0;
  while (*s && !isspc (*s)) {
    if (*s == '\\') {
      if (!*++s)
	return -1;
      /* Ignore backslash newline, plus any subsequent whitespace */
      else if (*s == '\n') {
	while (*++s && isspcnnl (*s))
	  ;
	continue;
      }
    }
    if (d)
      d[n] = *s;
    n++;
    s++;
  }
  if (d)
      d[n] = '\0';
  n++;
  *sp = s;
  return n;
}

static int
argcpy (char *d, char **s)
{
  return argcpy_i (d, s);
}

static int
arglen (char **s)
{
  return argcpy_i (NULL, s);
}

int
vectorlen (char **v)
{
  int n;
  for (n = 0; v[n]; n++)
    ;
  return n;
}

char **
vectordup (char **vi)
{
  char **vo;
  char *co;
  int nv, nc;

  for (nc = nv = 0; vi[nv]; nv++)
    nc += strlen (vi[nv]) + 1;
  vo = xmalloc (++nv * sizeof (*vo) + nc);
  co = (char *) (vo + nv);
  for (nv = 0; vi[nv]; nv++) {
    vo[nv] = co;
    strcpy (co, vi[nv]);
    co += strlen (vi[nv]) + 1;
  }
  vo[nv++] = NULL;
  assert (co = (char *) (vo + nv) + nc);
  return vo;
}

/* We can't actually assume argv is NULL terminated, unfortunetely, so
 * we copy argv before using parseopt. */
char **
vectorize_argv (char **argv, int argc)
{
  char **av;
  assert (argc >= 0);
  av = xmalloc (sizeof (*av) * (argc + 1));
  memcpy (av, argv, argc * sizeof (*av));
  av[argc] = NULL;
  return av;
}

char **
vectorize_line (char **strp)
{
  char **v;
  char *p;
  char *d;
  int n, i, narg, nchr;

  /* Skip leading space (including blank lines) */
  p = *strp;
  while (isspc (*p) || (p[0] == '\\' && p[1] == '\n'))
    p++;
  *strp = p;

  if (!*p)
    return NULL;

  narg = nchr = 0;
  while ((n = arglen (&p)) > 0) {
    narg++;
    nchr += n;
  }
  if (n < 0)
    return NULL;

  v = xmalloc ((1 + narg) * (sizeof (char *)) + nchr);
  d = (char *) v + (1 + narg) * (sizeof (char *));

  p = *strp;
  for (i = 0; i < narg; i++) {
    v[i] = d;
    d += argcpy (d, &p);
  }
  v[i] = NULL;

  assert (d = (char *) v + (1 + narg) * (sizeof (char *)) + nchr);

  *strp = p;
  return v;
}

static int
findopt (char **v, char opt, int witharg, char **val)
{
  char fl[] = {'-', opt, '\0'};
  char **nv;

  for (; !*v || strcmp (*v, fl); v++)
    if (!*v || !strcmp (*v, "--"))
      return 0;
  if (witharg) {
    if (!v[1])
      return -1;
    *val = v[1];
    nv = v + 2;
  }
  else
    nv = v+1;
  while ((*v++ = *nv++))
    ;
  return 1;
}

static void
ensurenomore (char **v, void (*errfn) (void *, const char *, ...),
	      void *errarg, char opt)
{
  char *junk;
  if (findopt (v, opt, 1, &junk)) {
    errfn (errarg, "duplicate '-%c' option\n", opt);
    while (findopt (v, opt, 1, &junk) == 1)
      ;
  }
}

int
convertint (int *out, char *in)
{
  char *e;
  long l;
  l = strtol (in, &e, 0);
  if (*e)
    return -1;
  *out = l;
  return 0;
}

int
parseopt (char **v, void (*errfn) (void *, const char *, ...),
	  void *errarg, ...)
{
  va_list ap;
  char opt;
  enum opt_type type;
  char *val;
  int *intp;
  char **vv;
  int ret = 0;

  va_start (ap, errarg);

  while ((opt = va_arg (ap, char))) {
    switch ((type = va_arg (ap, enum opt_type))) {
    case O_FLAG:
      intp = va_arg (ap, int *);
      for (*intp = 0; findopt (v, opt, 0, NULL); (*intp)++)
	;
      break;
    case O_INT:
      intp = va_arg (ap, int *);
      switch (findopt (v, opt, 1, &val)) {
      case 1:
	if (convertint (intp, val) < 0) {
	  ret = -1;
	  errfn (errarg, "'-%c' given invalid integer '%s'\n", opt, val);
	}
	ensurenomore (v, errfn, errarg, opt);
	break;
      case -1:
	ret = -1;
	errfn (errarg, "option '-%c' requires an integer argument\n", opt);
      }
      break;
    case O_STR:
      switch (findopt (v, opt, 1, va_arg (ap, char **))) {
      case 1:
	ensurenomore (v, errfn, errarg, opt);
	break;
      case -1:
	ret = -1;
	errfn (errarg, "option '-%c' requires an argument\n", opt);
	break;
      }
      break;
    default:
      ret = -1;
      fatal ("parseopt: unknown opt_type %d\n", type);
    }
  }
  va_end (ap);

  for (vv = v; *vv; vv++)
    if ((*vv)[0] == '-') {
      if (strcmp (*vv, "--")) {
	ret = -1;
	errfn (errarg, "unknown option '%s'\n", *vv);
      }
      else {
	while ((vv[0] = vv[1]))
	  vv++;
	return ret;
      }
    }
  return ret;
}

void
warncount (void *_nerrp, const char *fmt, ...)
{
  int *nerrp = _nerrp;
  va_list ap;

  if (nerrp)
    (*nerrp)++;
  va_start (ap, fmt);
#if __GNUC__ >= 2
  /* Can you believe the shit one has to go through to get an
   * __attributes__ (((printf...))) function to take an empty format
   * string? */
  (*(void (*) ())&warn) ("");
#else /* !GCC */
  warn ("");
#endif /* !GCC */
  verr (fmt, ap);
  va_end (ap);
}

#if 0
int
main (int argc, char **argv)
{
  char **v;
  char *fo;
  const char *f;
#if 0
  int a = 0, b = 0;
  char *c = "default";
#endif

  setprogname (argv[0]);
  if (argc != 2)
    fatal ("need one arg\n");

  if (strcmp (argv[1], "-"))
    f = fo = getfile (argv[1]);
  else
    f = fo = getinput (0);

  if (!f)
    fatal ("%s: %s\n", argv[1], strerror (errno));

  while ((v = vectorize_line (&f))) {
    warn ("LINE:\n");
    while (*v)
      warn (" ARG: %s\n", *v++);
  }
  xfree (fo);

#if 0
  v = vectorize (argv[1]);
  parseopt (v, parseopt_err, NULL,
	    'a', O_FLAG, &a,
	    'b', O_INT, &b,
	    'c', O_STR, &c,
	    '\0');
  warn ("a = %d; b = %d; c = \"%s\";\n", a, b, c);

  while (*v)
    warn ("ARG: %s\n", *v++);
#endif

  return 0;
}
#endif
