/* $Id: suio.h,v 1.1 1998/12/07 20:43:02 fubobb Exp $ */

/* Handy-dandy all-purpose maintainer of iovec structures.  Handles
 * partial writev's.  Allocates memory for small writes.  Can even be
 * asked to free memory for you at some later point.
 */

#ifndef _SUIO_H_
#define _SUIO_H_

#include <stdarg.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <memory.h>
#include <assert.h>

#include "arpc/misc.h"
#include "arpc/queue.h"

#ifndef UIO_MAXIOV
#define UIO_MAXIOV 1024
#endif /* UIO_MAXIOV */

#define MINVEC 4

#ifdef SMALL_LIMITS
/* Small limits to help find reallocation bugs */
#define SBUFSIZ 32
#define NDEFIOV 2
#else /* !SMALL_LIMITS */
#define MALLOCRESV 12  /* Maximum # bytes malloc reserves for each 2^n block */
#define SBUFSIZ (256 - sizeof (void *) - sizeof (int) - MALLOCRESV)
#define NDEFIOV ((384 - sizeof (struct s_buf) - 6 * sizeof (int) \
                 - 8 * sizeof (void *) - MALLOCRESV)/sizeof (struct iovec))
#endif /* !SMALL_LIMITS */

struct s_buf {
  struct s_buf *b_next;
  int b_uiono;
  char b_data[SBUFSIZ];
};

struct callfree {
  u_int cf_niov;                      /* Free when uio_nremiov >= cf_niov */
  void *cf_freeme;                    /* Memory that should be freed */
  TAILQ_ENTRY (callfree) cf_link;     /* Linked list */
};

struct suio {
  struct iovec *uio_mem;              /* iovec memory to be freed */
  struct iovec *uio_iov;              /* an iovec array suitable for writev */
  u_int uio_iovcnt;                   /* Number of used iovec structures */
  u_int uio_iovmax;                   /* usable iov before calling newiov */
  u_int uio_resid;                    /* Number of bytes in iovec structure */
  u_int uio_nremiov;                  /* Number of iov's removed by rembytes */
  TAILQ_HEAD (, callfree) uio_cf;     /* Memory that needs to be freed */
  struct s_buf *uio_buf1;             /* Pointer to oldest used buffer */
  struct s_buf *uio_bufp;             /* Pointer to current buffer */
  char *uio_dp;                       /* Pointer to scratch data in buffers */
  char *uio_endp;                     /* End of last iov data */
  int uio_dspace;                     /* Space in current scratch buffer */
  u_int uio_defbfree : 1;             /* Set when uio_defbuf is free */
  u_int uio_memalloced : 1;           /* Set when uio_mem must be freed */
  /* Some initial space to malloc with a suio */
  struct s_buf uio_defbuf;            /* First buffer for converted numbers */
  struct iovec uio_defiov[NDEFIOV];   /* If fewer than NDEFIOV iovecs */
};

extern struct suio *suio_alloc (void);
extern void suio_free (struct suio *);
extern void suio_rembytes (struct suio *, int);
#ifndef DMALLOC
extern char *suio_flatten (struct suio *);
#else /* DMALLOC */
extern char *__suio_flatten (struct suio *, const char *, int);
#define suio_flatten(uio) \
	__suio_flatten(uio, "(suio_flatten)" __FILE__, __LINE__);
#endif /* DMALLOC */
extern void suio_uprintf (struct suio *, const char *, ...)
     __attribute__ ((format (printf, 2, 3)));
extern void suio_vuprintf (struct suio *, const char *, va_list);
extern char *__suio_newbuf (struct suio *);
extern void __suio_newiov (struct suio *);
extern void __suio_fill (struct suio *, char, u_int);
extern void __suio_copy (struct suio *, char *, u_int);
extern void suio_cat (struct suio *, const struct suio *);
extern void suio_move (struct suio *, struct suio *);

static inline void
__suio_addiov (struct suio *uio, char *data, u_int len)
{
  if (len == 0)
    return;
  if (data == uio->uio_endp)
    uio->uio_iov[uio->uio_iovcnt-1].iov_len += len;
  else {
    struct iovec *iov;
    if (uio->uio_iovcnt >= uio->uio_iovmax)
      __suio_newiov (uio);
    iov = &uio->uio_iov[uio->uio_iovcnt++];
    iov->iov_base = data;
    iov->iov_len = len;
    uio->uio_bufp->b_uiono++;
  }
  uio->uio_resid += len;
  uio->uio_endp = data + len;
}

static inline char *
suio_getspace (struct suio *uio, int space)
{
  if (space <= uio->uio_dspace)
    return (uio->uio_dp);
  assert ((unsigned) space <= SBUFSIZ);
  return (__suio_newbuf (uio));
}

static inline char *
suio_getspace_aligned (struct suio *uio, int space)
{
  int pad = (4UL - (u_long) uio->uio_dp) & 3;
  if (uio->uio_dspace < pad + space)
    __suio_newbuf (uio);
  else {
    uio->uio_dspace -= pad;
    uio->uio_dp += pad;
  }
  return suio_getspace (uio, space);
}

static inline void
suio_copy (struct suio *uio, char *data, u_int len)
{
  char *dp;

  if (len > uio->uio_dspace) {
    __suio_copy (uio, data, len);
    return;
  }
  dp = uio->uio_dp;
  memcpy (dp, data, len);
  __suio_addiov (uio, dp, len);
  uio->uio_dspace -= len;
  uio->uio_dp = dp + len;
}

static inline void
suio_print (struct suio *uio, char *data, int len)
{
  if (len < MINVEC && uio->uio_dp != data)
    suio_copy (uio, data, len);
  else {
    __suio_addiov (uio, data, len);
    if (uio->uio_dp == data) {
      uio->uio_dspace -= len;
      uio->uio_dp += len;
    }
  }
}

static inline void
suio_fill (struct suio *uio, char c, int n)
{
  if (n <= uio->uio_dspace) {
    if (n > 0) {
      memset (uio->uio_dp, c, n);
      suio_print (uio, uio->uio_dp, n);
    }
  }
  else
    __suio_fill (uio, c, n);
}

/* Ask suio_rembytes to call free at some later point */
static inline void
suio_callfree (struct suio *uio, void *ptr)
{
  struct callfree *cf = xmalloc (sizeof (struct callfree));
  cf->cf_niov = uio->uio_nremiov + uio->uio_iovcnt;
  cf->cf_freeme = ptr;
  TAILQ_INSERT_TAIL (&uio->uio_cf, cf, cf_link);
}

#endif /*! _SUIO_H_ */
