#include <stdio.h>
#include "DEBUG.h"
#include "auth_db.h"
#include "auth_db_aux.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>

extern const char user_pattern[25];
extern const char group_pattern[25];
extern const char group_mem_pattern[25];

auth_database *auth_setup(char *user_fn, char *group_fn)
{
  FILE *fp;
  auth_database *db;
  auth_user *user;
  auth_group *group;
  auth_userList *users;
  auth_groupList *groups;
  
  DEBUGP(DEBUG_CONCISE,
	 ("auth_setup(%s, %s): entering\n", user_fn, group_fn));
  

  /* Make the pattern matching of db files more configurable */

  /* User database:
         Username : UID : Permissions : Public Key
  */
  sprintf (user_pattern, "%%.%ds : %%d : %%d : %%.%ds\n", 
	   MAX_USERNAME_LEN, MAX_KEY_LEN);

  /* Group database:
         Groupname: GID : Owner UID \n {
  */
  sprintf (group_pattern, "%%.%ds : %%d : %%d\n{\n", 
	   MAX_GROUPNAME_LEN);

  /* Group member:
         UID : Group Key for this user 
  */
  sprintf (group_mem_pattern, "%%d : %%.%ds\n", MAX_KEY_LEN);
  

  if ((db = (auth_database *)malloc(sizeof(auth_database))) == NULL) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_setup: can't malloc db\n"));
    return NULL;
  }

  db->user_fn = strdup(user_fn);

  /* Read users */
  if ((fp = fopen(user_fn, "r")) == NULL) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_setup(%s): can't open user database\n", user_fn));
    free (db);
    return NULL;
  }

  /* In the future, might want to check stat mtime so as
     not to clobber changes in the user file */


  if ((users = (auth_userList *)malloc(sizeof(auth_userList)))
      == NULL) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_setup: can't malloc users\n"));
    free(db);
    return NULL;    
  }

  db->users = users;

  memset (users, 0, sizeof(auth_userList));
    
  while((user = read_user(fp)) != NULL)
    {
      DEBUGP(DEBUG_CONCISE, ("Read a user\n"));
      if (users->head == NULL) {

	/* First user */
	users->head = users->tail = user;
	user->prev = NULL;
	user->next = NULL;

      } else {

	/* Link in another user */
	users->tail->next = user;
	user->prev = users->tail;
	user->next = NULL;
	users->tail = user;

      }

      users->count++;
    }

  fclose (fp);

  
  /* read groups */
  db->group_fn = strdup(group_fn);

  if ((fp = fopen(group_fn, "r")) == NULL) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_setup(%s): can't open group database\n", group_fn));
    /* Need to free stuff here */
    return NULL;
  }

  if ((groups = (auth_groupList *)malloc(sizeof(auth_groupList)))
      == NULL) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_setup: can't malloc groups\n"));
    /* Need to free more stuff here */
    free(db);
    return NULL;    
  }

  db->groups = groups;

  memset (groups, 0, sizeof(auth_groupList));

  while((group = read_group(fp, users)) != NULL)
    {
      if (groups->head == NULL) {

	/* First group */
	groups->head = groups->tail = group;
	group->prev = NULL;
	group->next = NULL;

      } else {

	/* Link in another group */
	groups->tail->next = group;
	group->prev = groups->tail;
	group->next = NULL;
	groups->tail = group;

      }

      groups->count++;
    }

  fclose (fp);

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_setup(%s, %s): exiting\n",  user_fn, group_fn));
  return(db);
}

void auth_shutdown(auth_database *db) 
{

  DEBUGP(DEBUG_CONCISE, ("auth_shutdown(%d): entering\n", db));

  /* Sync changes to disk */
  sync_db(db);

  /* Need to destroy the rest of state here */

  free (db->user_fn);  
  free (db->group_fn);
  free (db);
  DEBUGP(DEBUG_CONCISE, ("auth_shutdown(%d): exiting\n", db));

}

int auth_addUser(auth_database *db, char *name, int uid, int perms, char *key_content, int key_len)
{
  auth_user *user;
  
  DEBUGP(DEBUG_CONCISE, 
	 ("auth_addUser(%d, %s, %d, %d, %s, %d): entering\n",
	  db, name, uid, perms, key_content, key_len));
  
  user = new_user();
  assert (user != NULL);

  /* Check for duplicate user */
  if ((auth_lookupUser(db->users, uid) != NULL) ||
      (auth_lookupUserByName(db->users, name) != NULL)) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_addUser(%d, %s, %d, %d, %s, %d): duplicate UID or username\n",
	    db, name, uid, perms, key_content, key_len));
    destroy_user(user);
    return -1;
  }

  
  if(db->users->tail != NULL)
    {
      db->users->tail->next = user;
      user->prev = db->users->tail;
    }
  else
    db->users->head = user;
  db->users->tail = user;


  user->uid = uid;
  user->name = strdup(name);
  user->perms = perms;
  user->key = new_key();
  user->next = NULL;


  user->key->content = strdup(key_content);
  user->key->len = key_len;

  
  sync_db(db);

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_addUser(%d, %s, %d, %d, %s, %d): exiting\n",
	  db, name, uid, perms, key_content, key_len));

  return 0;  
}


/* return values
   0  user not found
   1  deleted successfully
   -1 user owns a group, error
   */
int auth_deleteUser(auth_database *db, int uid)
{
  auth_user *user;
  auth_group *group;


  DEBUGP(DEBUG_CONCISE, 
	 ("auth_deleteUser(%d, %d): entering\n", db, uid));

  user = auth_lookupUser(db->users, uid);

  if(user == NULL)
    {
      DEBUGP(DEBUG_CONCISE, ("auth_deleteUser(%d, %d): user not found\n",
			     db, uid));
      return 0;
    }

  /* can't delete user if he still owns a group.
     give an error if he does. */
  group = db->groups->head;
  while(group != NULL)
    {
      if(group->owner_uid == uid)
	{
	  DEBUGP(DEBUG_CONCISE, 
		 ("auth_deleteUser(%d, %d): user owns group %d\n",
		  db, uid, group->gid));
	  return -1;
	}
      group = group->next;
    }


  /* not a group owner so let's delete him */
  if(db->users->head == user)
    db->users->head = user->next;
  else
    user->prev->next = user->next;

  if(db->users->tail == user)
    db->users->tail = user->prev;
  else
    user->next->prev = user->prev;

  destroy_user(user);

  
  /* Now remove the user from any groups that might contain it 

     Loop through groups and check each one.
     Do deletion when necessary
     */


  group = db->groups->head;
  while(group != NULL)
    {
      int i;
      auth_user **list = group->user;
      for(i=0; i<group->num_users; i++)
	{
	  if(list[i]->uid == uid)
	    {
	      for(;i<group->num_users-1; i++)
		{
		  list[i] = list[i+1];
		  group->key[i] = group->key[i+1];
		}
	      group->num_users--;
	    }
	}
      group = group->next;
    }

  sync_db(db);


  DEBUGP(DEBUG_CONCISE, 
	 ("auth_deleteUser(%d, %d): exiting\n", db, uid));

  return 1;

}


int auth_user2uid(auth_database *db, char *name)
{
  auth_user *user;

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_user2uid(%d, %s): entering\n", db, name));

  if ((user = auth_lookupUserByName(db->users, name)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_user2uid(%d, %s): no such user\n", db, name));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_user2uid(%d, %s): exiting\n", db, name));

  return user->uid;
}

char* auth_uid2user(auth_database *db, int uid)
{
  auth_user *user;

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_uid2user(%d, %d): entering\n", db, uid));

  if ((user = auth_lookupUser(db->users, uid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_uid2user(%d, %d): no such UID\n", db, uid));
    return NULL;
  }

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_uid2user(%d, %d): exiting\n", db, uid));

  return user->name;
}


/* returns the public key of uid */
char* auth_getUserKey(auth_database *db, int uid, int key_len)
{
  char *key;
  auth_user *user;

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_getuserKey(%d, %d, %d): entering\n", db, uid, key_len));

  if ((user = auth_lookupUser(db->users, uid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_getuserKey(): can't find UID %d\n", uid));
    return NULL;
  }

  key = strdup(user->key->content);
  DEBUGP(DEBUG_CONCISE, ("auth_getUserKey: key = %s\n", key));

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_getuserKey(%d, %d, %d): exiting\n", db, uid, key_len));

  return key;
}


/* sets the public key of uid */
int auth_setUserKey(auth_database *db, int uid, char *key_content, int key_len)
{
  auth_user *user;

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_setUserKey(%d, %d, %s, %d): entering\n", db, uid, key_content, key_len));

  /* Not done */

  user = auth_lookupUser(db->users, uid);
  if(user == NULL)
    return -1;

  strncpy(user->key->content, key_content, key_len);
  sync_db(db);

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_setUserKey(%d, %d, %s, %d): exiting\n", db, uid, key_content, key_len));


  return 0;
}

/* returns the permisions of uid */
int auth_getPerms(auth_database *db, int uid)
{
  auth_user *user;

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_getPerms(%d, %d): entering\n", db, uid));


  if ((user = auth_lookupUser(db->users, uid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_getPerms(): no such UID %d\n", uid));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_getPerms(%d, %d): exiting\n", db, uid));

  return user->perms;
}


/* sets uid's permisions to perms.
   obliterates old perms.
   */
int auth_setPerms(auth_database *db, int uid, int perms)
{
  auth_user *user;

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_setPerms(%d, %d, %d): entering\n", db, uid, perms));

  if ((user = auth_lookupUser(db->users, uid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_setPerms(): no such UID %d\n", uid));
    return -1;
  }

  user->perms = perms;
  sync_db(db);

  DEBUGP(DEBUG_CONCISE, 
	 ("auth_setPerms(%d, %d, %d): exiting\n", db, uid, perms));

  return 0;
}

  

int auth_addGroup(auth_database *db, char *name, int gid, int owner_uid)
{
  auth_group *group;

  DEBUGP(DEBUG_CONCISE,
	 ("auth_addGroup(%d, %s, %d, %d): entering\n", db, name, gid, owner_uid));

  /* Verify the group doesn't already exist */
  if ((auth_lookupGroupByName(db->groups, name) != NULL) ||
      (auth_lookupGroup(db->groups, gid) != NULL)) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_addGroup(): group already exists %d %s\n", gid, name));
    return -1;
  }

  group = new_group();
  assert(group != NULL);

  strncpy(group->name, name, MAX_GROUPNAME_LEN);
  group->gid = gid;
  group->owner_uid = owner_uid;

  /* Insert group */
  if(db->groups->tail == NULL)
    db->groups->head = group;
  else
    db->groups->tail->next = group;
  group->prev = db->groups->tail;
  db->groups->tail = group;
    
  sync_db(db);
  DEBUGP(DEBUG_CONCISE,
	 ("auth_addGroup(%d, %s, %d, %d): exiting\n", db, name, gid, owner_uid));

  return 0;

}

/* return values
   -1  no such gid
    1  deletion occured           */
int auth_deleteGroup(auth_database *db, int gid)
{
  auth_group *group;

  DEBUGP(DEBUG_CONCISE, ("auth_deleteGroup(%d, %d): entering\n", db, gid));

  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_deleteGroup(): no such GID %d\n", gid));
    return -1;
  }

  
  
  /* Fencepost error if head or tail */
  if(group->prev == NULL) /* deleting head */
    db->groups->head = group->next;
  else
    group->prev->next = group->next;
  
  if(group->next == NULL) /* deleting tail */
    db->groups->tail = group->prev;
  else
    group->next->prev = group->prev;
  
  /* if head = tail = group, then head = tail = NULL now */
  
  destroy_group(group);
  sync_db(db);

  
  DEBUGP(DEBUG_CONCISE, ("auth_deleteGroup(%d, %d): exiting\n", db, gid));
  
  return 1;
}

int auth_group2gid(auth_database *db, char* name)
{
  struct auth_group *group;

  DEBUGP(DEBUG_CONCISE, ("auth_group2gid(%d, %s): entering\n", db, name));

  if ((group = auth_lookupGroupByName(db->groups, name)) == NULL) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_group2gid(): no such group name %s\n", name));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE, ("auth_group2gid(%d, %s): exiting\n", db, name));

  return group->gid;
}

char* auth_gid2group(auth_database *db, int gid)
{
  auth_group *group;
  char *name;

  DEBUGP(DEBUG_CONCISE, ("auth_gid2group(%d, %d): entering\n", db, gid));

  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_gid2group(): no such GID %d\n", gid));
    return NULL;
  }
  
  name = strdup(group->name);

  DEBUGP(DEBUG_CONCISE, ("auth_gid2group(%d, %d): exiting\n", db, gid));

  return name;
}


/* returns the group key for gid encrypted with uid's key */
char* auth_getGroupKey(auth_database *db, int gid, int uid, int key_len)
{
  auth_group *group;
  char *key_content;
  int i;

  DEBUGP(DEBUG_CONCISE,
	 ("auth_getGroupKey(%d, %d, %d, %d): entering\n", db, gid, uid, key_len));

  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_getGroupKey(): no such GID %d\n", gid));
    return NULL;
  }


  for (i=0; i < group->num_users; i++) {
    if (group->user[i]->uid == uid) {
      key_content = strdup (group->key[i]->content);
      key_len = group->key[i]->len;

      DEBUGP(DEBUG_CONCISE,
	     ("auth_getGroupKey(%d, %d, %d, %d): exiting\n", db, gid, uid, key_len));
      return key_content;
    }
  }

  DEBUGP(DEBUG_CONCISE,
	 ("auth_getGroupKey(): no such group member %d\n", uid));

  return NULL;
}


int auth_addGroupMember(auth_database *db, int gid, int uid, char *key_content, int key_len)
{
  auth_group *group;
  auth_user *user;
  auth_key *key;

  DEBUGP(DEBUG_CONCISE,
	 ("auth_addGroupMember(%d, %d, %d, %s, %d): entering\n", db, gid, uid, key_content, key_len));

  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_addGroupMember(): no such GID %d\n", gid));
    return -1;
  }

  if ((user = auth_lookupUser(db->users, uid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_addGroupMember(): no such UID %d\n", uid));
    return -1;
  }

  if (group->num_users >= MAX_USERS_PER_GRP) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_addGroupMember(): maxed out members\n"));
    return -1;
  }

  key = new_key();
  assert (key != NULL);

  key->content = strdup (key_content);
  key->len = key_len;

  group->user[group->num_users] = user;
  group->key[group->num_users] = key;

  group->num_users++;

  sync_db(db);

  DEBUGP(DEBUG_CONCISE,
	 ("auth_addGroupMember(%d, %d, %d, %s, %d): exiting\n", db, gid, uid, key_content, key_len));

  return 0;
}

int auth_deleteGroupMember(auth_database *db, int gid, int uid)
{
  auth_group *group;

  DEBUGP(DEBUG_CONCISE,
	 ("auth_deleteGroupMember(%d, %d, %d): entering\n", db, gid, uid));

  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_deleteGroupMember(): no such GID %d\n", gid));
    return -1;
  }

  /* Not done */

  /*  while(groups->group->gid != gid)
    {
      if(groups->next == NULL)
	return -1;
      groups = groups->next;
    }
  list = groups->group->user_list;
  while((list->next != NULL) && (list->next->uid != uid))
    list = list->next;
  del = list->next;
  list->next = list->next->next;
  free(del);
  */

  sync_db(db);

  DEBUGP(DEBUG_CONCISE,
	 ("auth_deleteGroupMember(%d, %d, %d): exiting\n", db, gid, uid));

  return 1;
}

int auth_changeGroupKey(auth_database *db, int gid, int uid, char *key_content, int key_len)
{
  auth_group *group;

  DEBUGP(DEBUG_CONCISE,
	 ("auth_changeGroupKey(%d, %d, %s, %d): entering\n", db, gid, key_content, key_len));

  /* Not done...not sure if this function should even exist */

  if (key_len > MAX_KEY_LEN) {
    DEBUGP(DEBUG_CONCISE,
	   ("auth_changeGroupKey(): key_len %d too long\n", key_len));
    return -1;
  }

  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_changeGroupKey(): no such GID %d\n", gid));
    return -1;
  }

  /*  group->key->len = key_len;
  memset (group->key->content, 0, MAX_KEY_LEN + 1);
  strncpy (group->key->content, key_content, MAX_KEY_LEN);
  */
  sync_db(db);

  DEBUGP(DEBUG_CONCISE,
	 ("auth_changeGroupKey(%d, %d, %s, %d): exiting\n", db, gid, key_content, key_len));

  return 0;
}

int auth_changeGroupOwner(auth_database *db, int gid, int owner_uid)
{
  auth_group *group;
  auth_user *user;

  DEBUGP(DEBUG_CONCISE,
	 ("auth_changeGroupOwner(%d, %d, %d): entering\n", db, gid, owner_uid));

  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_changeGroupOwner(): no such GID %d\n", gid));
    return -1;
  }

  if ((user = auth_lookupUser(db->users, owner_uid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_changeGroupOwner(): no such owner UID %d\n", owner_uid));
    return -1;
  }

  group->owner_uid = owner_uid;
  
  sync_db(db);

  DEBUGP(DEBUG_CONCISE,
	 ("auth_changeGroupOwner(%d, %d, %d): exiting\n", db, gid, owner_uid));

  return 0;
}


int auth_userInGroup(auth_database *db, int uid, int gid)
{
  auth_group *group;
  int i = 0;

  DEBUGP(DEBUG_CONCISE, ("auth_userInGroup(%d, %d, %d): entering\n",db, uid, gid));
  if ((group = auth_lookupGroup(db->groups, gid)) == NULL) {
    DEBUGP(DEBUG_CONCISE, 
	   ("auth_userInGroup(): no such GID %d\n", gid));
    return -1;
  }
  
  while(i < group->num_users)
    {
      if(group->user[i]->uid == uid)
	return 1;
      i++;
    }
    
  
  DEBUGP(DEBUG_CONCISE, ("auth_userInGroup(%d, %d, %d): exiting\n",db, uid, gid));

 
 return 0;
}

