/******************************************
 *  auth_db.h
 *
 *  definitions for authorization server
 ******************************************/

#ifndef _AUTH_DB_H
#define _AUTH_DB_H

#include "csfsauth_prot.h"

#define CSFSAUTH_GRPFILE "/etc/csfsauth_group"
#define CSFSAUTH_USERFILE "/etc/csfsauth_users"

/* Rep of an arbitrary key in ASCII */
typedef struct {
  char *content;
  int len;
} auth_key;

/* cookie describes state, max size of return value */
typedef struct {
  void *state; /* State we last were.  E.g., a pointer to next element */
  int maxsize; /* Max # bytes to return */
  int gen; /* Generation number to detect unatomic changes, staleness */
} auth_cookie;

/* Structure of a particular user */

typedef struct auth_user {
  char *name;
  int uid;
  int perms;
  auth_key *key;
  struct auth_user *next;
  struct auth_user *prev;
} auth_user;

typedef struct {
  int gen;       /* Generation number to detect stale cookies */
  int count;     /* Number of users */
  auth_user *head;
  auth_user *tail;
} auth_userList;

/* Information for a particular group */
typedef struct auth_group {
  char *name;
  int gid;
  int owner_uid;
  int gen;  /* Generation number to detect stale list of members */
  auth_user *user[MAX_USERS_PER_GRP];
  auth_key *key[MAX_USERS_PER_GRP];
  int num_users;   /* Number of members */
  struct auth_group *next;
  struct auth_group *prev;
} auth_group;

typedef struct {
  int gen;  /* Generation number to detect stale cookies */
  int count;     /* Number of groups */
  auth_group *head;
  auth_group *tail;
} auth_groupList;

/* In-memory representation of database */
typedef struct {
  char *user_fn;    /* File of users */
  char *group_fn;   /* File of groups */
  auth_groupList *groups;
  auth_userList *users;
} auth_database;


/* Initialize auth server. Use db files specified by
   filenames. Read entire db info into internal struct.
   */
auth_database*   auth_setup(char *user_fn, char *group_fn);

/* Close down the database, sync to disk */
void             auth_shutdown(auth_database *db);

int              auth_addUser(auth_database *db, char *name, int uid, int perms, char *key_content, int key_len);
int              auth_deleteUser(auth_database *db, int uid);


/* returns the uid associated with name */
int              auth_user2uid(auth_database *db, char *name);

/* returns the username associated with uid */
char*            auth_uid2user(auth_database *db, int uid);

char*            auth_getUserKey(auth_database *db, int uid, int key_len);  /* Make return value opaque */
int              auth_setUserKey(auth_database *db, int uid, char *key_content, int key_len);
int              auth_getPerms(auth_database *db, int uid);
int              auth_setPerms(auth_database *db, int uid, int perms);

int              auth_addGroup(auth_database *db, char *name, int gid, int owner_uid);
int              auth_deleteGroup(auth_database *db, int gid);
int              auth_group2gid(auth_database *db, char *name);
char*            auth_gid2group(auth_database *db, int gid);
char*            auth_getGroupKey(auth_database *db, int gid, int uid, int key_len);

int              auth_addGroupMember(auth_database *db, int gid, int uid, char *key_content, int key_len);
int              auth_deleteGroupMember(auth_database *db, int gid, int uid);
int              auth_changeGroupKey(auth_database *db, int gid, int uid, char *key_content, int key_len);
int              auth_changeGroupOwner(auth_database *db, int gid, int owner_uid);
int              auth_userInGroup(auth_database *db, int uid, int gid);

#endif
