#include <stdio.h>
#include "DEBUG.h"
#include "auth_db.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>

/* internal functions */

const char user_pattern[25];
const char group_pattern[25];
const char group_mem_pattern[25];

auth_key* new_key(void) {
  auth_key* key = malloc(sizeof(auth_key));

  DEBUGP(DEBUG_VERBOSE, ("new_key(): entering\n"));

  key->content = (char*)malloc (sizeof(char) * (MAX_KEY_LEN +1));
  if(key->content == NULL)
    {
      free (key);
      return NULL;
    }
  memset (key->content, 0, MAX_KEY_LEN + 1);
  key->len = 0;
  
  DEBUGP(DEBUG_VERBOSE, ("new_key: exiting\n"));

  return (key);
}

void destroy_key(auth_key *key) {
  DEBUGP(DEBUG_VERBOSE, ("destroy_key(%d): entering\n", key));

  if (key == NULL) {
    DEBUGP(DEBUG_CONCISE, ("destroy_key(%d): key NULL\n", key));
    return;
  }

  free (key->content);
  free (key);
  DEBUGP(DEBUG_VERBOSE, ("destroy_key(%d): exiting\n", key));
}

auth_user* new_user(void)
{
  auth_user *user;

  DEBUGP(DEBUG_VERBOSE, ("new_user(): entering\n"));

  user = (auth_user*)malloc(sizeof(auth_user));
  if(user == NULL)
    return NULL;

  memset(user, 0, sizeof(auth_user));

  user->name = (char*)malloc(sizeof(char) * (MAX_USERNAME_LEN+1));
  if(user->name == NULL) {
    free (user);
    return NULL;
  }

  memset (user->name, 0, MAX_USERNAME_LEN + 1);

  if ((user->key = new_key()) == NULL) {
    free (user->name);
    free (user);
    return NULL;
  }


  DEBUGP(DEBUG_VERBOSE, ("new_user(): exiting\n"));

  return user;
}

void destroy_user(auth_user *user)
{

  DEBUGP(DEBUG_VERBOSE, ("destroy_user(): entering\n"));

  if(user == NULL) {
    DEBUGP(DEBUG_CONCISE, ("destroy_user: attempt to destroy NULL\n"));
    return;
  }

  free (user->name);
  destroy_key(user->key);
  free (user);

  DEBUGP(DEBUG_VERBOSE, ("destroy_user(): exiting\n"));
}


auth_group* new_group(void)
{
  auth_group *group = (auth_group*)malloc(sizeof(auth_group));
  if(group != NULL)
    {
      memset (group, 0, sizeof(auth_group));
      group->name = (char*)malloc(sizeof(char) * (MAX_GROUPNAME_LEN + 1));
      if((group->name == NULL))
	{
	  free(group);
	  return NULL;
	}
      memset (group->name, 0, MAX_GROUPNAME_LEN + 1);
    }

  return group;
}

void destroy_group(auth_group *group) {
  DEBUGP(DEBUG_VERBOSE, ("destroy_group(): entering\n"));

  if(group == NULL) {
    DEBUGP(DEBUG_CONCISE, ("destroy_group: attempt to destroy NULL\n"));
    return;
  }

  free (group->name);
  free (group);

  DEBUGP(DEBUG_VERBOSE, ("destroy_group(): exiting\n"));

}

auth_user* read_user(FILE *user_fh)
{
  auth_user *user; 
  int i;

  DEBUGP(DEBUG_VERBOSE, ("read_user(%d): entering\n", user_fh));

  if ((user = new_user()) == NULL) {
    DEBUGP(DEBUG_CONCISE,
	   ("read_user(%d): new_user() failed\n", user_fh));
    return NULL;
  }

  i = fscanf(user_fh, "%s : %d : %d : %s\n",/*(const char *)user_pattern,*/
	     user->name, 
	     &(user->uid), 
	     &(user->perms),
	     user->key->content);

  if ((i!=4) || (i==EOF)) {
    DEBUGP(DEBUG_CONCISE, ("read_user(%d): fscanf error %d %s %d %d %s\n",
			   user_fh, 
			   i,
			   user->name,
			   user->uid,
			   user->perms,
			   user->key->content));
    destroy_user(user);
    return NULL;
  }
    DEBUGP(DEBUG_CONCISE, ("read_user(%d): read user %s %d %d %s\n",
			   user_fh, 
			   user->name,
			   user->uid,
			   user->perms,
			   user->key->content));

  user->key->len = strlen(user->key->content);

  DEBUGP(DEBUG_VERBOSE, ("read_user(%d): exiting\n", user_fh));

  return user;
}



auth_user* auth_lookupUser(auth_userList *users, int uid)
{
  auth_user *user;
  
  DEBUGP(DEBUG_VERBOSE, ("auth_lookupUser(%d): entering\n", uid));
  
  /* First user */
  user = users->head;

  while(user != NULL)
    {
      if(user->uid == uid)
	break;
      user = user->next;
    }

  DEBUGP(DEBUG_VERBOSE, ("auth_lookupUser(%d): exiting\n", uid));
  return user;
}


auth_user* auth_lookupUserByName(auth_userList *users, char* name)
{
  auth_user *user;

  DEBUGP(DEBUG_VERBOSE,
	 ("auth_lookupUserByName(%s): entering\n", name));
  
  user = users->head;

  while(user != NULL)
    {
      if(strncmp(user->name, name, MAX_USERNAME_LEN) == 0)
	break;
      user = user->next;
    }

  DEBUGP(DEBUG_VERBOSE,
	 ("auth_lookupUserByName(%s): exiting\n", name));

  return user;
}

/* Requires that all users read in already */
auth_group* read_group(FILE *group_fh, auth_userList *users)
{
  auth_group *group;
  auth_user *user;
  auth_key *key;
  int i, uid;
  char c;

  DEBUGP(DEBUG_VERBOSE, ("read_group(%d): entering\n", group_fh));

  group = new_group();

  i = fscanf(group_fh, "%s : %d : %d\n{\n",/*(const char *)group_pattern,*/
		 group->name,
		 &(group->gid),
		 &(group->owner_uid));
  if ((i != 3) || (i == EOF)) {
    destroy_group(group);
    return NULL;
  }

  group->num_users = 0;

  while (1) {

    if((c = fgetc(group_fh)) == '}')
      break;
    else
      ungetc(c, group_fh);
    

    if (group->num_users >= MAX_USERS_PER_GRP) {
      DEBUGP (DEBUG_CONCISE, 
	      ("read_group(): Maxed out group members in %s", group->name));
      destroy_group(group);
      return NULL;
    }

    key = new_key();

    fscanf(group_fh, /* (const char *)group_mem_pattern */ 
	   "%d : %s\n", &uid, key->content);
    key->len = strlen (key->content);


    /* verify user exists */
    if ((user = auth_lookupUser(users, uid)) == NULL) { 
      DEBUGP(DEBUG_CONCISE,
	     ("read_group(): no such UID %d in group %d\n", uid, group->gid));
      destroy_group(group);
      return NULL;
    }

    group->user[group->num_users] = user;
    group->key[group->num_users] = key;

    group->num_users++;
  }
  
  DEBUGP(DEBUG_VERBOSE, ("read_group(%d): exiting\n", group_fh));

  return group;
}

auth_group* auth_lookupGroup(auth_groupList *groups, int gid)
{
  auth_group *group;
  DEBUGP(DEBUG_VERBOSE, ("auth_lookupGroup(%d): entering\n", gid));
  
  group = groups->head;

  while(group != NULL)
    {
      if(group->gid == gid)
	break;
      group = group->next;
    }

  DEBUGP(DEBUG_VERBOSE, ("auth_lookupGroup(%d): exiting\n", gid));

  return group;
}

auth_group* auth_lookupGroupByName(auth_groupList *groups, char *name)
{
  auth_group *group;

  DEBUGP(DEBUG_VERBOSE,
	 ("auth_lookupGroupByName(%s): entering\n", name));
    
  group = groups->head;

  while(group != NULL)
    {
      if(strncmp(group->name, name, MAX_GROUPNAME_LEN) == 0)
	break;
      group = group->next;
    }
  
  DEBUGP(DEBUG_VERBOSE, ("auth_lookupGroupByName(%s): exiting\n", name));
  return group;
}


/* Sync the database to disk */
void sync_db(auth_database *db)
{
  auth_user *user;
  auth_group *group;
  int i;

  FILE *db_file = fopen(db->user_fn, "w");

  DEBUGP(DEBUG_CONCISE, ("sync_db: entering\n"));

  
  /* Should check here if database has been modified since loading.
     E.g., if a user hand modified the database */

  /* Sync users */
  user = db->users->head;

  while(user != NULL)
    {
      fprintf(db_file, (const char *)user_pattern,
	      user->name,
	      user->uid,
	      user->perms,
	      user->key->content);
      DEBUGP(DEBUG_CONCISE, ("sync_db: writing users\n"));
      printf(user_pattern,
	      user->name,
	      user->uid,
	      user->perms,
	      user->key->content);
      user = user->next;
    }
  fclose(db_file);


  /* Sync groups */

  /* Should check here if database has been modified since loading.
     E.g., if a user hand modified the database */

  db_file = fopen(db->group_fn, "w");
  group = db->groups->head;

  while(group != NULL)
    {
      fprintf(db_file, group_pattern, 
	      group->name, 
	      group->gid,
	      group->owner_uid);

      for (i = 0; i < group->num_users; i++)
	{
	  fprintf(db_file, group_mem_pattern,
		  group->user[i]->uid,
		  group->key[i]->content);
	}
      fprintf(db_file, "}\n");
      group = group->next;
    }
  fclose(db_file);
  DEBUGP(DEBUG_CONCISE,("sync_db: exiting\n"));

}







