/* Borrowed from dm's SFS */
/* $Id: rw.c,v 1.1 1999/02/16 18:37:51 fubobb Exp $ */

/* Portions of this file were derived from rsa.c in ssh-1.2.12, which
 * was covered by a license saying:
 *
 * As far as I am concerned, the code I have written for this software
 * can be used freely for any purpose.  Any derived versions of this
 * software must be clearly marked as such, and if the derived work is
 * incompatible with the protocol description in the RFC file, it must
 * be called by a name other than "ssh" or "Secure Shell".  */

/* Rabin-Williams public key algorithm.  From:
 *    A Modification of the RSA Public-Key Encryption Procedure.
 *    H. C. Williams, IEEE Transactions on Information Theory,
 *    Vol. IT-26, No. 6, November, 1980.
 *
 * The cipher is based on Rabin's original signature scheme.  This is
 * NOT RSA, and is NOT covered by the RSA patent.  (That's not to say
 * they couldn't try to sue you anyway for using this, but it's
 * unlikely.  However, if you use this code, you must agree to assume
 * all responsibility for patent infringement.)
 *
 * Two prime numbers p, q, chosen such that:
 *    p = 3 mod 8
 *    q = 7 mod 8
 *
 * [Note:  This means pq = 5 mod 8, so J(2, pq) = -1.]
 *
 * Define N and k as:
 *    N = pq
 *    k = 1/2 * (1/4 * (p-1) * (q-1) + 1)
 *
 * The public key is (N).
 * The private key is (N, k).
 *
 * Define the following four operations.  (Note D2 can only be
 * performed with the secret key.)
 *
 *          { 4(2M+1)       if J(2M+1, N) = 1
 * E1(M) =  { 2(2M+1)       if J(2M+1, N) = -1
 *          { key cracked!  if J(2M+1, N) = 0 (completely improbable)
 *
 * E2(M) =  M^2 % N
 *
 * D2(M) =  M^k % N
 *
 *          { (M/4-1)/2      when M = 0 mod 4
 * D1(M) =  { ((N-M)/4-1)/2  when M = 1 mod 4
 *          { (M/2-1)/2      when M = 2 mod 4
 *          { ((N-M)/2-1)/2  when M = 3 mod 4
 *
 * Public key operations on messages M < (N-4)/8:
 *   Encrypt:  E2(E1(M))
 *   Decrypt:  D1(D2(M))
 *   Sign:     D2(E1(M))
 *   Verify:   D1(E2(M))
 */

#include "misc.h"
#include "rw.h"

int rw_verbose;

struct rw_priv {
  mpz_t n;            /* Modulus (pq) */
  mpz_t p;            /* Small prime */
  mpz_t q;            /* Large prime */
  mpz_t u;            /* q^(-1) mod p */
  mpz_t kp;           /* ((p-1)(q-1)+4)/8 mod p */
  mpz_t kq;           /* ((p-1)(q-1)+4)/8 mod q */
};

int
rw_set_verbose (int new)
{
  int old = rw_verbose;
  rw_verbose = new;
  return old;
}

/* Returns a prime number of the specified number of bits.  The number
 * will have the highest bit set, the two lowest bits set, and the
 * third lowest bit set to bit2.
 */
#include "small_primes.h"
static void
rw_random_prime(MP_INT *ret, RandomState *state, unsigned int bits, int bit2)
{
  MP_INT start, aux;
  unsigned int num_primes;
  int *moduli;
  long difference;

  mpz_init(&start);
  mpz_init(&aux);

 retry:

  /* Pick a random integer of the appropriate size. */
  mpz_random_integer(&start, state, bits);

  /* Set the highest bit. */
  mpz_setbit (&start, bits - 1);
  /* ... and make sure we are 3 or 7 mod 8 */
  mpz_setbit (&start, 0);
  mpz_setbit (&start, 1);
  if (bit2)
    mpz_setbit (&start, 2);
  else
    mpz_clrbit (&start, 2);

  /* Initialize moduli of the small primes with respect to the given
     random number. */
  moduli = xmalloc(num_small_primes * sizeof (moduli[0]));
  if (bits < 16)
    num_primes = 0; /* Don't use the table for very small numbers. */
  else {
    for (num_primes = 0; small_primes[num_primes] != 0; num_primes++) {
      mpz_mod_ui(&aux, &start, small_primes[num_primes]);
      moduli[num_primes] = mpz_get_ui(&aux);
    }
  }

  /* Look for numbers that are not evenly divisible by any of the small
     primes. */
  for (difference = 0; ; difference += 8) {
    unsigned int i;

    if (difference > 0x70000000) {
      /* Should never happen, I think... */
      if (rw_verbose)
	fprintf(stderr, 
		"rw_random_prime: failed to find a prime, retrying.\n");
      xfree(moduli);
      goto retry;
    }

    /* Check if it is a multiple of any small prime.  Note that this
       updates the moduli into negative values as difference grows. */
    for (i = 0; i < num_primes; i++) {
      while (moduli[i] + difference >= small_primes[i])
	moduli[i] -= small_primes[i];
      if (moduli[i] + difference == 0)
	break;
    }
    if (i < num_primes)
      continue; /* Multiple of a known prime. */

    /* It passed the small prime test (not divisible by any of them). */
    if (rw_verbose) {
      printf(".");
      fflush(stdout);
    }

    /* Compute the number in question. */
    mpz_add_ui(ret, &start, difference);

    /* Perform the fermat test for witness 2.  This means:
       it is not prime if 2^n mod n != 2. */
    mpz_set_ui(&aux, 2);
    mpz_powm(&aux, &aux, ret, ret);
    if (mpz_cmp_ui(&aux, 2) == 0) {
      /* Passed the fermat test for witness 2. */
      if (rw_verbose) {
	printf("+");
	fflush(stdout);
      }
      /* Perform a more tests.  These are probably unnecessary. */
      if (mpz_probab_prime_p(ret, 20))
	break; /* It is a prime with probability 1 - 2^-40. */
    }
  }

  /* Found a (probable) prime.  It is in ret. */
  if (rw_verbose) {
    printf("+ (distance %ld)\n", difference);
    fflush(stdout);
  }

  /* Free the small prime moduli; they are no longer needed. */
  xfree(moduli);

  /* Sanity check: does it still have the high bit set (we might have
     wrapped around)? */
  mpz_div_2exp(&aux, ret, bits - 1);
  if (mpz_get_ui(&aux) != 1) {
    if (rw_verbose)
      fprintf(stderr, "rw_random_prime: high bit not set, retrying.\n");
    goto retry;
  }
  mpz_clear(&start);
  mpz_clear(&aux);
  /* Return value already set in ret. */
}

/* Preparing a message for signing/encryption.  The size of the
 * messager (in bits) is stored as the lower 16 bits.  The upper bits
 * (of which there must be at least MIN_RANDOM_BITS left over), are
 * filled with random data.
 */

#define MIN_RANDOM_BITS 64  /* Minimum random padding */
#define NUM_RESVD_BITS 5    /* Number of bits E1 might add to a number */

static int
rw_set_message (mpz_t z, RandomState *state, int bits, mpz_t m)
{
  int padlen;
  int len;
  mpz_t t;

  len = mpz_sizeinbase (m, 2);
  padlen = bits - len - 16;
  if (len > 0xffff || padlen < MIN_RANDOM_BITS) {
    warn ("rw_set_message: message too long\n");
    return -1;
  }

  mpz_mul_2exp (z, m, 16);
  mpz_add_ui (z, z, len);
  mpz_init (t);
  mpz_random_integer (t, state, padlen);
  mpz_mul_2exp (t, t, len + 16);
  mpz_ior (z, z, t);
  mpz_clear (t);
  return 0;
}


static int
rw_get_message (mpz_t m, mpz_t z)
{
  int len;
  int bits;

  len = mpz_get_ui (z) & 0xffff;
  bits = mpz_sizeinbase (z, 2);
  if (len + 16 + MIN_RANDOM_BITS > bits)
    return -1;
  mpz_fdiv_q_2exp (m, z, 16);
  mpz_mod_2exp (m, m, len);
  return 0;
}

static int
E1 (mpz_t out, mpz_t in, mpz_t n)
{
  mpz_mul_2exp (out, in, 1);
  mpz_add_ui (out, out, 1);
  switch (mpz_jacobi (out, n)) {
  case 1:
    mpz_mul_2exp (out, out, 2);
    break;
  case -1:
    mpz_mul_2exp (out, out, 1);
    break;
  case 0:
    warn ("Jacobi 0.  Key factored?\n");
    return -1;
    break;
  }
  if (mpz_cmp (out, n) >= 0) {
    warn ("E1: input too large\n");
    return -1;
  }
  return 0;
}

static void
E2 (mpz_t out, mpz_t in, mpz_t n)
{
  mpz_mul (out, in, in);
  mpz_mod (out, out, n);
}

/* Calculate out = in^k % n.  Use Chinese remainder theorem for speed. */
static void
D2 (mpz_t out, mpz_t in, rw_priv *sk, int rsel)
{
  mpz_t op, oq;

  mpz_init (op);
  mpz_init (oq);

  /* find op, oq such that out % p = op, out % q = oq */
  mpz_powm (op, in, sk->kp, sk->p);
  mpz_powm (oq, in, sk->kq, sk->q);

  /* rsel selects which of 4 square roots */
  if( rsel & 1 ) {
     mpz_sub (op, sk->p, op);
  }
  if( rsel & 2 ) {
     mpz_sub (oq, sk->q, oq);
  }

  /* out = (((op - oq) * u) % p) * q + oq; */
  mpz_sub (out, op, oq);
  mpz_mul (out, out, sk->u);
  mpz_mod (out, out, sk->p);
  mpz_mul (out, out, sk->q);
  mpz_add (out, out, oq);

  mpz_clear (op);
  mpz_clear (oq);
}

static void
D1 (mpz_t out, mpz_t in, mpz_t n)
{
  switch (mpz_get_ui (in) & 3) {
  case 0:
    mpz_sub_ui (out, in, 4);
    mpz_fdiv_q_2exp (out, out, 3);
    break;
  case 1:
    mpz_sub (out, n, in);
    mpz_sub_ui (out, out, 4);
    mpz_fdiv_q_2exp (out, out, 3);
    break;
  case 2:
    mpz_sub_ui (out, in, 2);
    mpz_fdiv_q_2exp (out, out, 2);
    break;
  case 3:
    mpz_sub (out, n, in);
    mpz_sub_ui (out, out, 2);
    mpz_fdiv_q_2exp (out, out, 2);
    break;
  }
}

int
rw_encrypt (mpz_t z, RandomState *state, rw_pub *pk, mpz_t m)
{
  int msgbits = mpz_sizeinbase (pk, 2) - NUM_RESVD_BITS;

  if (rw_set_message (z, state, msgbits, m))
    return -1;

  E1 (z, z, pk);
  E2 (z, z, pk);

  return 0;
}

int
rw_decrypt (mpz_t m, rw_priv *sk, mpz_t z)
{
  D2 (m, z, sk, 0);
  D1 (m, m, sk->n);
  return rw_get_message (m, m);
}


/* RW sign and verify with and without message recovery */
/* Redundancy function based on The Exact Security of Digital */
/* Signatures--How to Sign with RSA and Rabin by Mihir Bellare */
/* and Phillip Rogaway, 1996 */
#define K_0 (SHA1_BITSIZE)
#define K_0BYT (SHA1_SIZE)

/* In paper, g:{0,1}^k_1 -> {0,1}^{k-k_1-1} */
/* Here argument and return are complete bytes. */
/* Let || denote concatenation.  */
/* return reslen bytes of sha1(const||0||x)||sha1(const||1||x)||... */
static void 
g (const u_char *x, int xlen, u_char *res, int reslen)
{
   static u_int32_t gconst = 0xfaeb9513;
   int i = 0;
   sha1_ctx c;
   int resbytes = 0;
   
   while (resbytes < reslen) {
      u_int8_t h[SHA1_SIZE];
      u_int8_t *hb;
      sha1_init (&c);
      sha1_update (&c, (char*)&gconst, sizeof (gconst));
      sha1_update (&c, (char*)&i, sizeof (i));
      sha1_update (&c, x, xlen);
      sha1_final (h, &c);
      i++;
      hb = h;
      while (resbytes < reslen && hb < (h + SHA1_SIZE)) {
         *res++ = *hb++;
         resbytes++;
      }
   }
}
static void 
bs_xor (u_char *a, const u_char *b, int len)
{
   u_char *end = a + len;
   while (a < end) *a++ ^= *b++;
}
#if 0
/* debugging */
static void
bs_print (const u_char *data, int len)
{
   int i;
  for (i = 0; i < len; i++)
     fprintf(stderr, "%02x", data[i]);
}
#endif

static u_int
iov_count (const struct iovec *iov, int iovcnt)
{
   const struct iovec *endp = iov + iovcnt;
   u_int ret = 0;
   for (; iov < endp; iov++)
      ret += iov->iov_len;
   return ret;
}
static u_char *
iov_flatten (const struct iovec *iov, int iovcnt)
{
   const struct iovec *endp = iov + iovcnt;
   u_char *ret = xmalloc (iov_count (iov, iovcnt));
   u_char *dp = ret;
   for (; iov < endp; dp+= iov->iov_len, iov++)
      memcpy (dp, iov->iov_base, iov->iov_len);
   return ret;
}
int 
rw_sign( mpz_t z, const struct iovec* msg, int msgcnt, rw_priv* sk, 
         RandomState *state, int msgrecovery)
{
   u_char r[K_0BYT];
   u_char w[SHA1_SIZE];
   u_char *gd;
   int glen;
   u_char *y;
   u_char *py;
   int i;
   sha1_ctx c;
   int redbits = mpz_sizeinbase (sk->n, 2) - NUM_RESVD_BITS;
   glen = (redbits+7)/8 - SHA1_SIZE;
   gd = xmalloc (glen);

   /* r <-R- {0,1}^k_0 */
   for (i = 0; i < K_0BYT; i++)
      *(r + i) = random_get_byte (state);

   /* w <--- sha1(M||r) */
  sha1_init (&c);
  for (i = 0; i < msgcnt; i++)
     sha1_update (&c, msg[i].iov_base, msg[i].iov_len);
  sha1_update (&c, r, K_0BYT);
  sha1_final (w, &c);
  /* r* <--- g_1 (w) ^ r */
  g (w, SHA1_SIZE, gd, glen);
  bs_xor (r, gd, K_0BYT);
  /* y <--- w || r* || g_2(w) */
  y = xmalloc ((redbits+7)/8);
  py = y;
  memcpy (py, w, SHA1_SIZE);
  py += SHA1_SIZE;
  memcpy (py, r, K_0BYT);
  py += K_0BYT;
  if (msgrecovery) {
     /* M* <--- g_2(w) ^ M */
     /* y <--- w || r* || M* */
     char *flatmsg = iov_flatten (msg, msgcnt);
     int flatlen = iov_count (msg, msgcnt);
     /* make sure message fits */
     assert (flatlen < glen - K_0BYT - 1);
     bs_xor ((gd + K_0BYT), flatmsg, flatlen);
     xfree (flatmsg);
  }
  memcpy (py, gd + K_0BYT, glen - K_0BYT);
  xfree (gd);

  /* Set the mpz_t and get rid of the extra low bits */
  mpz_set_raw (z, y, (redbits+7)/8);
  mpz_tdiv_q_2exp (z, z, 8 - redbits%8);
  xfree (y);
  assert (mpz_sizeinbase (z, 2) <= redbits);

  E1 (z, z, sk->n);
  D2 (z, z, sk, random_get_byte(state));

  return 0;
}
int rw_verify(char* msg, int len, mpz_t s, rw_pub* pk, int msgrecovery)
{
   u_char *gd;
   int glen;
   int ret = 0;
   const u_char *w;
   u_char computedw[SHA1_SIZE];
   u_char *r;
   u_char *gamma;
   sha1_ctx c;
   int redbits = mpz_sizeinbase (pk, 2) - NUM_RESVD_BITS;
   u_char *sig = xmalloc ((redbits+7)/8);
   static char mask[] = {0xFF, 0xFE, 0xFC, 0xF8, 0xF0, 0xE0, 0xC0, 0x80};

   E2 (s, s, pk);
   D1 (s, s, pk);

   /* byte align mpz_t */
   mpz_mul_2exp (s, s, 8 - redbits%8);
   mpz_get_raw ( sig, (redbits+7)/8, s);
   glen = (redbits+7)/8 - SHA1_SIZE;
   gd = xmalloc (glen);
   /* Break y into w, r* and gamma (M*) */
   w = sig;
   r = sig + SHA1_SIZE;
   gamma = sig + SHA1_SIZE + K_0BYT;
   g (w, SHA1_SIZE, gd, glen);
   /* r <--- r* ^ g_1 (w) */
   bs_xor (r, gd, K_0BYT);
   /* zero irrelevent bits of g2 */
   *(gd + glen - 1) &= mask[8 - redbits%8];
   if (msgrecovery) {
      /* M <--- M* ^ g_2(w) */
     assert (len < glen - K_0BYT - 1); /* make sure message fits */
     bs_xor (gamma, gd + K_0BYT, len);
     memcpy (msg, gamma, len);
   } else {
      /* if (g_2(w) != gamma) then reject */
      if (memcmp (gd + K_0BYT, gamma, glen - K_0BYT)) ret = -1;
   }
   sha1_init (&c);
   sha1_update (&c, msg, len);
   sha1_update (&c, r, K_0BYT);
   sha1_final (computedw, &c);
   /* if (sha1 (M||r) != w) then reject */
   if (memcmp (w, computedw, SHA1_SIZE)) ret = -1;

   xfree (gd);
   xfree (sig);
   return ret;
}

void mpz_sha1( u_char* h, mpz_t z)
{
  sha1_ctx c;
  int bylen = (mpz_sizeinbase(z,2)+7)/8;
  char* zraw = (char*)xmalloc( bylen );
 
  mpz_get_raw( zraw, bylen, z );
  sha1_init (&c);
  sha1_update (&c, zraw, bylen);
  sha1_final (h, &c);
  xfree( zraw );
}

static int
rw_precompute (rw_priv *sk)
{
  int ret = 0;
  mpz_t k, p1, q1;

  mpz_init (k);
  mpz_init (p1);
  mpz_init (q1);

  /* Calculate modulus N = pq */
  mpz_mul (sk->n, sk->p, sk->q);

  /* Calculate k = ((p-1)(q-1)/4 + 1)/2 = ((p-1)(q-1) + 4)/8 */
  mpz_sub_ui (p1, sk->p, 1);
  mpz_sub_ui (q1, sk->q, 1);
  mpz_mul (k, p1, q1);
  mpz_add_ui (k, k, 4);
  mpz_fdiv_q_2exp (k, k, 3);

  /* Calculate kp = k % (p-1), and kq = k % (q-1) */
  mpz_mod (sk->kp, k, p1);
  mpz_mod (sk->kq, k, q1);

  /* Calculate u such that (uq) % p = 1 */
  mpz_invert (sk->u, sk->q, sk->p);

  mpz_clear (k);
  mpz_clear (p1);
  mpz_clear (q1);

  return ret;
}

void
rw_priv2pub (rw_pub *pk, rw_priv *sk)
{
  mpz_set (pk, sk->n);
}

void
rw_freepriv (rw_priv *sk)
{
  mpz_clear (sk->n);
  mpz_clear (sk->p);
  mpz_clear (sk->q);
  mpz_clear (sk->u);
  mpz_clear (sk->kp);
  mpz_clear (sk->kq);
  xfree (sk);
}

rw_priv *
rw_keygen (RandomState *state, int bits)
{
  rw_priv *sk = xmalloc (sizeof (*sk));

  mpz_init (sk->n);
  mpz_init (sk->p);
  mpz_init (sk->q);
  mpz_init (sk->u);
  mpz_init (sk->kp);
  mpz_init (sk->kq);

  bits >>= 1;
  rw_random_prime (sk->p, state, bits, 0);
  rw_random_prime (sk->q, state, bits, 1);

  /* Make p smaller than q */
  if (mpz_cmp (sk->p, sk->q) > 0) {
    mpz_t t;
    mpz_init_set (t, sk->p);
    mpz_set (sk->p, sk->q);
    mpz_set (sk->q, t);
    mpz_clear (t);
  }

  rw_precompute (sk);
  return sk;
}

/*
 * Key conversion functions.
 */

char *
rw_pub2ascii (rw_pub *pk)
{
  return mpz_get_str (NULL, 16, pk);
}

int
rw_ascii2pub (rw_pub *pk, const char *str)
{
  return mpz_set_str (pk, str, 16);
}

static const char rwak_ppref[] = "CSFS-RWpriv-V1; p = 0x";
static const char rwak_qpref[] = "; q = 0x";
static const char rwak_end[] = ".";

char *
rw_priv2ascii (rw_priv *sk)
{
  char *buf = xmalloc (sizeof (rwak_ppref) + sizeof (rwak_qpref)
		       + sizeof (rwak_end) - 2
		       + mpz_sizeinbase (sk->p, 16)
		       + mpz_sizeinbase (sk->q, 16));
  char *p = buf;
  strcpy (p, rwak_ppref);
  p += strlen (p);
  mpz_get_str (p, 16, sk->p);
  p += strlen (p);
  strcpy (p, rwak_qpref);
  p += strlen (p);
  mpz_get_str (p, 16, sk->q);
  p += strlen (p);
  strcpy (p, rwak_end);
  return buf;
}

rw_priv *
rw_ascii2priv (const char *ibuf)
{
  char *buf = xstrdup (ibuf);
  rw_priv *sk = xmalloc (sizeof (*sk));
  char *p1, *p2;

  mpz_init (sk->n);
  mpz_init (sk->p);
  mpz_init (sk->q);
  mpz_init (sk->u);
  mpz_init (sk->kp);
  mpz_init (sk->kq);

  if (strncmp (buf, rwak_ppref, sizeof (rwak_ppref) - 1))
    goto error;
  p1 = buf + sizeof (rwak_ppref) - 1;
  if (! (p2 = strstr (p1, rwak_qpref)))
    goto error;
  *p2 = '\0';
  if (mpz_set_str (sk->p, p1, 16))
    goto error;
  p1 = p2 + sizeof (rwak_qpref) - 1;
  if (! (p2 = strstr (p1, rwak_end)))
    goto error;
  *p2 = '\0';
  if (mpz_set_str (sk->q, p1, 16))
    goto error;
  xfree (buf);
  rw_precompute (sk);
  return sk;

error:
  xfree (buf);
  rw_freepriv (sk);
  return NULL;
}

/* I/O from files */

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

int
rw_priv2file (const char *path, rw_priv *sk)
{
  int fd;
  char *ask;

  fd = open (path, O_CREAT|O_EXCL|O_TRUNC|O_WRONLY, 0600);
  if (fd < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    return -1;
  }
  ask = rw_priv2ascii (sk);
  if (write (fd, ask, strlen (ask)) < 0
      || write (fd, "\n", 1) < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    close (fd);
    unlink (path);
    bzero (ask, strlen (ask));
    xfree (ask);
    return -1;
  }
  if (close (fd) < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    close (fd);
    unlink (path);
    bzero (ask, strlen (ask));
    xfree (ask);
    return -1;
  }
  bzero (ask, strlen (ask));
  xfree (ask);
  return 0;
}

int
rw_pub2file (const char *path, rw_pub *sk)
{
  int fd;
  char *ask;

  fd = open (path, O_CREAT|O_EXCL|O_TRUNC|O_WRONLY, 0644);
  if (fd < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    return -1;
  }
  ask = rw_pub2ascii (sk);
  if (write (fd, ask, strlen (ask)) < 0
      || write (fd, "\n", 1) < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    close (fd);
    unlink (path);
    bzero (ask, strlen (ask));
    xfree (ask);
    return -1;
  }
  if (close (fd) < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    close (fd);
    unlink (path);
    bzero (ask, strlen (ask));
    xfree (ask);
    return -1;
  }
  bzero (ask, strlen (ask));
  xfree (ask);
  return 0;
}

rw_priv *
rw_file2priv (const char *path)
{
  int fd;
  char *buf;
  struct stat sb;
  rw_priv *sk;

  fd = open (path, O_RDONLY);
  if (fd < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    return NULL;
  }
  if (fstat (fd, &sb) < 0) {
    close (fd);
    warn ("%s: %s\n", path, strerror (errno));
    return NULL;
  }
  if (sb.st_size > 8192) {
    close (fd);
    warn ("%s: file too large\n", path);
    return NULL;
  }
  buf = xmalloc (sb.st_size + 1);
  if (read (fd, buf, sb.st_size) != sb.st_size) {
    close (fd);
    bzero (buf, sb.st_size + 1);
    xfree (buf);
    warn ("%s: short read\n", path);
    return NULL;
  }
  buf[sb.st_size] = '\0';
  if (! (sk = rw_ascii2priv (buf))) {
    close (fd);
    bzero (buf, sb.st_size + 1);
    xfree (buf);
    warn ("%s: bad contents\n", path);
    return NULL;
  }
  close (fd);
  bzero (buf, sb.st_size + 1);
  xfree (buf);
  return sk;
}

/* Returns 0 if ok, -1 if error */
int
rw_file2pub (rw_pub *sk, const char *path)
{
  int fd;
  char *buf;
  struct stat sb;
  int r;

  fd = open (path, O_RDONLY);
  if (fd < 0) {
    warn ("%s: %s\n", path, strerror (errno));
    return -1;
  }
  if (fstat (fd, &sb) < 0) {
    close (fd);
    warn ("%s: %s\n", path, strerror (errno));
    return -1;
  }
  if (sb.st_size > 8192) {
    close (fd);
    warn ("%s: file too large\n", path);
    return -1;
  }
  buf = xmalloc (sb.st_size + 1);
  if (read (fd, buf, sb.st_size) != sb.st_size) {
    close (fd);
    bzero (buf, sb.st_size + 1);
    xfree (buf);
    warn ("%s: short read\n", path);
    return -1;
  }
  buf[sb.st_size] = '\0';
  if ((r = rw_ascii2pub (sk, buf)) < 0) {
    close (fd);
    bzero (buf, sb.st_size + 1);
    xfree (buf);
    warn ("%s: bad contents\n", path);
    return -1;
  }
  close (fd);
  bzero (buf, sb.st_size + 1);
  xfree (buf);
  return r;
}
