/* Locking code taken from W. Richard Stevens and modified by
 * Jeff Hu.  
 *
 * This complete set of code from Unix Network Programming (1990)
 * by W. Richard Stevens can be found at 
 * http://www.kohala.com/~rstevens/unp.html
 * 
 * This file implements a lock by using a file lock.  Note that this
 * locking mechanism relies on the O_EXCL with the O_CREAT flag
 * resulting in an atomic file creation operation.  Check your system
 * man page to see if this mechanism works.  
 *
 * One way to enforce a mandatory lock (when open is not atomic) is to
 * use the link() system call.  The basic idea is to create a file with
 * open and create a link to the file.  Because link fails if the file
 * exists, it can be used when open is not guaranteed to be atomic.
 * If you use this scheme, you should still check the man page just
 * to make sure.
 *  
 * In general when using a file to lock a resource, you should consult
 * the man pages of your system to see what operations you may use to
 * guarantee atomicity.  In some cases, the locking mechanism may
 * work if you are a user but fail if you are the superuser.
 */

#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>

#define	LOCKFILE	"lock"

int create_lock() {
  return(0);
}

void destroy_lock(int fd) {
  unlink(LOCKFILE);
}

void my_lock(int fd)
{
  int	tempfd;
  
  /*
   * Try to create the lock file, using open() with both
   * O_CREAT (create file if it doesn't exist) and O_EXCL
   * (error if create and file already exists).
   * If this fails, some other process has the lock.
   */
  
  while ( (tempfd = open(LOCKFILE, O_RDWR|O_CREAT|O_EXCL, 0666)) < 0) {
    if (errno != EEXIST)
      perror("open error for lock file");
  }
  close(tempfd);
}

void my_unlock(int fd)
{
  if (unlink(LOCKFILE) < 0)
    perror("unlink error for lock file");
}
