/*
 * Some example code for how to use System V shared memory with 
 * a lock.
 * 
 * This code uses shared memory to have the parent process tell a
 * child how long to sleep.  A mutex is used to serialize access to
 * the shared memory segment.  In the interest of clearly showing how
 * a lock and shared memory might be used for interprocess
 * communication (IPC), I have tried to keep it as simple as possible.
 * Hence the code does not actually produces a clear race condition
 * without locking.  Instead, I have attempted to use the locks to
 * show how two processes would interact with a lock shared between
 * them.
 *
 * Author: Jeff Hu <jhu@mit.edu> 
 */

#include <errno.h>              /* errno and error codes */
#include <stdio.h>              /* for printf() */
#include <unistd.h>             /* for fork() */
#include <signal.h>             /* for kill(), sigsuspend(), others */
#include <sys/ipc.h>            /* for all IPC function calls */
#include <sys/shm.h>            /* for shmget(), shmat(), shmctl() */
#include <string.h>             /* for memset() */
#include <sys/types.h>

/* A few constants that are used for this example. 
 *   SHM_KEY is the key in a key-value pair that associates a number
 *           with a corresponding shared memory segment.
 *   SHARED_SIZE corresponds to the size (in integers) that we wish our 
 *           segment of shared memory to be.
 *   MAX_SLEEP the maximum number of seconds that the child process
 *           can sleep.
 */
#define SHM_KEY         0x4321
#define SHARED_SIZE          1
#define MAX_SLEEP         2000

/* Some variables that are useful to use as globals.
 *   g_shm_id is the identifier of a shared memory segment created by
 *         shmget().
 *   g_shm is a pointer to the shared memory segment.
 *   pid is the return value from the fork command.  It is useful
 *         for determining which process is the parent, and which is 
 *         the child.
 *   lock corresponds to an object upon which a lock is maintained.
 *         In the case of a semaphore, the lock is a semaphore number.
 *         In the case of a file lock, the lock is a file descriptor.
 */
int g_shm_id;
int *g_shm;
int pid;
int lock;


/* Some function prototypes.
 * 
 *   cleanup removes the shared memory and semaphore.  Note that
 *       IPC structures cannot be be contained within the address
 *       space of a process since they are by definition used between
 *       processes.  As a result, the completion of a process does
 *       not automatically result in the clean up of IPC structures.
 *       The clean up must be done explicitly through functions that
 *       deallocate the structures.
 *
 *       The shell commands 'ipcs' and 'ipcrm' can also be used to
 *       list and clean up existing IPC structures.  
 *
 *   signal_handler is called to handle some of the common signals
 *       and to make sure that the IPC structures are destroyed before
 *       closing the processes.
 */
void cleanup();
void signal_handler(int sig);

int create_lock();                  /* Creates the lock */
void destroy_lock( int lock );      /* Destroys the lock */
void my_lock( int lock );           /* Lock a resource */
void my_unlock( int lock );         /* Unlock a resource */

/*
 * The main function
 */
int main( int argc, char *argv[] )
{
  /*
   * Some needed variables
   */
  int c;

  /* Catch the following signals and indicate that the signal_handler 
   * should be called to handle the signals 
   */
  signal(SIGINT, signal_handler);
  signal(SIGQUIT, signal_handler);
  signal(SIGTERM, signal_handler);
  
  /* Create the lock */
  lock = create_lock();

  /* Create the shared memory segment with global read/write permissions */
  if( (g_shm_id = shmget( SHM_KEY, sizeof(int) * SHARED_SIZE, 
			  IPC_CREAT | 0666)) == -1 ) {
    perror("shmget");
    exit(1);
  }

  /* Attach the shared memory segment to the data segment of the process.
   * In other words, this function puts the shared memory segment into
   * the address space of the process.
   */
  if( (g_shm = (int *)shmat(g_shm_id, NULL, 0)) == (int *)-1 ) {
    perror("shmat");
    exit(1);
  }
  /* Initialize the shared memory */
  memset( (void *) g_shm, 0, sizeof(int) * SHARED_SIZE );

  /* Fork into two duplicate processes */
  if( (pid = fork()) == -1 ) {
    perror("fork()");
    exit(-1);
  }
  
  /* Check return value to determine if process is parent or child */
  if( pid == 0 ) {
    /* I am a child */

    while( (*g_shm) != -1 ) {
      my_lock(lock);                  /* Get a lock */

      /* Check to see if an exit message was received by the child. */
      if( (*g_shm) != -1 ) {
	printf("Child is holding lock for %d seconds.\n", *g_shm);
	sleep(*g_shm);
      }
      else
	printf("Child has the lock.\n");

      printf("Child is releasing the lock.\n");
      my_unlock(lock);                /* Unlock the lock */

      sleep(1);        /* Give the parent time to get the lock.
			  Note that this is essentially a bad 
		          hack.  I am assuming that the scheduler
		          will pre-empt this process given a 1
		          second sleep time. */
    }

    exit(0);
  } 
  
  /* I am the parent */
  while( (*g_shm) != -1 ) {
    my_lock(lock);                    /* Get a lock */

    printf("\nParent has the lock.\n");
    printf("Specify the number of seconds the child should hold the lock.\n");
    printf("Press enter to release it or enter 'q' to quit.\n");

    /* Get the value from the keyboard indicating how long the child
     * process should sleep.
     */
    (*g_shm) = 0;               
    while((c = getchar()) != '\n') {
      if( c >= '0' && c <= '9' ) {  /* Parse a number from stdin */
	(*g_shm) = (*g_shm) * 10 + (c - '0');

	if( *g_shm > MAX_SLEEP )    /* Enforce the maximum sleep time. */
	  (*g_shm) = MAX_SLEEP;
      }
      else if( c == 'q' ) {         /* Send a quit value to the child. */
	(*g_shm) = -1;
	while( getchar() != '\n' );  /* Read to the end of the line */
	break;
      }
      else {
	(*g_shm) = 1;
	printf("Invalid duration.  Setting default to 1 second.\n");
	while( getchar() != '\n' );  /* Read to the end of the line */
	break;
      }
	
    }
    printf("Parent is releasing the lock.\n");
    my_unlock(lock);

    sleep(1);        /* Give the child time to get the lock.
			Note that this is essentially a bad 
			hack.  I am assuming that the scheduler
			will pre-empt this process given a 1
			second sleep time. */
  }

  /* Received instruction to quit */
  sleep(1);       /* Give child time to get the message */

  cleanup();
  exit(0);
}


/*
 * This code is a signal handler.  It is called after a signal  
 * and closes things down gracefully.
 */
void signal_handler(int sig) {
  /* Make sure only parent thread performs the handler functions.
     Child just dies now. */
  if( pid == 0 )
    exit(0);

  printf("Signal caught...\n");

  cleanup();
  exit(0);
}

void cleanup() 
{
  destroy_lock(lock); 

  /*
   * Delete the shared memory
   */
  if( shmctl(g_shm_id,IPC_RMID,NULL) != 0 ) {
    perror("Couldn't remove the shared memory segment!\n");
  }
}

  
