#include <errno.h>              /* errno and error codes */
#include <stdio.h>              /* for printf() */
#include <sys/ipc.h>            /* for all IPC function calls */
#include <sys/sem.h>            /* for semget(), semop(), semctl() */
#include <sys/types.h>

/* A few constants that are used for this example. 
 *   SEM_KEY is the key in a key-value pair that associates a number
 *           with a corresponding set of semaphores.
 */
#define SEM_KEY         0x4322

/* Define the lock and unlock operations.  The sembuf structure defines
 * an operation on a semaphore set: the sem_num field indicates which
 * semaphore in the set and the sem_op field indicates what operation.
 * The rules for sem_ops are:
 *    1) if sem_op is 0, block until the semaphore value is 0.
 *    2) if sem_op is negative, block until the value of the semaphore 
 *       is greater than or equal to the absolute value of sem_op.  Then 
 *       decrement the semaphore value by the absolute value of sem_op.
 *    3) if sem_op is positive, the amount indicated is added to
 *       the semaphore value.
 *
 * In this case, the sem_op fields are set to -1 for locking and
 * +1 for unlocking.  Hence, when the semaphore is +1, it is
 * unlocked whereas it is locked when its value is 0.
 *
 * The sem_flg fields are set to SEM_UNDO so that when a process
 * holding a lock exits unexpectedly, the locks that it holds will
 * be released.
 *
 * The reason we can define multiple operations on semaphores
 * is to guarantee atomicity of our actions on semaphores.
 * Given the rules for sem_op, a number of ways to use semaphores exist.
 * However, those "ways" perform multiple operations on semaphores
 * and must be atomic.
 *
 * See the man pages for semop or W. Richard Steven's UNIX Network 
 * Programming for more information.
 */
struct sembuf	g_lock_sembuf[1] =  { {0, -1, SEM_UNDO} };
struct sembuf	g_unlock_sembuf[1] = { {0, 1, SEM_UNDO} };

int create_lock() 
{
  int sem;

  /* Create the semaphore with permissions allowing everyone to
   * read and write to the semaphore.  
   */	
  if( ( sem = semget( SEM_KEY, 1, IPC_CREAT | 0666 ) ) == -1 ) {
    perror("semget");
    exit(-1);
  }
  
  /* Initialize the semaphore.  When created, its value is 0 which
   * corresponds to a locked mutex.  Thus, we add 1 to the current
   * value to unlock the mutex.
   */
  if( semop( sem, g_unlock_sembuf, 1 ) == -1 ) {
    perror("semop");
    exit(1);
  }
  
  return(sem);
}

void destroy_lock( int sem ) 
{
  /*
   * Delete the semaphore
   */
  if( semctl( sem, 0, IPC_RMID, 0) != 0 ) {
    perror("Couldn't remove the semaphore!\n");
  }
}


/* Lock via a semaphore.
 */
void my_lock(int sem) {
  if( semop( sem, g_lock_sembuf, 1 ) == -1 ) {
    perror("semop in lock");
    exit(-1);
  }
}

/* Unlock via a semaphore.
 */
void my_unlock(int sem) {
  if( semop( sem, g_unlock_sembuf, 1 ) == -1 ) {
    perror("semop in unlock");
    exit(-1);
  }
}

