/**
** Portions borrowed from CSFSSD - X-File System
**
** Patrick McCormick
** Kevin Fu
** Joy Nicholson
**
** April 23, 1997
**
** $Id: csfssd.c,v 1.14 1999/04/30 20:56:35 drew42 Exp $
**
** $Log: csfssd.c,v $
** Revision 1.14  1999/04/30 20:56:35  drew42
** fixed statfs
**
** Revision 1.13  1999/04/29 19:27:03  fubob
** i can successfully mkfs a disk:
**
** 	./mkfs 16 16 /tmp/disk
**
** and mount it:
**
** 	./csfssd /tmp/disk
**
** yeah. Zzzz
**
** Revision 1.12  1999/04/28 23:43:44  fubob
** almost done with blockmapping [bmap()] routine and block allocation
** routines.  Painly computation of singly/doubly indirect pointers.
** yuck.
**
** Revision 1.11  1999/04/28 14:50:51  fubob
** wow, I had a major bug in file.c.  Not done fixing, but almost.  Forgot
** entirely the data block allocation routines.
**
** Revision 1.10  1999/04/27 03:36:06  fubob
** fixed lots of little typos
**
** Revision 1.9  1999/04/26 16:14:23  fubob
** fixed silly mistakes (printf problems), fixed inode size to 512 bytes.
**
** Revision 1.8  1999/04/23 14:59:08  fubob
** added new RPCs: unlink, link, create (to get/remove inodes)
**
** pretty much finished inode.c functions.  fixed partition formatting
** procedures in csfssd.c and sb.c
**
** Revision 1.7  1999/04/23 04:49:15  fubob
** finished csfssd.c functions.  finished resetinode in inode.c
**
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "csfssd.h"
/* #include "csfscd.h" */ /* Need the dirent structure... */
#include "disk.h"
#include "DEBUG.h"
#include "ASSERT.h"

/* Kludge to avoid including csfscd.h */

typedef struct dirent_s {
    u32   inode_idx;                   
    u32   len;                 /* if this is zero, this is inactive */
    char  name[248];   /* the name entry */
} dirent_t;

int CSFSSD_mount(csfssd_t *csfssd, char *filename, u8 mount_flags) {

  u32 numblocks = 0;

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_mount(0x%p,%s,%d): Entering\n", 
	  csfssd, filename, mount_flags));
  
  ASSERT(csfssd != NULL);
  ASSERT(filename != NULL);
  

  csfssd->mount_flags = mount_flags;
  
  if ( (csfssd->disk = openDisk(filename, &numblocks)) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_mount: openDisk failed\n"));
    return -1;
  }
  
  DEBUGP(DEBUG_VERBOSE, ("CSFSSD_mount: got file pointer\n"));
  
  if ( (csfssd->sb = (sb_t*)malloc(sizeof(sb_t))) == NULL ) {	
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_mount: malloc(sb_t) failed\n"));
    closeDisk(csfssd->disk);
    return -1;
  }

  csfssd->sb = (sb_t *)malloc(sizeof(sb_t));
  memset (csfssd->sb, 0, sizeof(sb_t));
  csfssd->sb->num_blocks = numblocks;

  if (CSFSSD_readSB(csfssd->sb, csfssd->disk) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_mount: readSB failed.\n"));
    free(csfssd->sb);
    closeDisk(csfssd->disk);
    return -1;
  }

  csfssd->freemap = NULL;

  /* If not cleanly unmounted, first recover disk */
  if (csfssd->sb->dirty == TRUE) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_mount: FS not cleanly mounted.  Checking consistency.\n"));
    if (CSFSSD_fsck(csfssd->sb, csfssd->disk) < 0 ) {
      CSFSSD_panic("CSFSSD_mount: fsck can't recover\n");
    }
  }
  
  if (CSFSSD_readFreemap(csfssd->sb, csfssd->disk, &(csfssd->freemap)) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_mount: readFreemap failed.\n"));
    free(csfssd->sb);
    closeDisk(csfssd->disk);
    return -1;
  }
  
  if (CSFSSD_mcountFreeBlocks(csfssd->sb, csfssd->freemap, 
			      &(csfssd->sb->num_free_blocks)) < 0 ) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_mount: mcountFreeBlocks failed.\n"));
    free(csfssd->freemap);
    free(csfssd->sb);
    closeDisk(csfssd->disk);
    return -1;
  }

  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_mount: csfssd->freeblock_cnt = %lu\n", 
	  csfssd->sb->num_free_blocks));
  
  /* Mark dirty */

  csfssd->sb->dirty = TRUE;
  if (CSFSSD_writeSB(csfssd->sb, csfssd->disk) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_mount: writeSB failed.\n"));
    free(csfssd->freemap);
    free(csfssd->sb);
    closeDisk(csfssd->disk);
    return -1;
  }

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_mount: exiting\n"));
  return 0;
}


int CSFSSD_init(char *filename, u32 numblocks, u32 numinodes)
{
  csfssd_t csfssd;
  inode_t root;   
  dirent_t dirent[2];
  u8 block[CSFSSD_BLOCK_SIZE];
  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_init(%s,%d,%d): Entering\n", 
	  filename, numblocks, numinodes));
  
  ASSERT(filename != NULL);
  ASSERT((numblocks % 8) == 0);
  ASSERT((numinodes % (CSFSSD_BLOCK_SIZE/CSFSSD_INODE_SIZE)) == 0);


  if (createDisk(filename, numblocks) < 0) {
    DEBUGP(1, ("CSFSSD_init: createDisk failed.\n"));
    return -1;
  }


  if ((csfssd.disk = openDisk(filename, (u32 *)&numblocks)) == NULL) {
    DEBUGP(1, ("CSFSSD_init: openDisk failed.\n"));
    return -1;
  }

  
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_init: got file for disk\n"));
  
  if ( (csfssd.sb = (sb_t*)malloc(sizeof(sb_t))) == NULL ) {	
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: malloc(sb_t) failed.\n"));
    closeDisk(csfssd.disk);
    return -1;
  }
  
  /* now we create an SB */
  
  if (CSFSSD_createSB(csfssd.sb, numblocks, numinodes) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_createSB failed.\n"));
    closeDisk(csfssd.disk);
    return -1;
  }    
  
  if (CSFSSD_writeSB(csfssd.sb, csfssd.disk) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_writeSB failed.\n"));
    free(csfssd.sb);
    closeDisk(csfssd.disk);
    return -1;
  }
  
  
  /* make the freemap */
  
  csfssd.freemap = NULL;
  if (CSFSSD_createFreemap(&csfssd) < 0 ) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_createFreemap failed.\n"));
    free(csfssd.sb);
    closeDisk(csfssd.disk);
    return -1;
  }

  /* First block is 0.  Find out remaining blocks for data */
  ASSERT(numblocks > csfssd.sb->first_data_block);
  csfssd.sb->num_free_blocks = numblocks - csfssd.sb->first_data_block;
  
  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_init: freeblocks = %lu, numblocks = %lu, "
	  "first_data_block = %lu\n",
	  csfssd.sb->num_free_blocks,
	  csfssd.sb->num_blocks,
	  csfssd.sb->first_data_block));
  
  
  if (CSFSSD_writeFreemap(csfssd.sb, csfssd.disk, csfssd.freemap) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_writeFreemap failed.\n"));
    free(csfssd.freemap);
    free(csfssd.sb);
    closeDisk(csfssd.disk);
    return -1;
  }

  /* create the root inode */
  
  if (CSFSSD_resetInode(&root, CSFSSD_ROOT_INODE) < 0 ) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_resetInode failed.\n"));
    free(csfssd.freemap);
    free(csfssd.sb);
    closeDisk(csfssd.disk);
    return -1;
  }

  root.link_count = 2;
  root.mode = 0777 | S_IFDIR;

  /* Here we should generate a key and add it to the auth server
     for each user, and set up the key in the lock boxes */

  if (CSFSSD_writeInode(&csfssd, &root, CSFSSD_ROOT_INODE) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_writeInode failed.\n"));
    free(csfssd.freemap);
    free(csfssd.sb);
    closeDisk(csfssd.disk);
    return -1;
  }



  /* Bootstrap the directory contents, which are by
     encrypted in a key accessible by the 'authusers' group.
     This group contains all authenticated users in the system
  */

  /* Kludge as if client daemon */
  ASSERT(sizeof(dirent_t) == 256);
  dirent[0].inode_idx = CSFSSD_ROOT_INODE;  /* point to ourselves */
  dirent[1].inode_idx = CSFSSD_ROOT_INODE;  /* point to parent */
  dirent[0].len = 1;
  dirent[1].len = 2;
  memcpy(dirent[0].name, ".", 1);
  memcpy(dirent[1].name, "..", 2);

  memset(block, 0, CSFSSD_BLOCK_SIZE);
  
  memcpy(block, dirent, sizeof(dirent_t)*2);


  if (CSFSSD_writeData(&csfssd, CSFSSD_ROOT_INODE, 0, 1, block) != 1) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_writeData failed.\n"));
    free(csfssd.freemap);
    free(csfssd.sb);
    closeDisk(csfssd.disk);
    return -1;
  }


  /* TBD
    Set integrity fields in inode, update inode on disk
  */

  if (CSFSSD_writeInode(&csfssd, &root, CSFSSD_ROOT_INODE) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_init: CSFSSD_writeInode failed.\n"));
    free(csfssd.freemap);
    free(csfssd.sb);
    closeDisk(csfssd.disk);
    return -1;
  }

  closeDisk(csfssd.disk);
  
  return 0;
}


int CSFSSD_shutdown(csfssd_t *csfssd)
{
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_shutdown(0x%p): Entering.\n", csfssd));

    ASSERT(csfssd != NULL);
    
    if (CSFSSD_sync(csfssd) < 0 ) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_shutdown: sync failed\n"));
      return (-1);
    }

    if (closeDisk(csfssd->disk) < 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_shutdown: closeDisk failed\n"));
      return (-1);
    }

    if (csfssd->sb != NULL)
      free(csfssd->sb);

    if (csfssd->freemap != NULL) 
      free(csfssd->freemap);

    DEBUGP(DEBUG_CONCISE, ("CSFSSD_shutdown: exiting\n"));

    return(0);
}

int CSFSSD_sync(csfssd_t *csfssd) 
{
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_sync: Entering.\n"));

    ASSERT(csfssd != NULL);

    if ( syncDisk(csfssd->disk) < 0 ) {
      return (-1);
    }

    DEBUGP(DEBUG_CONCISE, ("CSFSSD_sync: Returning.\n"));
    return(0);
}

int CSFSSD_statfs(csfssd_t *csfssd, u16 *bsize, u32 *blocks, 
		  u32 *bfree, u32 *bavail) {
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_statfs: Entering.\n"));
  
  ASSERT(csfssd != NULL);
  ASSERT(csfssd->sb != NULL);

  *bsize  = CSFSSD_BLOCK_SIZE;
  *blocks =  csfssd->sb->num_blocks;
  *bfree  = *bavail = csfssd->sb->num_free_blocks;

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_statfs: exiting.\n"));
  return (0);
}

void CSFSSD_panic(char *message)
{
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_panic: Entering.\n"));

    printf("CSFSSD PANIC: %s", message);

    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_panic: Exiting process with -1.\n"));

    exit(1);
}

