/**
** Portions borrowed from XFS - X-File System
**
** Patrick McCormick
** Kevin Fu
** Joy Nicholson
**
** April 23, 1997
**
** $Id: dirent.c,v 1.3 1999/02/16 18:35:50 fubobb Exp $
**
** $Log: dirent.c,v $
** Revision 1.3  1999/02/16 18:35:50  fubobb
** CSFS current status
**
** Revision 1.2  1998/11/20 03:47:01  fubobb
** server daemon code.
** sine RPC et suffragio.
**
** Revision 1.1  1998/11/12 23:35:28  fubobb
** Server daemon.
**
** Code for the on-disk operations (inodes, data blocks, etc).
**
** Yeah, automake file finally work!
**
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>
#include <sys/errno.h>

#include "csfssd.h"
#include "libdebug/DEBUG.h"
#include "libdebug/ASSERT.h"

/*
** MUSIC: KMFDM, NIHIL. REALLY LOUD.
**        and White Zombie. And NIN Closer. -Pat
** Don't forget the wonderful sounds of Wally Pleasant! -Fubob
*/

#define IS_ACTIVE_DIRENT(x) ((x)->namelen)
#define IS_EMBED_SYMLINK(x) ((x)->inode == CSFSSD_DIRENT_INODE)

#define DIRENTS_PER_BLOCK (CSFSSD_BLOCK_SIZE/sizeof(dirent_t))

/*
** REQUIRES: valid csfssd, dir, block #
** EFFECTS: returns numdirs dirents in data.
** RETURNS: number of directory entries read.
*/
u32 CSFSSD_grab_dirblock(csfssd_t *csfssd, inode_t *dir, u32 block,
			 dirent_t *data)
{
    u32 real;

    ASSERT(csfssd != NULL);
    ASSERT(dir    != NULL);
    ASSERT(data   != NULL);

    real = CSFSSD_readData(csfssd, dir, block*CSFSSD_BLOCK_SIZE,
			   CSFSSD_BLOCK_SIZE, (char*)data);

    DEBUGP(7, ("CSFSSD_grab_dirblock: CSFSSD_readData returns %d.\n", real));

    return (real / sizeof(dirent_t));
}

/*
 * Counts up the number of used directory entries in a directory inode. 
 * Eg, does not count empty fragments
 */
int CSFSSD_count_dirent(csfssd_t *csfssd, inode_t *dir)
{
  int i, j, curr_numents, total_numents;
  dirent_t dirblock[DIRENTS_PER_BLOCK];
  /*  dirent_t dirblock[DIRENTS_PER_BLOCK+1];*/
  
  i = 0;
  total_numents = 0;
  
  while (i < dir->numblocks) {
    curr_numents = CSFSSD_grab_dirblock(csfssd, dir, i, dirblock);
    for (j=0; j < curr_numents; j++) {
      /* Check if directory entry is active */
      if (IS_ACTIVE_DIRENT( &(dirblock[j]) ))
	total_numents++;
    }	
  }
  
  return total_numents;
}

/*
** REQUIRES: Valid csfssd, dir, name of size namelen. dir must be directory 
**           inode.
** data has sizeof(dirent_t)*4 allocated.
** EFFECTS: Searches for a file with the exact name given, returns the dirent
** (maximum four for symlinks) in data. If not found, offset is set to -1
** (0xFFFFFFFF). 
** RETURNS: numdirs on success, 0 on failure.
*/
int CSFSSD_find_dirent(csfssd_t *csfssd, inode_t *dir, char *name, int namelen,
		       dirent_t *data, u32 *offset)
{
  u32 i, j;
  dirent_t dirblock[DIRENTS_PER_BLOCK];
  /*    dirent_t dirblock[DIRENTS_PER_BLOCK+1]; */
  u16 numents;
  u32 numdirs;
#ifdef DEBUG
  char dummy[1024];
#endif
  
  DEBUGP(9, ("CSFSSD_find_dirent: Entering with name = \"%s\"\n", name));
  
  ASSERT(csfssd  != NULL);
  ASSERT(dir     != NULL);
  ASSERT(name    != NULL);
  ASSERT(namelen > 0);
  ASSERT(data    != NULL);
  ASSERT(offset  != NULL);
  
  i = 0;
  numents = 1;
  
  while (i < dir->numblocks) {
    numdirs = CSFSSD_grab_dirblock(csfssd, dir, i, dirblock);
    
    for (j = 0; j < numdirs; j++) {
      /* if this is a good entry */ 
      if ( IS_ACTIVE_DIRENT( &(dirblock[j]) ) &&
	   !IS_EMBED_SYMLINK( &(dirblock[j]) )  &&
	   (strncmp(dirblock[j].name, name, namelen) == 0) ) {
	
	DEBUGP(5, ("CSFSSD_find_dirent: found name = \"%s\".\n", name));
	
	memcpy(data, &(dirblock[j]), sizeof(dirent_t)*numents);
	*offset = (i * CSFSSD_BLOCK_SIZE) + (j * sizeof(dirent_t));
	
	DEBUGP(5, ("CSFSSD_find_dirent: success, returning %d and offset = %lu,\n", 
		   numents, *offset));
	
	return numents;
      }
#ifdef DEBUG
      memcpy(dummy, dirblock[j].name, dirblock[j].namelen);
      dummy[dirblock[j].namelen] = 0;
      DEBUGP(9, ("CSFSSD_find_dirent: skipping name = \"%s\".\n", dummy));
#endif
    }
    i++;
  }
  
  DEBUGP(9, ("CSFSSD_find_dirent: Failed search, returning -1.\n", name));
  
  return 0;
}

/*
** REQUIRES: Valid csfssd, dir, data, offset.
** EFFECTS: Writes the dirents to the directory at the given location.
** RETURNS: number of dirents written.
*/
int CSFSSD_write_dirent(csfssd_t *csfssd, inode_t *dir, dirent_t *data, u32 offset, 
		     u16 numents)
{
    u32 written;
    int retval;

    DEBUGP(9, ("CSFSSD_write_dirent: Entering with offset = %lu, numents = %d\n",
	       offset, numents));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(data != NULL);
    
    written = CSFSSD_writeData(csfssd, dir, offset, numents * sizeof(dirent_t), (char*)data);

    DEBUGP(5, ("CSFSSD_write_dirent: CSFSSD_writeData returned %lu.\n", written));

    ASSERT((written % sizeof(dirent_t)) == 0);

    retval = written/sizeof(dirent_t);

    DEBUGP(9, ("CSFSSD_write_dirent: Returning %d.\n", retval));

    return retval;
}

/*
** REQUIRES: Valid csfssd, dir, data.
** EFFECTS: Writes the dirent block to the directory, somewhere.
** Dirent group will be placed inside one directory block,
** so no need to worry about splitting symlinks.
** RETURNS: Number of dirents written.
*/
int CSFSSD_add_dirent(csfssd_t *csfssd, inode_t *dir, dirent_t *data, u16 numents)
{
    int i, j, startblk;
    dirent_t dirblock[DIRENTS_PER_BLOCK+1];
    u16 written, numdirs;

    DEBUGP(9, ("CSFSSD_add_dirent: Entering with numents = %d\n", numents));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(data != NULL);

    i = 0;

    while (i < dir->numblocks) {
	numdirs = CSFSSD_grab_dirblock(csfssd, dir, i, dirblock);
	
	j = 0;
	startblk = -1;

	while (j < numdirs) {	    
	    if (!IS_ACTIVE_DIRENT(&(dirblock[j]))) {  /* empty slot */

		if (startblk < 0) 
		    startblk = j;
		
		if ( numents <= (j - startblk + 1) ) { /* right size! */

		    /* write to the location starting at block i, entry startblk */

		    written = CSFSSD_write_dirent(csfssd, dir, data, 
					       (i * CSFSSD_BLOCK_SIZE) + 
					       (startblk * sizeof(dirent_t)),
					       numents);

		    DEBUGP(5, ("CSFSSD_add_dirent: Wrote %d dirents in existing block.\n", written));

		    return written;
		}
		
	    }
	    else {
		startblk = -1;
	    }
	    j++;
	}

	/* okay, if we had less than max numdirs then we can add some. */

	if ((DIRENTS_PER_BLOCK - numdirs) >= numents) {  /* we can fit */
	    written = CSFSSD_write_dirent(csfssd, dir, data, 
				       (i * CSFSSD_BLOCK_SIZE) + (numdirs * sizeof(dirent_t)),
				       numents);

	    DEBUGP(5, ("CSFSSD_add_dirent: Wrote %d dirents at offset (%d,%d) in new spaces," 
		       "existing block.\n", written, i, numdirs));
	    
	    return written;
	}
    }

    /* now we just give up and use the next block over */

    written = CSFSSD_write_dirent(csfssd, dir, data, 
			       (i * CSFSSD_BLOCK_SIZE),
			       numents);
    
    DEBUGP(5, ("CSFSSD_add_dirent: Wrote %d dirents in new block at offset (%d,0), returning.\n", 
	       written, i));
    
    return written;
	    
}

/*
 * REQUIRES: Valid csfssd, dir, name.
 * EFFECTS: Creates a new directory entry and inode for the file/directory.
 * RETURNS: 0 on success, -errno on failure.
 */
int CSFSSD_create_new_thing(csfssd_t *csfssd, inode_t *dir, char *name, 
			    int namelen, u32 mode, inode_t **new_inode_ptr)
{
    int err;
    u32 new_inode_num;
    dirent_t *dirent;
    inode_t *inode;

    DEBUGP(9, ("CSFSSD_create_new_thing: Entering with namelen = %u, mode = %o.\n",namelen,mode));

    /* Check that parameters are valid */

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    ASSERT(new_inode_ptr != NULL);


    if (!S_ISDIR(dir->mode)) {
	DEBUGP(1, ("CSFSSD_create_new_thing: dir is not a directory; returning -ENOTDIR.\n"));
	return -ENOTDIR;
    }


    if (namelen > CSFSSD_MAX_NAMELEN) {
	DEBUGP(1, ("CSFSSD_create_new_thing: namelen > %d; returning -ENAMETOOLONG.\n",
		   CSFSSD_MAX_NAMELEN));
	return -ENAMETOOLONG;
    }

    /* malloc the new inode for the file */

    if ( (*new_inode_ptr = (inode_t*)malloc(sizeof(inode_t))) == NULL) {
	DEBUGP(1, ("CSFSSD_create_new_thing: malloc(inode_t) returns NULL! Panicking.\n"));
	CSFSSD_panic("Memory allocation error in CSFSSD_create_new_thing.");
	return -1;
    }    

    /* malloc the dirent */

    if ( (dirent = (dirent_t*)malloc(sizeof(dirent_t))) == NULL) {
	DEBUGP(1, ("CSFSSD_create_new_thing: malloc(dirent_t) returns NULL! Panicking.\n"));
	CSFSSD_panic("Memory allocation error in CSFSSD_create_new_thing.");
	return -1;
    }

    /* Build a new inode and they will come */

    if ( (*new_inode_ptr = (inode_t*)malloc(sizeof(inode_t))) == NULL) {
	DEBUGP(1, ("CSFSSD_create_new_thing: malloc(inode_t) returns NULL! Panicking.\n"));
	CSFSSD_panic("Memory allocation error in CSFSSD_create_new_thing.");
	/* never returns, but to make control flow more solid: */
	return -1;
    }

    /* Now we can find a new inode to exploit for our nefarious purposes. */

    if ( ( err = CSFSSD_allocInode(csfssd, *new_inode_ptr, &new_inode_num, dir->inode_idx) < 0) ) {
	DEBUGP(1, ("CSFSSD_create_new_thing: CSFSSD_allocInode returns %d! Returning %d.\n", err, err));
	return err;
    }
    
    /* New inode is in new_inode, and the inode itself has been reset. */

    inode = *new_inode_ptr;

    inode->mode = mode;

    dir->atime = CSFSSD_get_time();
    dir->mtime = CSFSSD_get_time();
    
    /* Now we add a dirent. */

#ifdef DEBUG
    memset(dirent, DEBUG_FILL, sizeof(dirent_t));
#endif

    dirent->inode = new_inode_num;
    inode->link_count++;  /* we point to the new guy. */
    inode->free = 0;
    dirent->namelen = namelen;
    memcpy(dirent->name, name, namelen);  /* not zero-terminated. */

    if (CSFSSD_writeInode(csfssd,inode,new_inode_num) < 0) {
	return -1;
    }

    if (CSFSSD_writeInode(csfssd,dir,dir->inode_idx) < 0) {
	return -1;
    }

    /* Add this to the directory */
    
    if ((err = CSFSSD_add_dirent(csfssd, dir, dirent, 1)) < 1) {
	DEBUGP(1, ("CSFSSD_create_new_thing: CSFSSD_add_dirent returns %d. Returning -1.\n", err));
	return -1;
    }

    DEBUGP(9, ("CSFSSD_create_new_thing: Returning 0.\n"));

    return 0;
}

int CSFSSD_createFile(csfssd_t *csfssd, inode_t *dir, char *name, int namelen, 
		   u32 mode, inode_t **new_inode_ptr)
{
    int err;
    
    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    ASSERT(new_inode_ptr != NULL);

    DEBUGP(9, ("CSFSSD_createFile: Entering with dir = %p, name = \"%s\", "
	       "namelen = %d, mode = %o\n", dir, name, namelen, mode));

    mode &= (~S_IFMT);  /* make this a regular file. */    
    mode |= S_IFREG;
    
    if ( ( err = CSFSSD_create_new_thing(csfssd, dir, name, namelen, mode, new_inode_ptr) ) < 0 ) {
	DEBUGP(1, ("CSFSSD_createFile: CSFSSD_create_new_thing returned %d, returning.\n", err));
	return err;
    }

    DEBUGP(9, ("CSFSSD_createFile: Returning 0.\n"));

    return 0;
}

int CSFSSD_renameFile(csfssd_t *csfssd, inode_t *olddir, char *oldname, 
		      int oldnamelen, inode_t *newdir, char *newname, 
		      int newnamelen) {

  DEBUGP(9, ("CSFSSD_renameFile: Entering.\n"));

  

  DEBUGP(9, ("CSFSSD_renameFile: Exiting.\n"));

  return 0;
}

int CSFSSD_lookup(csfssd_t *csfssd, inode_t *dir, char *name, int namelen, 
		   inode_t **new_inode_ptr)
{
    dirent_t dirents[4];
    int numdirents;
    u32 offset;

    DEBUGP(9, ("CSFSSD_lookup: Entering with dir = %p, name = \"%s\", "
	       "namelen = %d\n", dir, name, namelen));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    ASSERT(new_inode_ptr != NULL);

    if (!S_ISDIR(dir->mode)) {
	DEBUGP(1, ("CSFSSD_lookup: dir->mode %o is not a directory; returning -ENOTDIR.\n",dir->mode));
	return -ENOTDIR;
    }

    if (namelen > CSFSSD_MAX_NAMELEN) {
	DEBUGP(1, ("CSFSSD_lookup: namelen > %d; returning -ENAMETOOLONG.\n",
		   CSFSSD_MAX_NAMELEN));
	return -ENAMETOOLONG;
    }

    numdirents = CSFSSD_find_dirent(csfssd, dir, name, namelen, dirents, &offset);

    if (numdirents == 0) {
	DEBUGP(9, ("CSFSSD_lookup: Didn't find file, Returning -ENOENT.\n"));
	return -ENOENT;
    }
    
    /* grab the inode and copy it. */

    if ( (*new_inode_ptr = (inode_t*)malloc(sizeof(inode_t))) == NULL ) {
	DEBUGP(1, ("CSFSSD_lookup: failed malloc(sizeof(inode_t)), panicking.\n"));
	CSFSSD_panic("CSFSSD_lookup: failed malloc");
	return -1;
    }

    CSFSSD_readInode(csfssd, *new_inode_ptr, dirents[0].inode);

    DEBUGP(9, ("CSFSSD_lookup: Found \"%s\", Returning 0.\n", name));

    return 0;
}

int CSFSSD_makeDir(csfssd_t *csfssd, inode_t *dir, char *name, int namelen,
		u32 mode, inode_t **new_inode_ptr)
{
    int err;
    dirent_t dirent[2]; /* for . and .. */

    DEBUGP(9, ("CSFSSD_makeDir: Entering with dir = %p, name = \"%s\", "
	       "namelen = %d, mode = %o\n", dir, name, namelen, mode));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    ASSERT(new_inode_ptr != NULL);

    mode &= (~S_IFMT);  /* make this a regular file. */    
    mode |= S_IFDIR;

    if ( ( err = CSFSSD_create_new_thing(csfssd, dir, name, namelen, mode, new_inode_ptr) ) < 0 ) {
	DEBUGP(1, ("CSFSSD_makeDir: CSFSSD_create_new_thing returned %d, returning.\n", err));
	return err;
    }

    /* Now we can write to the inode and create the first two dirents. */

    dirent[0].inode = (*new_inode_ptr)->inode_idx;  /* point to ourselves */
    (*new_inode_ptr)->link_count++;  /* up the refcount */
    dirent[1].inode = dir->inode_idx;  /* point to parent */
    dir->link_count++;  /* up the refcount */
    dirent[0].namelen = 1;
    dirent[1].namelen = 2;
    memcpy(dirent[0].name, ".", 1);
    memcpy(dirent[1].name, "..", 2);

    if ((err = CSFSSD_add_dirent(csfssd, *new_inode_ptr, dirent, 2)) < 1) {
	DEBUGP(1, ("CSFSSD_makeDir: CSFSSD_add_dirent returns %d. Returning -1.\n"));
	return -1;
    }

    DEBUGP(9, ("CSFSSD_makeDir: Returning 0.\n"));

    return 0;
}

int CSFSSD_removeDir(csfssd_t *csfssd, inode_t *dir, char *name, int namelen)
{
    inode_t *inode;
    dirent_t dirent;
    u32 offset;
    int total;

    DEBUGP(9, ("CSFSSD_removeDir: Entering with dir = %p, name = \"%s\", "
	   "namelen = %d\n", dir, name, namelen));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    
    /* find the dirent & inode for target */

    if ( CSFSSD_find_dirent(csfssd, dir, name, namelen, &dirent, &offset) == 0) {
	return -ENOENT;  /* can't find it! */
    }

    /* We have found the inode and it is us. */

    CSFSSD_readInode(csfssd, inode, dirent.inode);

    /* check to see if we can delete target. */

    if ( (total = CSFSSD_count_dirent(csfssd, inode)) > 2 ) {
	DEBUGP(1, ("CSFSSD_removeDir: Directory has %d dirents; returning -ENOTEMPTY.\n", total));
	return -ENOTEMPTY;
    }

    /* XXX - uncaring about link_count we are. */

    /* Copy that, command; all clear to release moorings. */

    /* deactivate dirent. */
    
    dirent.namelen = 0;

    CSFSSD_write_dirent(csfssd, dir, &dirent, offset, 1);  /* commit the directory entity. */
    
    /* then CSFSSD_deallocInode() the inode */

    CSFSSD_deallocInode(csfssd, dirent.inode);
    
    DEBUGP(9, ("CSFSSD_removeDir: Returning 0.\n"));

    return 0;
}

int CSFSSD_unlink(csfssd_t *csfssd, inode_t *dir, char *name, int namelen)
{
    inode_t inode;
    dirent_t dirent;
    u32 offset;

    DEBUGP(9, ("CSFSSD_unlink: Entering with dir = %p, name = \"%s\", "
	       "namelen = %d\n", dir, name, namelen));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    ASSERT(S_ISDIR(dir->mode));
    
    /* find the dirent & inode for target */

    if ( CSFSSD_find_dirent(csfssd, dir, name, namelen, &dirent, &offset) == 0) {
	return -ENOENT;  /* can't find it! */
    }

    /* We have found the inode and it is us. */

    CSFSSD_readInode(csfssd, &inode, dirent.inode);

    /* Copy that, command; all clear to release moorings. */

    /* deactivate dirent. */
    
    dirent.namelen = 0;

    CSFSSD_write_dirent(csfssd, dir, &dirent, offset, 1);  /* commit the directory entity. */

    inode.link_count--; /* de-ref */

    if (inode.link_count == 0) {  /* all gone */
    	/* then CSFSSD_deallocInode() the inode */
	
	CSFSSD_deallocInode(csfssd, dirent.inode);
    }
    else {
	/* write the inode back. */
	CSFSSD_writeInode(csfssd, &inode, dirent.inode);
    }

    DEBUGP(9, ("CSFSSD_unlink: Returning 0.\n"));

    return 0;
}

int CSFSSD_link(csfssd_t *csfssd, inode_t *dir, char *name, int namelen,
	     u32 target_inodenum)
{
    dirent_t dirent;

    DEBUGP(9, ("CSFSSD_link: Entering with dir = %p, name = \"%s\", "
	       "namelen = %d, target_inodenum = %p\n", dir, name, namelen, target_inodenum));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    ASSERT(S_ISDIR(dir->mode));
    
    /* fix up dirent for the new link */

    dirent.inode = target_inodenum;
    dirent.namelen = namelen;
    memcpy(dirent.name, name, namelen);

    CSFSSD_add_dirent(csfssd, dir, &dirent, 1);

    DEBUGP(9, ("CSFSSD_link: Returning 0.\n"));

    return 0;
}

int CSFSSD_readDir(csfssd_t *csfssd, inode_t *dir, int start_ent, int num_ents, dirent_t *dirent)
{
    dirent_t dirblock[DIRENTS_PER_BLOCK+1];
    int i, currblock, curr_numents, total_numents, j;

    ASSERT(csfssd != NULL);
    ASSERT(dirent != NULL);
    ASSERT(dir != NULL);

    i = currblock = total_numents = 0;

    DEBUGP(9, ("CSFSSD_readDir: Entering; start_ent = %u, num_ents = %u.\n",start_ent,num_ents));
    DEBUGP(9, ("CSFSSD_readDir: dir->numblocks = %u\n",dir->numblocks));

    for (currblock = 0; 
	 (currblock < dir->numblocks) && (total_numents < num_ents); 
	 currblock++) {  /* not past end */

	curr_numents = CSFSSD_grab_dirblock(csfssd, dir, currblock, dirblock);

	DEBUGP(9, ("CSFSSD_readDir: currblock = %u, curr_numents = %u, i = %i, total = %u.\n", 
		   currblock, curr_numents, i, total_numents));

	for (j = 0; (j < curr_numents) && (total_numents < num_ents); j++) {  
	    /* bail if we hit our endpoint */

	    /* Check if directory entry is active */

	    if (dirblock[j].namelen!=0) {
		if (i >= start_ent) {    /* ready to start copying? */
		    memcpy(&(dirent[total_numents]), &(dirblock[j]), sizeof(dirent_t));  
		    /* copy it into output buffer */
		    total_numents++;
		}
		i++;
	    }
	}
    }

    DEBUGP(9, ("CSFSSD_readDir: Returning %u.\n",total_numents));

    return total_numents;
}

int CSFSSD_symlink(csfssd_t *csfssd, inode_t *dir, char *name, int namelen,
		char *target_inodenum, int tgtlen)
{
    int err;
    u32 bytes;
    inode_t *inode;
    u32 mode;

    DEBUGP(9, ("CSFSSD_symlink: Entering with dir = %p, name = \"%s\", "
	       "namelen = %d, target_inodenum = \"%s\", tgtlen = %d\n", 
	       dir, name, namelen, target_inodenum, tgtlen));

    ASSERT(csfssd != NULL);
    ASSERT(dir != NULL);
    ASSERT(name != NULL);
    ASSERT(target_inodenum != NULL);

    mode = 0777 | S_IFLNK;

    if (( err = CSFSSD_create_new_thing(csfssd, dir, name, namelen, mode, &inode)) < 0) {
	DEBUGP(1, ("CSFSSD_makeDir: CSFSSD_create_new_thing returned %d, returning.\n", err));
	return err;
    }

    if ((bytes = CSFSSD_writeData(csfssd, dir, 0, tgtlen, target_inodenum)) < tgtlen) {
	DEBUGP(1, ("CSFSSD_makeDir: CSFSSD_writedata returned %lu, returning.\n", bytes));
	return -1;
    }
    
    DEBUGP(9, ("CSFSSD_symlink: Returning 0.\n"));

    return 0;
}

int CSFSSD_readsymlink(csfssd_t *csfssd, inode_t *dir, char *srcname, 
		       int srcnamelen, char *tgtname, int tgtnamelen) {


  DEBUGP(9, ("CSFSSD_readsymlink: Entering.\n"));

  DEBUGP(9, ("CSFSSD_readsymlink: Exiting.\n"));

  return tgtnamelen;
}

int CSFSSD_followlink(csfssd_t *csfssd, inode_t *dir, char *name, 
		      int namelen, inode_t *inode) {

  DEBUGP(9, ("CSFSSD_followlink: Entering.\n"));

  DEBUGP(9, ("CSFSSD_followlink: Exiting.\n"));

  return 0;
}
