/**
** Portions borrowed from XFS - X-File System
**
** Patrick McCormick
** Kevin Fu
** Joy Nicholson
**
** April 23, 1997
**
** $Id: file.c,v 1.13 1999/04/29 19:27:03 fubob Exp $
**
** $Log: file.c,v $
** Revision 1.13  1999/04/29 19:27:03  fubob
** i can successfully mkfs a disk:
**
** 	./mkfs 16 16 /tmp/disk
**
** and mount it:
**
** 	./csfssd /tmp/disk
**
** yeah. Zzzz
**
** Revision 1.12  1999/04/29 15:51:06  fubob
** bmap() stuff
**
** Revision 1.11  1999/04/29 15:22:09  fubob
** freemap block allocation mode modulized
**
** bmap done for direct and singly indirect data pointers.  Not done
** for doubly indirect.  untested, and really ugly code with lots of
** #define'd macros. yum yum.
**
** Revision 1.10  1999/04/29 05:47:14  fubob
** i hate bmap.
**
** Revision 1.9  1999/04/28 23:43:44  fubob
** almost done with blockmapping [bmap()] routine and block allocation
** routines.  Painly computation of singly/doubly indirect pointers.
** yuck.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "disk.h"
#include "csfssd.h"
#include "csfssd_prot.h"
#include "DEBUG.h"
#include "ASSERT.h"
#include "misc/types.h"

/* Internal procedures */

/* Translates block_id to either
   direct==0, sindirect==1, dindir==2 */
#define dp_level(block_id) ( \
   ((block_id) > CSFSSD_NUM_DIRECT + \
    (CSFSSD_BLOCK_SIZE/sizeof(csfssd_dp_t))*CSFSSD_NUM_SINDIRECT) + \
   ((block_id) > CSFSSD_NUM_DIRECT + \
    (CSFSSD_BLOCK_SIZE/sizeof(csfssd_dp_t))*CSFSSD_NUM_SINDIRECT + \
    (CSFSSD_BLOCK_SIZE/sizeof(csfssd_dp_t)) \
    *(CSFSSD_BLOCK_SIZE/sizeof(csfssd_dp_t)) \
    *CSFSSD_NUM_DINDIRECT))
     
/* Sets *blocknum to dp and allocates new space if necessary 
   Sets flag = 1 if we allocate a block 
*/
#define get_dp(dp, alloc, flag) \
     if (((*blocknum = (dp)) == 0) \
	 && ((alloc))) { \
       (flag) = 1; \
       if (CSFSSD_allocBlocks(csfssd, 0, 1, \
			      &((dp))) < 0) { \
	 DEBUGP(DEBUG_CONCISE, ("CSFSSD_bmap: allocBlocks failed\n")); \
	 return -1; \
       } \
       *blocknum = (dp); \
     }


#define CSFSSD_NUM_PTRS (CSFSSD_BLOCK_SIZE/sizeof(csfssd_dp_t))

/* Requires: csfssd, inode, and blocknum allocated.  csfssd, inode,
 * block_id, and alloc set.  Only works for data blocks, not
 * inode/freemap/super blocks.
 *
 * Effects: Maps the block_id'th block of the file to a physical block
 * on disk.  The resultant can be used as an argument to readBlock()
 * and writeBlock().  If alloc=1, will allocate new blocks if necessary (only
 * makes sense when calling bmap and then writing blocks).  Otherwise
 * resolving an unused block_id implies a block of zeros (a hole) and
 * the blocknum = 0 is returned.
 *
 * Returns: The physical block number of this data block relative to
 * the start of the disk in the blocknum pointer.  Returns 0 if ok, -1
 * otherwise.  
 */
u32 CSFSSD_bmap(csfssd_t *csfssd, inode_t *inode, u32 block_id, 
		u8 alloc, u32 *blocknum) 
{
  u32 temp_index;
  csfssd_dp_block_t *dpb;
  u32 s_blocknum, d_blocknum; /* Singly/doubly indrect blocks */
  u8 level, flag = 0;
  

  DEBUGP(DEBUG_VERBOSE, ("CSFSSD_bmap(%p,%p,%d,%d,%p): Entering\n",
			 csfssd, inode,  block_id, alloc, blocknum));

  level = dp_level(block_id);

  if (level == 0) {

    /* Data goes in a direct data block */

    DEBUGP(DEBUG_VERBOSE, ("CSFSSD_bmap: direct block\n"));
    get_dp (inode->block_ptr[block_id].dp, alloc, flag);


    DEBUGP(DEBUG_VERBOSE, ("CSFSSD_bmap: flag %d, alloc %d, blocknum %d\n",
			   flag, alloc, *blocknum));
    

    /* Write new DP's to inode table if newly allocated */
    if (flag && alloc
	&& (CSFSSD_writeInode(csfssd, inode, inode->inode_idx) < 0)) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_bmap: can't write inode.\n"));
      return -1;
    }
    
  } else if (level == 1) {

    /* Data goes through singly indirect block */
    DEBUGP(DEBUG_VERBOSE, ("CSFSSD_bmap: sindirect block\n"));

    /* Find which singly indirect data block pointer */
    temp_index = CSFSSD_NUM_DIRECT - 1 +
      (block_id - CSFSSD_NUM_DIRECT)/CSFSSD_NUM_PTRS;

    get_dp(inode->block_ptr[temp_index].dp, alloc, flag);

    /* Write new DP block if newly allocated */
    if (alloc && flag
	&& (CSFSSD_writeInode(csfssd, inode, inode->inode_idx) < 0)) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_bmap: can't write inode.\n"));
      return -1;
    }

    s_blocknum = *blocknum;

    /* If a hole in the file, we're done */
    if (s_blocknum != 0) {

      /* Retrieve the singly indirect data pointer block */
      dpb = malloc(CSFSSD_BLOCK_SIZE);

      if (flag) {
	/* Newly allocated DP block, hence a hole */
	memset(dpb, 0, CSFSSD_BLOCK_SIZE);
	
      } else {
	/* Already existing block, read from disk */
	if (readBlock(csfssd->disk, s_blocknum, (u8 *)dpb) < 0) {
	  DEBUGP(DEBUG_CONCISE, ("CSFSSD_bmap: readBlock of %d failed\n",
				 s_blocknum));
	  return -1;
	}
      }

      /* The singly indirect DP block now is in dpb */

      /* Find block_id's member of this singly indirect data block */
      temp_index = (block_id - CSFSSD_NUM_DIRECT) % CSFSSD_NUM_PTRS;
	
      /* See if pointer to data block allocated */
      get_dp(dpb->dp[temp_index], alloc, flag);
	
      /* Write new DP block if newly allocated */
      if (alloc && flag
	  && (writeBlock(csfssd->disk, s_blocknum, dpb) < 0)) {
	DEBUGP(DEBUG_CONCISE, ("CSFSSD_bmap: can't write sindir block\n"));
	return -1;
      }
    }    
  } else if (level == 2) {

    /* Doubly direct */
    DEBUGP(DEBUG_VERBOSE, ("CSFSSD_bmap: dindirect block\n"));
    return -1;

  } else {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_bmap: invalid level %d\n", level));
    return -1;
  }


  DEBUGP(DEBUG_VERBOSE, 
	 ("CSFSSD_bmap: Exiting block_id %d, inode_idx %d -> disk block %d\n",
	  block_id, inode->inode_idx, *blocknum));
  return 0;
}


/* Public procedures */

u16 CSFSSD_readData(csfssd_t *csfssd, u32 inode_idx,
		    u32 first_block_id, u16 numblocks, u8 *blocks)
{
  inode_t inode;
  u32 i, blocknum;

   
  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_readData(%p,%d,%d,%d,%p): Entering\n",
	  csfssd, inode_idx, first_block_id, numblocks, blocks));
  
  ASSERT(csfssd   != NULL);
  ASSERT(blocks   != NULL);
  
  
  if (CSFSSD_readInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_readData: can't read inode.\n"));
    return 0;
  }
  
  
  /* NEEDS FIELD TO CHECK UID HERE !!! */
  
  if (first_block_id + numblocks >
      inode.size/CSFSSD_BLOCK_SIZE +
      ((inode.size % CSFSSD_BLOCK_SIZE) != 0)) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_readData: read past end of file. truncating read\n"));
    numblocks = MAX (inode.size/CSFSSD_BLOCK_SIZE 
		     + ((inode.size % CSFSSD_BLOCK_SIZE) != 0)
		     - first_block_id,
		     0);
  }
  
  
  for (i = 0; i < numblocks; i++) {

    if (CSFSSD_bmap(csfssd, &inode, first_block_id + i, 0, &blocknum)
	< 0) {
      DEBUGP(DEBUG_CONCISE,
	     ("CSFSSD_readData: can't bmap %ld\n", first_block_id + i));
      return 0;
    }


    if (blocknum == 0) {
      memset((u8 *)(blocks + CSFSSD_BLOCK_SIZE * i), 0, CSFSSD_BLOCK_SIZE);
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_readData: hole in file, returning block of 0's\n"));

    } else {

      if (readBlock (csfssd->disk, blocknum,
		     (u8 *)(blocks + CSFSSD_BLOCK_SIZE * i)) < 0) {
	DEBUGP(DEBUG_CONCISE,
	       ("CSFSSD_readData: can't read block id %ld\n", i));
	return 0;
      }
    }
  }
  
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_readData: exiting\n"));
  
  return numblocks;
}

u16 CSFSSD_writeData(csfssd_t *csfssd, u32 inode_idx,
		     u32 first_block_id, u16 numblocks, u8 *blocks)
{
  inode_t inode;
  u32  i, blocknum;
  
  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_writeData(%p,%d,%d,%d,%p): Entering\n",
	  csfssd, inode_idx, first_block_id, numblocks, blocks));
  
  ASSERT(csfssd != NULL);
  ASSERT(blocks != NULL);

  if (CSFSSD_readInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeData: can't read inode.\n"));
    return 0;
  }
  
  
  /* NEEDS FIELD TO CHECK UID HERE !!! */
  
  if (first_block_id
      == inode.size/CSFSSD_BLOCK_SIZE + 
      ((inode.size % CSFSSD_BLOCK_SIZE) != 0)) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeData: appending to file\n"));

  } else if (first_block_id
	     > inode.size/CSFSSD_BLOCK_SIZE + 
	     ((inode.size % CSFSSD_BLOCK_SIZE) != 0)) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeData: warning, leaving gap, writing past EOF\n"));
  }
  
  
  for (i = 0; i < numblocks; i++) {
    if (CSFSSD_bmap(csfssd,
		    &inode,
		    first_block_id + i, 
		    1,
		    &blocknum) < 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeData: bmap error\n"));
      return 0;
    }
    if (writeBlock (csfssd->disk, blocknum,
		    (u8 *)(blocks + CSFSSD_BLOCK_SIZE * i)) < 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeData: can't write block id %ld\n", i));
      return 0;
    }
  }
  

  /* might want to update inode file size and mod times? nah, let CD handle it */

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeData: exiting\n"));

  return numblocks;
}

int CSFSSD_writeDPBlock(csfssd_t *csfssd, u32 inode_idx,
			u32 dp_block_id, u8 *dp_block) 
{
  inode_t inode;
  u8 level, child;

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_writeDPBlock(%p,%d,%d,%p): Entering\n",
	  csfssd, inode_idx, dp_block_id, dp_block));


  if (CSFSSD_readInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeDPBlock: can't read inode.\n"));
    return -1;
  }

  /* Need to insert safety checks to make sure the DP's exist. */
  
  level = ((u8 *)dp_block_id)[0];

  if (level == 0) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeDPBlock: setting 0th level DP block\n"));


    if ((child = ((u8 *)dp_block_id)[1]) >=
	CSFSSD_NUM_SINDIRECT + CSFSSD_NUM_DINDIRECT) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_writeDPBlock: child OOB %d\n", child));
      return -1;
    }


    if (inode.block_ptr[child].dp == 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeDPBlock: null DP\n"));
      return -1;
    }

    if (writeBlock(csfssd->disk, 
		   inode.block_ptr[child].dp,
		   dp_block) < 0) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_writeDPBlock: writeBlock %d failed\n",
	      inode.block_ptr[child].dp));
      return -1;
    }

  } else if (level == 1) {
    csfssd_dp_block_t block;

    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeDPBlock: setting 1st level DP block\n"));

    child = ((u8 *)dp_block_id)[1];

    if ((child >= CSFSSD_NUM_SINDIRECT + CSFSSD_NUM_DINDIRECT) ||
	(child < CSFSSD_NUM_SINDIRECT)) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_writeDPBlock: child OOB %d\n", child));
      return -1;
    }


    if (inode.block_ptr[child].dp == 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeDPBlock: null DP\n"));
      return -1;
    }


    if (readBlock(csfssd->disk, 
		  inode.block_ptr[child].dp,
		  (u8 *)&block) < 0) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_writeDPBlock: readBlock %d failed\n",
	      inode.block_ptr[child].dp));
      return -1;
    }

    
    /* Now set the real DP block */
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeDPBlock: really setting 1st level DP block\n"));
    
    child = ((u8 *)dp_block_id)[2];
    
    if (child >= CSFSSD_BLOCK_SIZE/sizeof(csfssd_dp_t)) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_writeDPBlock: child OOB %d\n", child));
      return -1;
    }


    ASSERT(sizeof(csfssd_dp_t) == 32);
    ASSERT(sizeof(csfssd_dp_block_t) == CSFSSD_BLOCK_SIZE);


    if ((child = block.dp[child]) == 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeDPBlock: null DP\n"));
      return -1;
    }

    
    if (writeBlock(csfssd->disk, 
		   child,
		   dp_block) < 0) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_writeDPBlock: readBlock %d failed\n",
	      child));
      return -1;
    }
    

  } else {

    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeDPBlock: invalid level %d.\n", level));
    return -1;

  }


  return 0;
}


int CSFSSD_readDPBlock(csfssd_t *csfssd, u32 inode_idx,
		       u32 dp_block_id, u8 **dp_block)
{
  inode_t inode;
  u8 level, child;


  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_readDPBlock(%p,%d,%d,%p): Entering\n",
	  csfssd, inode_idx, dp_block_id, dp_block));


    if (CSFSSD_readInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_readDPBlock: can't read inode.\n"));
    return -1;
  }

  /* Need to insert safety checks to make sure the DP's exist. */
  
  level = ((u8 *)dp_block_id)[0];

  if (level == 0) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_readDPBlock: setting 0th level DP block\n"));


    if ((child = ((u8 *)dp_block_id)[1]) >=
	CSFSSD_NUM_SINDIRECT + CSFSSD_NUM_DINDIRECT) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_readDPBlock: child OOB %d\n", child));
      return -1;
    }


    if (inode.block_ptr[child].dp == 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_readDPBlock: null DP\n"));
      return -1;
    }

    *dp_block = (u8 *)malloc(sizeof(csfssd_dp_block_t));
					    
    if (readBlock(csfssd->disk, 
		  inode.block_ptr[child].dp,
		  *dp_block) < 0) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_readDPBlock: writeBlock %d failed\n",
	      inode.block_ptr[child].dp));
      free(dp_block);
      return -1;
    }

  } else if (level == 1) {
    csfssd_dp_block_t block;

    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_readDPBlock: setting 1st level DP block\n"));

    child = ((u8 *)dp_block_id)[1];

    if ((child >= CSFSSD_NUM_SINDIRECT + CSFSSD_NUM_DINDIRECT) ||
	(child < CSFSSD_NUM_SINDIRECT)) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_readDPBlock: child OOB %d\n", child));
      return -1;
    }


    if (inode.block_ptr[child].dp == 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_readDPBlock: null DP\n"));
      return -1;
    }


    if (readBlock(csfssd->disk, 
		  inode.block_ptr[child].dp,
		  (u8 *)&block) < 0) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_readDPBlock: readBlock %d failed\n",
	      inode.block_ptr[child].dp));
      return -1;
    }

    
    /* Now set the real DP block */
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_readDPBlock: really setting 1st level DP block\n"));
    
    child = ((u8 *)dp_block_id)[2];
    
    if (child >= CSFSSD_BLOCK_SIZE/sizeof(csfssd_dp_t)) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_readDPBlock: child OOB %d\n", child));
      return -1;
    }


    ASSERT(sizeof(csfssd_dp_t) == 32);
    ASSERT(sizeof(csfssd_dp_block_t) == CSFSSD_BLOCK_SIZE);


    if ((child = block.dp[child]) == 0) {
      DEBUGP(DEBUG_CONCISE, ("CSFSSD_readDPBlock: null DP\n"));
      return -1;
    }

    *dp_block = (u8 *)malloc(sizeof(csfssd_dp_block_t));
    
    if (readBlock(csfssd->disk, 
		  child,
		  *dp_block) < 0) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_readDPBlock: readBlock %d failed\n",
	      child));
      free(dp_block);
      return -1;
    }
    

  } else {

    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_readDPBlock: invalid level %d.\n", level));
    return -1;

  }


  return 0;
}
