/**
** Portions borrowed from XFS - X-File System
**
** Patrick McCormick
** Kevin Fu
** Joy Nicholson
**
** April 23, 1997
**
** $Id: freemap.c,v 1.11 1999/04/29 19:27:03 fubob Exp $
**
** $Log: freemap.c,v $
** Revision 1.11  1999/04/29 19:27:03  fubob
** i can successfully mkfs a disk:
**
** 	./mkfs 16 16 /tmp/disk
**
** and mount it:
**
** 	./csfssd /tmp/disk
**
** yeah. Zzzz
**
** Revision 1.10  1999/04/29 15:22:09  fubob
** freemap block allocation mode modulized
**
** bmap done for direct and singly indirect data pointers.  Not done
** for doubly indirect.  untested, and really ugly code with lots of
** #define'd macros. yum yum.
**
** Revision 1.9  1999/04/29 05:47:14  fubob
** i hate bmap.
**
** Revision 1.8  1999/04/28 23:43:44  fubob
** almost done with blockmapping [bmap()] routine and block allocation
** routines.  Painly computation of singly/doubly indirect pointers.
** yuck.
**
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "disk.h"
#include "csfssd.h"
#include "csfssd_prot.h"
#include "DEBUG.h"
#include "ASSERT.h"


inline int CSFSSD_sizeofFreemap_inbytes(u32 numblocks) {
  
  ASSERT(numblocks > 2);
  ASSERT((numblocks % 8) == 0);

  /* Num of bytes freemap actually takes up */
  return(numblocks / 8);
}

inline int CSFSSD_sizeofFreemap_inblocks(u32 numblocks) {
  
  ASSERT(numblocks > 2);
  ASSERT((numblocks % 8) == 0);
  
  /* Num of blocks freemap actually takes up */
  return((numblocks / (CSFSSD_BLOCK_SIZE*8)) +
	 ((numblocks % (CSFSSD_BLOCK_SIZE*8)) != 0));  

  /* last operand exists because for almost all cases, the freemap
     will not fill up a whole number of blocks. */
}


int CSFSSD_createFreemap(csfssd_t *csfssd) {
  u32 *blockarray, i;

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_createFreemap(%p): Entering\n", csfssd));
  
  ASSERT(csfssd != NULL);    
  ASSERT(csfssd->sb != NULL);    
  ASSERT(csfssd->freemap == NULL);    

  /* Sets entire disk as unused by zeroing memory */
  if ( (csfssd->freemap =
	calloc(1, CSFSSD_sizeofFreemap_inbytes(csfssd->sb->num_blocks))) 
       == NULL) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_createFreemap: malloc error.\n"));
    return -1;
  }

  blockarray=malloc(sizeof(u32)*csfssd->sb->first_data_block);
  
  /* This should allocate block 0-2 for instance */
  if (CSFSSD_allocBlocks(csfssd, 0,
			 csfssd->sb->first_data_block,
			 blockarray) < 0) {
    
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_createFreemap: allocBlocks failed\n"));
    return -1;
  }
  
  /* Verify we got the first blocks on the disk for the metadata */
  for (i = 0; i < csfssd->sb->first_data_block; i++) {
    ASSERT(blockarray[i] == i);
  }

  free(blockarray);

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_createFreemap: exiting\n"));

  return 0;
}


int CSFSSD_readFreemap(sb_t *sb, disk_t *disk, void **freemap) {
  int currblock, freemapsize_inbytes, offset_inbytes;
  char block[CSFSSD_BLOCK_SIZE];
  char *freeptr;
  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_readFreemap(%p,%p,%p): Entering\n", 
	  sb, disk, freemap));

  ASSERT(sb   != NULL);
  ASSERT(disk != NULL);

  freemapsize_inbytes = CSFSSD_sizeofFreemap_inbytes(sb->num_blocks);

  if ( (freeptr = (char*)malloc(freemapsize_inbytes)) == NULL) {
      DEBUGP(1, ("CSFSSD_readFreemap: malloc error.\n"));
      return -1;
  }

  /* Each block of the freemap on disk */
  for(currblock = sb->first_freemap_block; 
      currblock < sb->first_inode_block; 
      currblock++) {
    
    if ( readBlock(disk, currblock, block) < 0 ) {
      DEBUGP(DEBUG_CONCISE,
	     ("CSFSSD_readFreemap: readBlock failed block %d\n",
	      currblock));
      return -1;      
    }
    
    /* Copy the block to the in-mem freemap */
    offset_inbytes = CSFSSD_BLOCK_SIZE * (currblock 
					  - sb->first_freemap_block);

    /* We use MIN size the freemap does not necessary use up an
       entire block in memory */
    memcpy(freeptr + offset_inbytes,
	   block,
	   MIN(CSFSSD_BLOCK_SIZE, 
	       freemapsize_inbytes - offset_inbytes));
  }
  
  *freemap = freeptr;
  
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_readFreemap: Exiting.\n"));

  return freemapsize_inbytes;
}

int CSFSSD_writeFreemap(sb_t *sb, disk_t *disk, void *freemap) {
  char block[CSFSSD_BLOCK_SIZE];
  int currblock, offset_inbytes, freemapsize_inbytes;

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_writeFreemap(%p,%p,%p): Entering\n",
	  sb, disk, freemap));

  ASSERT(sb      != NULL);
  ASSERT(disk    != NULL);
  ASSERT(freemap != NULL);


  freemapsize_inbytes = CSFSSD_sizeofFreemap_inbytes(sb->num_blocks);

  /* Write freemap block by block to disk */
  for(currblock = sb->first_freemap_block; 
      currblock < sb->first_inode_block; 
      currblock++){
    
    /* Zero-out in case freemap does not use up the whole block */
    memset(block, 0, CSFSSD_BLOCK_SIZE);

    offset_inbytes = CSFSSD_BLOCK_SIZE * (currblock - 
					  sb->first_freemap_block);

    /* We use MIN size the freemap does not necessary use up an
       entire block in memory */
    memcpy(block,
	   (char *)freemap + offset_inbytes,
	   MIN(CSFSSD_BLOCK_SIZE,
	       freemapsize_inbytes - offset_inbytes));

    if (writeBlock(disk, currblock, block) < 0 ) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_writeFreemap: writeBlock failed block %d\n",
	      currblock));
      return -1;      
    }
  }
  
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeFreemap: Exiting\n"));

  return 0;
}


int CSFSSD_dcountFreeBlocks(sb_t *sb, disk_t *disk, u32 *num_free_blocks) {

  void *freemap;
  u32 freemapsize_inbytes, res;

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_dcountFreeBlocks(%p,%p,%p): Entering\n",
	  sb, disk, num_free_blocks));

  ASSERT(sb   != NULL);
  ASSERT(disk != NULL);
  ASSERT(num_free_blocks != NULL);

  freemapsize_inbytes = CSFSSD_sizeofFreemap_inbytes(sb->num_blocks);

  res = CSFSSD_readFreemap(sb, disk, &freemap);

  if (res != freemapsize_inbytes) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_dcountFreeBlocks: Conflict CSFSSD_sizeofFreemap_inbytes"));
    DEBUGP(DEBUG_CONCISE,
	   (" %lu CSFSSD_readFreemap %lu\n",freemapsize_inbytes, res));
    free(freemap);
    return -1;
  }
  
  if (CSFSSD_mcountFreeBlocks(sb, freemap, num_free_blocks) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_dcountFreeBlocks: mcount failed"));
    return -1;
  }
  
  free(freemap);
  
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_dcountFreeBlocks: Exiting.\n"));
  
  return 0;
}

int CSFSSD_mcountFreeBlocks(sb_t *sb, void *freemap, u32 *num_free_blocks) 
{
  u32 freeblock_cnt = 0;
  u8 *freebyte;
  u32 freemap_walk, freemapsize_inbytes, b;
  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_mcountFreeBlocks(%p,%p,%p): Entering\n",
	  sb, freemap, num_free_blocks));
  
  ASSERT(sb != NULL);
  ASSERT(freemap != NULL);
  ASSERT(num_free_blocks != NULL);
  
  freemapsize_inbytes = CSFSSD_sizeofFreemap_inbytes(sb->num_blocks);
  
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_mcountFreeBlocks: freemapsize = %lu bytes\n", 
	     freemapsize_inbytes));
  
  for (freemap_walk = 0; 
       freemap_walk < freemapsize_inbytes;
       freemap_walk++) {
    
    freebyte=(u8 *)freemap + freemap_walk;
    
    for (b=0; b<8; b++) {
      
      freeblock_cnt += (( ((*freebyte)>>b) &1) == 1);
      
    }
    
  }
  
  *num_free_blocks = sb->num_blocks - freeblock_cnt;
  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_mcountFreeBlocks: Exiting. Free blocks according to"));
  DEBUGP(DEBUG_CONCISE,
	 (" disk %lu\n", *num_free_blocks));

  return 0;
}


int CSFSSD_allocBlocks(csfssd_t *csfssd, u32 nearblock, u16 numblocks,
			  u32 *blockarray) {
  u32 freemap_walk;
  u32 allocated_cnt = 0;
  u8 *freebyte; 
  int i;
  int bit_cnt = 0;

  DEBUGP(DEBUG_VERBOSE, 
	 ("CSFSSD_allocateBlocks(%p,%d,%d,%p): Entering\n", 
	  csfssd, nearblock, numblocks, blockarray));

  DEBUGP(DEBUG_VERBOSE, ("CSFSSD_allocateBlocks: sb->num_free_blocks %lu\n",
			 csfssd->sb->num_free_blocks));

  ASSERT(csfssd     != NULL);
  ASSERT(csfssd->sb != NULL);
  ASSERT(blockarray != NULL);

  if (numblocks > csfssd->sb->num_free_blocks) {
    /* Not enough blocks left */
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_allocateBlocks: Not enough blocks left.\n"));
    return -1;
  }

  nearblock = nearblock % csfssd->sb->num_blocks;
  /* Ensure 0 <= nearblock <= sb->num_blocks 
     Closure?  But I just met her! */

  freemap_walk = nearblock/8; /* Begin looking at this byte in the freemap */

  /* Confusing point: We must make sure not to use the extra bytes
     (if any) in blocks.  eg, the former if the number of blocks is not a
     multiple of CSFSSD_BLOCK_SIZE.
     */

  while (allocated_cnt < numblocks) {
    /* Load one byte of the freemap at a time, looking for free space */
    freebyte = (u8*)csfssd->freemap + freemap_walk;
    
    /* Check each bit for freeness */
    for (i = 0; 
	 (i < 8) && (allocated_cnt < numblocks) && 
	   (bit_cnt < csfssd->sb->num_blocks);
	 i++) {

      if (((*freebyte>>i) & 1) == 0) {

	/* We found a free block, mark as used */

	*freebyte = *freebyte | (1<<i);	

	blockarray[allocated_cnt] = freemap_walk*8 + i; 
	/* The represented block num */

	allocated_cnt++;
      }
      bit_cnt++;  /* Checks that we don't start checking bits twice */

    }
    freemap_walk++; /* It's ok to overflow */

    /* Catch going over maximum, not a u32 overflow check */
    freemap_walk = freemap_walk % (csfssd->sb->num_blocks/8);

    if (bit_cnt > csfssd->sb->num_blocks) {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_allocateBlocks: We exhausted freemap. numblocks "));
      DEBUGP(DEBUG_CONCISE,
	     ("%lu; freeblocks %lu; Boom.\n", 
	      numblocks, csfssd->sb->num_free_blocks));
      return -1;
    }
    
  }

  csfssd->sb->num_free_blocks -= allocated_cnt;

  if (CSFSSD_writeFreemap(csfssd->sb, csfssd->disk, csfssd->freemap) < 0)
    {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_allocateBlocks: writeFreemap failed\n"));
      return -1;
    }


  if (CSFSSD_writeSB(csfssd->sb, csfssd->disk) < 0)
    {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_allocateBlocks: writeSB failed\n"));
      return -1;
    }

  DEBUGP(DEBUG_VERBOSE, 
	 ("CSFSSD_allocateBlocks: Exiting. Blocks left %lu\n",
	  csfssd->sb->num_free_blocks));
  return 0;
}


int CSFSSD_deallocBlocks(csfssd_t *csfssd, u16 numblocks, 
			 u32 *blockarray) {
  int currblock, offset_inbits;
  u8 *freebyte;

  DEBUGP(DEBUG_VERBOSE, ("CSFSSD_freeBlocks(%p,%d,%p): Entering\n"));

  ASSERT(csfssd     != NULL);
  ASSERT(csfssd->sb != NULL);
  ASSERT(blockarray != NULL);

  /* Sanity check that we aren't freeing more blocks than we have allocated */
  if (numblocks > csfssd->sb->num_blocks - csfssd->sb->num_free_blocks) {
      DEBUGP(DEBUG_CONCISE,
	     ("CSFSSD_freeBlocks: Freeing more than possible.  Blocks "));
      DEBUGP(DEBUG_CONCISE,
	     ("used %d Want to free %d.\n",
	      csfssd->sb->num_blocks - csfssd->sb->num_free_blocks,
	      numblocks));
      return -1;
  }

  for (currblock = 0; currblock < numblocks; currblock++) {

    /* Which byte in the freemap */
    freebyte = csfssd->freemap + (blockarray[currblock]/8); 

    /* Offset within a byte of the freemap */
    offset_inbits = blockarray[currblock] % 8;
    
    /* Check that this bit is actually marked as used */
    if ( (((*freebyte) >> offset_inbits) & 1) == 0) {
      DEBUGP(DEBUG_CONCISE,
	     ("CSFSSD_allocateBlocks: Warning.  Freeing unused "));
      DEBUGP(DEBUG_CONCISE,
	     ("bit. block %d freebyte %d offset %d.\n",
	      currblock, freebyte, offset_inbits));
      return -1;
    }
    
    /* Mark the bit free! */
    *freebyte=(*freebyte) ^ (1 << offset_inbits);
  }

  csfssd->sb->num_free_blocks += numblocks;

  if (CSFSSD_writeFreemap(csfssd->sb, csfssd->disk, csfssd->freemap) < 0)
    {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_deallocBlocks: writeFreemap failed\n"));
      return -1;
    }


  if (CSFSSD_writeSB(csfssd->sb, csfssd->disk) < 0)
    {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_deallocBlocks: writeSB failed\n"));
      return -1;
    }



  DEBUGP(DEBUG_VERBOSE,
	 ("CSFSSD_freeBlocks: Exiting.  Blocks left %d\n",
	  csfssd->sb->num_free_blocks));
  return 0;
}

