/**
** Portions borrowed from CSFSSD - X-File System
**
** Patrick McCormick
** Kevin Fu
** Joy Nicholson
**
** April 23, 1997
**
** $Id: inode.c,v 1.16 1999/04/27 03:36:06 fubob Exp $
**
** $Log: inode.c,v $
** Revision 1.16  1999/04/27 03:36:06  fubob
** fixed lots of little typos
**
** Revision 1.15  1999/04/26 16:14:23  fubob
** fixed silly mistakes (printf problems), fixed inode size to 512 bytes.
**
** Revision 1.14  1999/04/24 17:28:00  fubob
** fixed more of inode.c, finished superblock routines in sb.c.  starting
** on freemap.
**
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>

#include "disk.h"
#include "csfssd.h"
#include "crypto.h"
#include "libdebug/DEBUG.h"
#include "libdebug/ASSERT.h"

int CSFSSD_resetInode(inode_t *inode, u32 inode_idx) {

  ASSERT(inode != NULL);

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_resetInode(%p,%d): Entering.\n", 
	  inode, inode_idx));

  /* This operation takes care of zeroizing all fields.
     below we only set fields which have non-zero values */
  memset (inode, 0, sizeof(inode_t));

  memset (inode->magic, CSFSSD_INODE_MAGIC, 51);
  inode->atime = inode->mtime = inode->ctime = CSFSSD_get_time();
 
  /* Note, the root inode is not encrypted, and by default has no
     integrity protection.  This is signified by having all keys == 0.
     The liklihood of the lockboxes ==0 for real keys is too small to
     consider */

  inode->inode_idx = inode_idx;
  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_resetInode: Exiting.\n"));
  
  return 0;
}


int CSFSSD_lockFile(csfssd_t *csfssd, u32 inode_idx, u16 uid, 
		    u16 *duration)
{
  u32 now;
  inode_t inode;

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_lockFile: entering\n"));

  if (CSFSSD_readInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_lockFile: can't read inode\n"));
    return -1;
  }

  now = CSFSSD_get_time();

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_lockFile: got time\n"));


  /* If file is not locked, lock it */
  /* need to check if UID is owner or GID group!!! */
  if((inode.locked_duration == 0) ||
     (inode.locked_at + inode.locked_duration < now))
    {

      DEBUGP(DEBUG_CONCISE,("CSFSSD_lockFile: getting lock\n"));
      inode.locked_duration = *duration;
      inode.locked_by = uid;
      inode.locked_at = now;
      DEBUGP(DEBUG_CONCISE,("CSFSSD_lockFile: got virgin lock\n"));


      if (CSFSSD_writeInode(csfssd, &inode, inode_idx) < 0) {
	DEBUGP(DEBUG_CONCISE, ("CSFSSD_lockFile: can't write inode\n"));
	return -1;
      }
      
    } else if ((inode.uid == uid) &&
	       (inode.locked_at + *duration - now
		< CSFSSD_MAX_LEASE_SECONDS)){
      
      /* If UID already locked it, relock it if we 
	 haven't gone past the max lease period */
      
      
      inode.locked_duration = *duration;
      
      if (CSFSSD_writeInode(csfssd, &inode, inode_idx) < 0) {
	DEBUGP(DEBUG_CONCISE, ("CSFSSD_lockFile: can't write inode\n"));
	return -1;
      }
    
    } else {
      *duration = inode.locked_at + inode.locked_duration - now;
      DEBUGP(DEBUG_CONCISE,
	     ("CSFSSD_lockFile: already locked for %d seconds\n", *duration));
      return -1;
    }
  
  DEBUGP(DEBUG_CONCISE, ("CSFSSD_lockFile: exiting\n"));
  return 0;
}

int CSFSSD_unlockFile(csfssd_t *csfssd, u32 inode_idx, u16 uid)
{
  inode_t inode;

  DEBUGP(DEBUG_CONCISE,("CSFSSD_unlockFile: entering\n"));

  if (CSFSSD_readInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_unlockFile: can't read inode\n"));
    return -1;
  }

  if(inode.locked_by != uid) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_unlockFile: not lock owner\n"));
    return -1;
  }

  inode.locked_duration = 0;
  inode.locked_by = 0;
  inode.locked_at = 0;

  if (CSFSSD_writeInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_unlockFile: can't write inode\n"));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE,("CSFSSD_unlockFile: exiting\n"));
  return 0;
}

int CSFSSD_allocInode(csfssd_t *csfssd, u32 *new_inode_idx, 
		      u32 dir_inode_idx) {
  inode_t an_inode_block[CSFSSD_INODES_PER_BLOCK];
  u32 dir_block, num_inode_blocks, i, blocknum =0;
  int res, j=0;
  bool DONE;

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_allocInode(%p,%p,%d): Entering.\n", 
	  csfssd, new_inode_idx, dir_inode_idx));

  ASSERT(csfssd          != NULL);
  ASSERT(csfssd->sb      != NULL);
  ASSERT(csfssd->disk    != NULL);
  ASSERT(csfssd->freemap != NULL);
  ASSERT(new_inode_idx   != NULL);
  ASSERT(sizeof(inode_t) == CSFSSD_INODE_SIZE);

  if (csfssd->sb->num_free_inodes == 0) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_allocInode: No inodes left\n"));
    return -1;
  }
    
  /* Where does the dir's block live w.r.t beginning of inode table */
  dir_block = dir_inode_idx / CSFSSD_INODE_SIZE;

  num_inode_blocks = csfssd->sb->num_inodes/CSFSSD_INODES_PER_BLOCK;

  /* Find an unused inode by searching near directory inode */
  DONE = FALSE;
  for (i=0; i < num_inode_blocks ; i++) {
    if (DONE) break;

    /* Loop through inode blocks in order, wrapping around  */
    blocknum = 
      csfssd->sb->first_inode_block +
      ((dir_block + i) % num_inode_blocks);

    /* read an inode block from disk into an_inode_block */
    if ((res = readBlock(csfssd->disk, blocknum, (char*)an_inode_block)) < 0)
      {
	DEBUGP(DEBUG_CONCISE, 
	       ("CSFSSD_allocInode: Yuck; readBlock returns %d.\n",
		res));
	return -1;
      }


    /* Look for a free inode */
    for (j=0; j < CSFSSD_INODES_PER_BLOCK; j++) {

      if (an_inode_block[j].link_count == 0) {
	*new_inode_idx = (i * CSFSSD_INODES_PER_BLOCK) + j;
	DONE = TRUE;
	break;
      }
    }
  }
  
  if (DONE != TRUE) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_allocInode: No more inodes left. Inode the feeling.\n"));
    return -1;
  }

  if (CSFSSD_resetInode (&(an_inode_block[j]), *new_inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_allocInode: can't reset inode\n"));
    return -1;
  }

  an_inode_block[j].link_count = 1;
  if ((res = writeBlock(csfssd->disk, blocknum, (char*)an_inode_block)) < 0)
    {
      DEBUGP(DEBUG_CONCISE, 
	     ("CSFSSD_allocInode: Yuck; writeBlock returns %d.\n",
	      res));
      return -1;
    }
  
  csfssd->sb->num_free_inodes--;

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_allocInode: Exiting.  Allocated inode %d\n",
	  *new_inode_idx));

  return 0;
}

int CSFSSD_deallocInode(csfssd_t *csfssd, u32 inode_idx) {
  inode_t inode;

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_deallocInode(%p,%d): Entering.\n",
	  csfssd, inode_idx));

  ASSERT(csfssd          != NULL);
  ASSERT(csfssd->sb      != NULL);
  ASSERT(csfssd->disk    != NULL);
  ASSERT(csfssd->freemap != NULL);

  if (CSFSSD_readInode(csfssd, &inode, inode_idx) < 0 ) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_deallocInode: Can't read inode\n"));
    return -1;
  } 

  if (inode.link_count > 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_deallocInode: link_count>0\n"));
    return -1;
  }

  if (CSFSSD_resetInode(&inode, inode_idx) < 0 ) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_deallocInode: Can't reset inode\n"));
    return -1;
  } 

  if (CSFSSD_writeInode(csfssd, &inode, inode_idx) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_deallocInode: Can't write inode\n"));
    return -1;
  } 

  csfssd->sb->num_free_inodes++;

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_deallocInode: Exiting.\n"));
  return 0;
}



int CSFSSD_writeInode(csfssd_t *csfssd, inode_t *inode, u32 inode_idx) {
  inode_t inode_block[CSFSSD_INODES_PER_BLOCK];
  u32 block_num;
  int res;

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_writeInode(%p,%p,%d): Entering\n",
	  csfssd, inode, inode_idx));

  ASSERT(csfssd          != NULL);
  ASSERT(csfssd->sb      != NULL);
  ASSERT(csfssd->disk    != NULL);
  ASSERT(csfssd->freemap != NULL);
  ASSERT(inode           != NULL);

  block_num = csfssd->sb->first_inode_block;
  block_num += inode_idx/CSFSSD_INODE_SIZE;  /* w.r.t. inode table */

  if (block_num > csfssd->sb->first_data_block) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeInode: Inode block out of bounds at"));
    DEBUGP(DEBUG_CONCISE, ("%lu, first inode at %lu.\n", block_num, 
			   csfssd->sb->first_inode_block));
    return -1;
  }

  if ((res = readBlock(csfssd->disk, block_num, (u8 *)inode_block)) < 0) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeInode: readBlock returns %d.\n", res));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_writeInode: writing inode in inode block %u\n",
	  block_num));

  memcpy((void*)&(inode_block[inode_idx % CSFSSD_INODES_PER_BLOCK]),
	 (void*)inode,
	 sizeof(inode_t));

  if ((res = writeBlock(csfssd->disk, block_num, (u8 *)inode_block)) < 0) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_writeInode: CSFSSD_writeBlock returns %d.\n", res));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeInode: Exiting.\n"));
  return 0;
}


int CSFSSD_readInode(csfssd_t *csfssd, inode_t *inode, u32 inode_idx) {
  inode_t inode_block[CSFSSD_INODES_PER_BLOCK];
  u32 block_num;
  int res;

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_readInode(%p,%p,%d): Entering\n",
			 csfssd, inode, inode_idx));

  ASSERT((CSFSSD_BLOCK_SIZE % sizeof(inode_t)) == 0);
  ASSERT(csfssd          != NULL);
  ASSERT(csfssd->sb      != NULL);
  ASSERT(csfssd->disk    != NULL);
  ASSERT(csfssd->freemap != NULL);
  ASSERT(inode           != NULL);

  block_num = csfssd->sb->first_inode_block;
  block_num += inode_idx/CSFSSD_INODES_PER_BLOCK;  /* Offset into inode table */

  if (block_num > csfssd->sb->first_data_block) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_readInode: Inode block out of bounds %d.\n", 
	    block_num));
    return -1;
  }

  if ((res = readBlock(csfssd->disk, block_num, (u8 *)inode_block)) < 0) {
    DEBUGP(DEBUG_CONCISE, 
	   ("CSFSSD_readInode: Yuck CSFSSD_readBlock returns %d.\n", res));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_readInode: reading from inode block %u\n", block_num));

  memcpy((void*)inode, 
	 (void*)&inode_block[inode_idx % CSFSSD_INODES_PER_BLOCK], 
	 sizeof(inode_t));

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_readInode: Exiting.\n"));
  return 0;
}


#ifdef 0
int CSFSSD_offset2dp(csfssd_t *csfssd, inode_t *file, u32 offset, 
		     csfssd_dp_t **dptr)
{
  int block_num;

  DEBUGP(DEBUG_CONCISE,("CSFSSD_offset2dp: entering\n"));
  
  block_num = offset/CSFSSD_BLOCK_SIZE;
  /*  *dptr = malloc(sizeof(csfssd_dp_t)); */
  if(block_num < CSFSSD_NUM_DIRECT)
    {
      *dptr = &(file->block_ptr[block_num]);
      return 0;
    }
  block_num -= CSFSSD_NUM_DIRECT;
  if (block_num < 
      CSFSSD_NUM_SINDIRECT*(CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE))
    {
      csfssd_dp_t dp;
      csfssd_dp_t indir_block[CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE];
      int sindirect_block_num = block_num 
	/ (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
      block_num = block_num % (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);

      dp = file->block_ptr[CSFSSD_NUM_DIRECT + sindirect_block_num];

      if ((readBlock(csfssd->disk, dp.dp, indir_block)) < 0) {
	return -1;
      }
      *dptr = &(indir_block[block_num]);
      return 0;
    }
  block_num -= CSFSSD_NUM_SINDIRECT*(CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
  if (block_num <
      CSFSSD_NUM_DINDIRECT*((CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE)^2))
    {
      int dindirect_block_num; /* which dindirect_block_num to look at */
      int dindirect_block_offset; /* where to find ptr to sindirect block */
      int sindirect_block_offset; /* where to find ptr to actual block */
      csfssd_dp_t dp;
      csfssd_dp_t indir_block[CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE];

      /* which dindirect block? */
      /* calculate offset */
      dindirect_block_num = block_num 
	/ ((CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE)^2);
      dp = file->block_ptr[CSFSSD_NUM_DIRECT
			  + CSFSSD_NUM_SINDIRECT
			  + dindirect_block_num];

      /* read in the dindirect block */
      if((readBlock(csfssd->disk, dp.dp, indir_block)) < 0) {
	return -1;
      }

      /* which sindirect block? */
      /* calculate offset */
      dindirect_block_offset = 
	(block_num / (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE)) 
	% (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
      dp = indir_block[dindirect_block_offset];

      /* read in sindirect block */
      if((readBlock(csfssd->disk, dp.dp, indir_block)) < 0) {
	return -1;
      }

      /* which block? */
      /* calculate offset */
      sindirect_block_offset = block_num 
	% (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
      dp = indir_block[sindirect_block_offset];

      /* get the IV for the actual block */
      *dptr = &(dp);
      return 0;
    }

  DEBUGP(DEBUG_CONCISE,("CSFSSD_offset2dp: exiting\n"));
  return -1;
}


/* sets block_offset to -1 if data is in the inode itself */
int CSFSSD_offset2block(csfssd_t *csfssd, inode_t *file, u32 offset,
			u32 *block_ptr, u32 *block_offset)
{
  int block_num;

  DEBUGP(DEBUG_CONCISE,("CSFSSD_offset2block: entering\n"));
  
  block_num = offset/CSFSSD_BLOCK_SIZE;
  /*  *dptr = malloc(sizeof(csfssd_dp_t)); */
  if(block_num < CSFSSD_NUM_DIRECT)
    {
      /* *block_ptr = (file->block_ptr[block_num]).dp;*/
      *block_offset = -1;/*block_num;*/
      return 0;
    }
  block_num -= CSFSSD_NUM_DIRECT;
  if (block_num <
      CSFSSD_NUM_SINDIRECT*(CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE))
    {
      csfssd_dp_t dp;
      csfssd_dp_t indir_block[CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE];
      int sindirect_block_num = block_num 
	/ (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
      block_num = block_num % (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);

      dp = file->block_ptr[CSFSSD_NUM_DIRECT + sindirect_block_num];

      if ((readBlock(csfssd->disk, dp.dp, indir_block)) < 0) {
	return -1;
      }
      *block_ptr = dp.dp;/*&(indir_block[block_num]);*/
      *block_offset = block_num;
      return 0;
    }
  block_num -= CSFSSD_NUM_SINDIRECT*(CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
  if (block_num < 
      CSFSSD_NUM_DINDIRECT*((CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE)^2))
    {
      int dindirect_block_num; /* which dindirect_block_num to look at */
      int dindirect_block_offset; /* where to find ptr to sindirect block */
      int sindirect_block_offset; /* where to find ptr to actual block */
      csfssd_dp_t dp;
      csfssd_dp_t indir_block[CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE];

      /* which dindirect block? */
      /* calculate offset */
      dindirect_block_num = block_num
	/ ((CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE)^2);
      dp = file->block_ptr[CSFSSD_NUM_DIRECT 
			  + CSFSSD_NUM_SINDIRECT
			  + dindirect_block_num];

      /* read in the dindirect block */
      if((readBlock(csfssd->disk, dp.dp, indir_block)) < 0) {
	return -1;
      }

      /* which sindirect block? */
      /* calculate offset */
      dindirect_block_offset =
	(block_num / (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE)) 
	% (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
      dp = indir_block[dindirect_block_offset];

      /* read in sindirect block */
      if((readBlock(csfssd->disk, dp.dp, indir_block)) < 0) {
	return -1;
      }

      /* which block? */
      /* calculate offset */
      sindirect_block_offset = block_num
	% (CSFSSD_BLOCK_SIZE/CSFSSD_BLOCK_PTR_SIZE);
      
      *block_ptr = dp.dp;
      *block_offset = sindirect_block_offset;
      /*
      dp = indir_block[sindirect_block_offset];
      */
      /* get the IV for the actual block */
      /**dptr = &(dp);*/
      return 0;
    }

  DEBUGP(DEBUG_CONCISE,("CSFSSD_offset2block: exiting\n"));
  return -1;
}

#endif
