/**
** Superblock routines
**
** Danger Will Robinson! Portions borrowed from CSFSSD - X-File System
**
** Patrick McCormick
** Kevin Fu
** Joy Nicholson
**
** April 23, 1997
**
**
** $Id: sb.c,v 1.10 1999/04/29 19:27:03 fubob Exp $
**
** $Log: sb.c,v $
** Revision 1.10  1999/04/29 19:27:03  fubob
** i can successfully mkfs a disk:
**
** 	./mkfs 16 16 /tmp/disk
**
** and mount it:
**
** 	./csfssd /tmp/disk
**
** yeah. Zzzz
**
** Revision 1.9  1999/04/28 23:43:44  fubob
** almost done with blockmapping [bmap()] routine and block allocation
** routines.  Painly computation of singly/doubly indirect pointers.
** yuck.
**
** Revision 1.8  1999/04/28 14:50:51  fubob
** wow, I had a major bug in file.c.  Not done fixing, but almost.  Forgot
** entirely the data block allocation routines.
**
** Revision 1.7  1999/04/27 03:36:06  fubob
** fixed lots of little typos
**
** Revision 1.6  1999/04/26 16:14:23  fubob
** fixed silly mistakes (printf problems), fixed inode size to 512 bytes.
**
** Revision 1.5  1999/04/24 17:28:00  fubob
** fixed more of inode.c, finished superblock routines in sb.c.  starting
** on freemap.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "csfssd.h"
#include "csfssd_prot.h"
#include "disk.h"
#include "DEBUG.h"
#include "ASSERT.h"
#include "misc/types.h"


int CSFSSD_createSB(sb_t *sb, u32 numblocks, u32 numinodes) 
{
  int freemap_size_inblocks, inodes_size_inblocks;

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_createSB(%p,%d,%d): Entering\n",
	  sb, numblocks, numinodes));

  ASSERT(sb              != NULL);
  ASSERT(sizeof(inode_t) == CSFSSD_INODE_SIZE);
  ASSERT((numblocks % 8) == 0);    /* Ensure freemap bits are all used */

  /* Ensure inode table all used */
  ASSERT((numinodes % CSFSSD_INODES_PER_BLOCK) == 0);

  /* # blocks the freemap uses */
  freemap_size_inblocks = CSFSSD_sizeofFreemap_inblocks(numblocks);

  /* # blocks for inode table */
  inodes_size_inblocks = numinodes/CSFSSD_INODES_PER_BLOCK;

  memset(&sb->magic, CSFSSD_SUPER_MAGIC, 4);
  sb->num_blocks = numblocks;
  sb->num_inodes = numinodes;
  sb->first_freemap_block = CSFSSD_SUPER_BLOCK + 1;
  sb->first_inode_block = sb->first_freemap_block + freemap_size_inblocks;   
  sb->first_data_block= sb->first_inode_block + inodes_size_inblocks;
  sb->mount_count = 0;
  sb->dirty = FALSE;
  sb->lastchecked_time = CSFSSD_get_time();
  sb->num_free_blocks = numblocks;
  sb->num_free_inodes = numinodes;

  DEBUGP(DEBUG_CONCISE,
	 ("%lu 1st_inode_block %lu 1st_data_block %lu ",
	  numinodes, sb->first_inode_block, sb->first_data_block));
  DEBUGP(DEBUG_CONCISE, 
	 ("1st_freemap_block = %lu.\n",sb->first_freemap_block));

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_createSB: Exiting.\n"));
  return 0;
}

int CSFSSD_writeSB(sb_t *sb, disk_t *disk) {
  sb_disk_t sb_d; 
  int res;
  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_writeSB(%p,%p): Entering\n", sb, disk));

  ASSERT(disk != NULL);
  ASSERT(sb   != NULL);
  ASSERT(CSFSSD_BLOCK_SIZE == sizeof(sb_disk_t));


  /* Copy in-memory SB -> disk SB struct */
  memcpy(&sb_d.sb, sb, sizeof(sb_t)); 

  memset(&sb_d.magic, CSFSSD_SUPER_MAGIC, CSFSSD_SUPER_PADDING); 

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_writeSB: sb_d num_blocks %d, num_inodes %d\n", 
	 sb_d.sb.num_blocks, sb_d.sb.num_inodes));

  if ((res=writeBlock (disk, CSFSSD_SUPER_BLOCK, (u8 *)&sb_d)) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_writeSB: writeBlock error %d\n",res));
    return -1;
  }
  
  syncDisk(disk);

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_writeSB: Exiting\n"));

  return 0;
}

int CSFSSD_readSB(sb_t *sb, disk_t *disk) {
  sb_disk_t sb_d;
  int res;
  
  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_readSB(%p,%p): Entering\n", sb, disk));

  ASSERT(disk != NULL);
  ASSERT(sb   != NULL);
  ASSERT(CSFSSD_BLOCK_SIZE == sizeof(sb_disk_t));
  
  if ((res = readBlock(disk, CSFSSD_SUPER_BLOCK, (char *)&sb_d)) < 0) {
    DEBUGP(DEBUG_CONCISE,
	   ("CSFSSD_readSB: readBlock error %d\n",res));
    return -1;
  }

  DEBUGP(DEBUG_CONCISE,
	 ("CSFSSD_readSB: verify sb_d num_blocks %d, num_inodes %d\n", 
	  sb_d.sb.num_blocks, sb_d.sb.num_inodes));
  

  
  /* Copy in-memory SB -> disk SB struct */
  memcpy(sb, &(sb_d.sb), sizeof(sb_t)); 

  DEBUGP(DEBUG_CONCISE, 
	 ("CSFSSD_readSB: num_blocks=%lu num_inodes=%lu ",
	  sb->num_blocks, sb->num_inodes));
  DEBUGP(DEBUG_CONCISE,
	 ("1st_freemap_block=%lu 1st_inode_block=%lu 1st_data_block=%lu ",
	  sb->first_freemap_block,sb->first_inode_block,sb->first_data_block));

  DEBUGP(DEBUG_CONCISE,
	 ("num_free_blocks=%lu num_free_inodes=%lu\n",
	  sb->num_free_blocks, sb->num_free_inodes));

  DEBUGP(DEBUG_CONCISE, ("CSFSSD_readSB: Exiting\n"));
  return 0;
}


int CSFSSD_fsck(sb_t *sb, disk_t *disk) {
  
  /* so silly */
  sb->dirty = FALSE;
  sb->lastchecked_time = CSFSSD_get_time();

  /* Need to fix other fields and stuff in superblock */

  if (CSFSSD_writeSB(sb, disk) < 0) {
    DEBUGP(DEBUG_CONCISE, ("CSFSSD_fsck: writeSB failed\n"));
    return -1;
  }

  return 0;
}
