/* $Id: callout.c,v 1.1 1998/10/21 18:32:52 fubobb Exp $ */

#include <assert.h>
#include <sys/types.h>
#include <sys/time.h>

#include "private.h"
#include "redblack.h"

/* These might get defined somewhere */
#undef timespecclear
#undef timespeccmp
#undef timespecadd
#undef timespecsub

#ifndef HAVE_CLOCK_GETTIME
int
clock_gettime (int id, struct timespec *tp)
{
  struct timeval tv;
  gettimeofday (&tv, NULL);
  tp->tv_sec = tv.tv_sec;
  tp->tv_nsec = tv.tv_usec * 1000;
  return 0;
}
#endif /* !HAVE_CLOCK_GETTIME */

struct callout {
  union {
    struct callout *co_u_next;
    struct redblack_entry co_u_node;
  } co_u;
  struct timespec co_tv;
  void (*co_fn) (void *);
  void *co_arg;
};
#define co_next co_u.co_u_next
#define co_node co_u.co_u_node

static struct callout *cofree;

redblack_decl (cotree, callout, co_node, callout_cmp);

static struct cotree coroot;

static inline void
timespecclear (struct timespec *tvp)
{
  tvp->tv_sec = 0;
  tvp->tv_nsec = 0;
}

static inline int
timespeccmp (const struct timespec *a, const struct timespec *b)
{
  if (a->tv_sec == b->tv_sec)
    return (int) (a->tv_nsec - b->tv_nsec);
  return a->tv_sec > b->tv_sec ? 1 : -1;
}

static inline void
timespecsub (struct timespec *r, const struct timespec *a,
	    const struct timespec *b)
{
  r->tv_sec = a->tv_sec - b->tv_sec;
  r->tv_nsec = a->tv_nsec - b->tv_nsec;
  if (r->tv_nsec < 0) {
    r->tv_sec--;
    r->tv_nsec += 1000000000;
  }
}

static inline void
timespecadd (struct timespec *r, const struct timespec *a,
	    const struct timespec *b)
{
  r->tv_sec = a->tv_sec + b->tv_sec;
  r->tv_nsec = a->tv_nsec + b->tv_nsec;
  if (r->tv_nsec > 1000000000) {
    r->tv_sec++;
    r->tv_nsec -= 1000000000;
  }
}

int
callout_cmp (struct callout *a, struct callout *b)
{
  return timespeccmp (&a->co_tv, &b->co_tv);
}

struct callout *
callout_add (struct timespec *tvp, void (*cb) (void *), void *arg)
{
  struct callout *co;
  if ((co = cofree))
    cofree = co->co_next;
  else {
    assert (!(thread_current->t_ipl & SPL_MEM));
    co = xmalloc (sizeof (*co));
  }
  if (co) {
    co->co_tv = *tvp;
    co->co_fn = cb;
    co->co_arg = arg;
    cotree_insert (&coroot, co);
  }
  return co;
}

void
callout_free (struct callout *co)
{
  cotree_delete (&coroot, co);
  co->co_next = cofree;
  cofree = co;
#if 0
  xfree (co);
#endif
}

/* It's often convenient to add a callout at splhigh.  This call can
 * be made first, at splsig, to ensure that xmalloc doesn't fail
 * within callout_add.  */
void
callout_reserve (void)
{
  assert (thread_current->t_ipl & SPL_SIG);
  assert (!(thread_current->t_ipl & SPL_MEM));
  if (!cofree) {
    struct callout *co = xmalloc (sizeof (*co));
    co->co_next = NULL;
    cofree = co;
  }
}

int
callout_check (struct timespec *time_to_next)
{
  struct callout *co, *nco;
  struct timespec tv;
  int didcb;

  do {
    didcb = 0;
    clock_gettime (CLOCK_REALTIME, &tv);
    for (co = cotree_first (&coroot);
	 co && timespeccmp (&tv, &co->co_tv) >= 0;
	 co = nco) {
      didcb = 1;
      nco = cotree_next (co);
      cotree_delete (&coroot, co);
      co->co_fn (co->co_arg);
      xfree (co);
    }
  } while (didcb && co);

  if (co) {
    if (time_to_next)
      timespecsub (time_to_next, &co->co_tv, &tv);
    return 1;
  }
  return 0;
}

int
callout_next (struct timespec *time_to_next)
{
  struct callout *co = cotree_first (&coroot);
  struct timespec tv;

  clock_gettime (CLOCK_REALTIME, &tv);
  if (co) {
    if (time_to_next)
      timespecsub (time_to_next, &co->co_tv, &tv);
    return 1;
  }
  return 0;
}

static void
thread_wake_handler (void *_t)
{
  tcb *t = _t;
  t->t_wake = NULL;
  thread_awaken (t);
}

void
thread_waketime (struct timespec *tvp)
{
  int s = splsig ();

  assert (!thread_current->t_wake);
  thread_current->t_wake = callout_add (tvp, thread_wake_handler,
					thread_current);
  splx (s);
}

terr
thread_sleeptil (struct timespec *tvp)
{
  terr ret;
  int s = splsig ();

  thread_enq (&thread_sleepq, thread_current);
  thread_waketime (tvp);
  ret = thread_yield ();

  splx (s);
  return ret;
}

terr
thread_nanosleep (struct timespec *tvp, struct timespec *rtvp)
{
  struct timespec now, tv;
  terr ret;

  if (!tvp) {
    assert (!rtvp);
    thread_enq (&thread_sleepq, thread_current);
    return thread_yield ();
  }

  clock_gettime (CLOCK_REALTIME, &now);
  timespecadd (&tv, &now, tvp);
  ret = thread_sleeptil (&tv);

  if (rtvp) {
    clock_gettime (CLOCK_REALTIME, &now);
    if (timespeccmp (&tv, &now) > 0)
      timespecsub (rtvp, &tv, &now);
    else
      timespecclear (rtvp);
  }

  return ret;
}

static void
thread_alarm_handler (void *_t)
{
  tcb *t = _t;
  t->t_alarm = NULL;
  t->t_alerted = 1;
  thread_awaken (t);
}

int
thread_alarm (int sec)
{
  int ret = 0;
  struct timespec now;
  struct timespec tv;
  struct callout *co;
  int s;

  assert (sec >= 0);
  s = splsig ();

  clock_gettime (CLOCK_REALTIME, &now);

  if ((co = thread_current->t_alarm)) {
    timespecsub (&tv, &co->co_tv, &now);
    ret = tv.tv_sec + (tv.tv_nsec >= 500000000);
    /* Don't let an alarm fall through the cracks, the caller might
     * need to know he is deleting an alert. */
    if (ret <= 0)
      ret = 1;
    callout_free (co);
    thread_current->t_alarm = NULL;
  }

  if (sec) {
    now.tv_sec += sec;
    thread_current->t_alarm
      = callout_add (&now, thread_alarm_handler, thread_current);
  }

  splx (s);
  return ret;
}
