/* $Id: io.cc,v 1.1 1998/10/21 18:33:06 fubobb Exp $ */

extern "C" {
#include <assert.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
}

#include "io.hh"

#undef FD_MAX
#ifdef HAVE_WIDE_SELECT
const FD_MAX = 0x10000;
#else /* !HAVE_WIDE_SELECT */
const FD_MAX = FD_SETSIZE;
#endif /* !HAVE_WIDE_SELECT */

mutex iodesc::fdt_lock;
int iodesc::fdt_size;
iodesc **iodesc::fdt;

selmask iodesc::wrm;
selmask iodesc::wwm;
selmask iodesc::srm;
selmask iodesc::swm;
selmask iodesc::sem;
selmask iodesc::orm;
selmask iodesc::owm;
selmask iodesc::oem;
condition iodesc::seldone;

static inline void *
zextend (void *mem, int oldbytes, int newbytes)
{
  assert (newbytes > oldbytes);
  mem = xrealloc (mem, newbytes);
  bzero ((char *) mem + oldbytes, newbytes - oldbytes);
  return mem;
}

iodesc *
iodesc::lookup (int fdn)
{
  if (fdn < 0 || fdn >= FD_MAX)
    return NULL;

  fdt_lock.lock ();

  /* Extend fdt if necessary */
  if (fdn >= fdt_size) {
    fdt = (iodesc **) zextend (fdt, sizeof (iodesc *) * fdt_size,
			       sizeof (iodesc *) * (fdn + 1));
    fdt_size = fdn + 1;
  }

  iodesc *f = fdt[fdn];
  if (!f) {
    struct stat sb;
    int saved_errno = errno;
    if (fstat (fdn, &sb) >= 0) {
      switch (sb.st_mode & S_IFMT) {
      case S_IFIFO:
      case S_IFSOCK:
      case S_IFCHR:		// Includes INET sockets in Solaris
	f = new iononblock (fdn);
	break;
      default:
	f = new iodesc (fdn);
	break;
      }
      fdt[fdn] = f;
    }
    else
      errno = saved_errno;
  }
  if (f && f->closed)
    f = NULL;
  if (f)
    f->refcnt++;

  fdt_lock.unlock ();
  return f;
}

void
iodesc::release ()
{
  fdt_lock.lock ();
  if (!--refcnt && closed)
    delete this;
  fdt_lock.unlock ();
}

/* Free a file descriptor number.  Does not close file! */
int
iodesc::freefdno ()
{
  fdt_lock.lock ();
  if (closed) {
    fdt_lock.unlock ();
    return -1;
  }
  closed = 1;
  fdt[fd] = NULL;
  wrm.clr (fd);
  wwm.clr (fd);
  fd = -1;
  readok.broadcast ();
  writeok.broadcast ();
  fdt_lock.unlock ();
  return 0;
}

terr
iodesc::readwait (timespec *tvp)
{
  autolock l(fdt_lock);
  if (fd >= 0) {
    wrm.set (fd);
    return readok.alertwait (fdt_lock, tvp);
  }
  return TFAILED;
}

terr
iodesc::writewait (timespec *tvp)
{
  autolock l(fdt_lock);
  if (fd >= 0) {
    wwm.set (fd);
    return writeok.alertwait (fdt_lock, tvp);
  }
  return TFAILED;
}

static inline int
__max (int a, int b)

{
  return ((a < b) ? b : a);
}

void
iodesc::checkio (timeval *tv)
{
  int sel_size;

  fdt_lock.lock ();

  orm = srm;
  orm.set (wrm);
  owm = swm;
  owm.set (wwm);
  oem = sem;

  sel_size = __max(__max(orm.size(), owm.size()), oem.size());

  int n = ::select (sel_size, orm.fdset (sel_size), owm.fdset (sel_size),
		    oem.fdset (sel_size), tv);

  for (int i = 0; n > 0 && i < sel_size; i++) {
    if (orm.isset (i)) {
      n--;
      if (wrm.isset (i)) {
	wrm.clr (i);
	fdt[i]->readok.broadcast ();
      }
    }
    if (owm.isset (i)) {
      n--;
      if (wwm.isset (i)) {
	wwm.clr (i);
	fdt[i]->writeok.broadcast ();
      }
    }

    if (oem.isset (i))
      n--;
  }
  if (n < 0) {
    assert (errno == EINTR);  // XXX
    // A signal happened. Somehow, we need to propogate this
    // to the waiters
    orm.zero();
    owm.zero();
    oem.zero();

  }

  srm.zero ();
  swm.zero ();
  sem.zero ();

  if (n < 0)
    seldone.alert();
  else
    seldone.broadcast ();

  fdt_lock.unlock ();
}

/* To be called from C code */
extern void
thread_select (timeval *tv)
{
  iodesc::checkio (tv);
}

/* XXX - need to check value of tv */
int
iodesc::select (int nfds, fd_set *rfds, fd_set *wfds,
		fd_set *efds, timeval *tvp)
{
  static timeval ztv;
  timespec ts;
  selmask r(rfds, nfds);
  selmask w(wfds, nfds);
  selmask e(efds, nfds);
  int n;

  /* Run non-blocking select immediately, not just to get a quick
   * response if one is there, but also to verify the validity of all
   * the arguments.  */
  if ((n = ::select (nfds, r.fdset(nfds), w.fdset(nfds),
		     e.fdset(nfds), &ztv))
      || (tvp && !tvp->tv_sec && !tvp->tv_usec))
    goto leave;

  /* Okey, when should we wake up in real time... */
  if (tvp) {
    clock_gettime (CLOCK_REALTIME, &ts);
    ts.tv_sec += tvp->tv_sec;
    ts.tv_nsec += 1000 * tvp->tv_usec;
    if (ts.tv_nsec > 1000000000) {
      ts.tv_nsec -= 1000000000;
      ts.tv_sec += 1;
    }
  }

  r = selmask (rfds, nfds);
  w = selmask (wfds, nfds);
  e = selmask (efds, nfds);

  fdt_lock.lock ();
  for (;;) {
    srm.set (r);
    swm.set (w);
    sem.set (e);
    terr ret = seldone.alertwait (fdt_lock, tvp ? &ts : NULL);
    if (ret == TTIMEDOUT) {
      n = 0;
      break;
    } else if (ret == TALERTED) {
      errno = EINTR;
      n = -1;
      break;
    } else if (orm.isset (r) || owm.isset (w) || oem.isset (e)) {
      r.clrnot (orm);
      w.clrnot (owm);
      e.clrnot (oem);
      n = r.nset () + w.nset () + e.nset ();
      break;
    }
  }
  fdt_lock.unlock ();

 leave:
  int nbytes = (nfds + 7) / 8;
  if (n > 0) {
    if (rfds)
      memcpy (rfds, r.fdset (nfds), nbytes);
    if (wfds)
      memcpy (wfds, w.fdset (nfds), nbytes);
    if (efds)
      memcpy (efds, e.fdset (nfds), nbytes);
  } else if (n == 0) {
    if (rfds)
      bzero (rfds, nbytes);
    if (wfds)
      bzero (wfds, nbytes);
    if (efds)
      bzero (efds, nbytes);
  }

  return n;
}


/*
 * Redefinitions of I/O system calls
 */
#define IOCALL(rettype, call, fd, args)		\
{						\
  iodesc *f = iodesc::lookup (fd);		\
  rettype ret;					\
  if (!f) {					\
    errno = EINVAL;				\
    return -1;					\
  }						\
  int saved_errno = errno;			\
  ret = f->call args;				\
  f->release ();				\
  if (ret != -1)				\
    errno = saved_errno;			\
  return ret;					\
}

extern "C" {

inline ssize_t
upt_writev (int fd, const struct iovec *iov, int iovcnt)
{
  IOCALL (ssize_t, writev, fd, (iov, iovcnt));
}
inline ssize_t
upt_readv (int fd, const struct iovec *iov, int iovcnt)
{
  IOCALL (ssize_t, readv, fd, (iov, iovcnt));
}
ssize_t
upt_read (int fd, void *buf, size_t nbites)
{
  struct iovec iov;
  iov.iov_base = (char *) buf;
  iov.iov_len = nbites;
  return upt_readv (fd, &iov, 1);
}
ssize_t
upt_write (int fd, const void *buf, size_t nbites)
{
  struct iovec iov;
  iov.iov_base = (char *) buf;
  iov.iov_len = nbites;
  return upt_writev (fd, &iov, 1);
}

ssize_t
upt_recvmsg (int fd, msghdr *msg, int flags)
{
  IOCALL (ssize_t, recvmsg, fd, (msg, flags));
}
ssize_t
upt_sendmsg (int fd, const msghdr *msg, int flags)
{
  IOCALL (ssize_t, sendmsg, fd, (msg, flags));
}

int
upt_accept (int fd, sockaddr *addr, int *addrlen)
{
  IOCALL (int, accept, fd, (addr, addrlen));
}
int
upt_connect (int fd, const sockaddr *addr, int addrlen)
{
  IOCALL (int, connect, fd, (addr, addrlen));
}

off_t
upt_lseek (int fd, off_t offset, int whence)
{
  IOCALL (off_t, lseek, fd, (offset, whence));
}

/* XXX - file descriptors we haven't seen yet? */
int
upt_close (int fdn)
{
  iodesc *f = iodesc::lookup (fdn);
  if (!f || f->freefdno () < 0) {
    errno = EINVAL;
    return -1;
  }
  f->release ();
  return ::close (fdn);
}

int
upt_select (int nfds, fd_set *rfds, fd_set *wfds, fd_set *efds, timeval *tv)
{
  return iodesc::select (nfds, rfds, wfds, efds, tv);
}

}
