/* $Id: xmalloc.cc,v 1.1 1998/10/21 18:33:40 fubobb Exp $ */

/* Threadsafe malloc routines for UPT, based on libc's probably
 * non-reentrant malloc.  The strategy here is always to call malloc
 * at splhigh.  We could insetad throw a global lock around all memory
 * operations, but this would in turn make it impossible to call
 * malloc at splhigh.  (A preempted malloc holding the global lock
 * might otherwise cause supposedly atomic code to sleep for that
 * lock.)
 *
 * Note that it's a bad idea to call malloc (or at least xmalloc) at
 * splhigh anyway, since we can't run the user's xmalloc_handler code
 * at splhigh, either.  (It might sleep, or do something slow for
 * which it should be preempted.)  As a compromise, we call out to
 * thread_spl_xmalloc_handler when xmalloc runs out of memory at IPL.
 * That function just aborts.  This ensures that we can call xmalloc
 * at splhigh in printf-like things for debugging and code and panic
 * code, which are two situations where an abort doesn't really
 * matter.  */

#include "private.h"

#undef realloc
#undef free
#undef malloc
#undef xrealloc
#undef xfree
#undef xmalloc
#undef new
#undef delete

#ifdef DMALLOC
dmalloc_t dmalloc;
#endif /* DMALLOC */

void (*xmalloc_handler) (void);

static volatile int xhctr;
static mutex xhlock;
static condition xhstart;
static condition xhdone;

void
thread_xmalloc_begin (void *)
{
  xhlock.lock ();
  splx (SPL_MEM);
  for (;;) {
    xhstart.wait (xhlock);
    if (!xmalloc_handler)
      thread_panic ("Out of memory with no xmalloc_handler.\n");
    xmalloc_handler ();
    xhctr++;
    xhdone.broadcast ();
  }
}

static void
outofmemory (int ipl)
{
  if (ipl & SPL_MEM) {
    thread_spl_xmalloc_handler ();
    return;
  }

  int lastctr = xhctr;
  xhlock.lock ();
  if (lastctr == xhctr) {
    xhstart.signal ();
    xhdone.wait (xhlock);
  }
  xhlock.unlock ();
}

#ifndef DMALLOC

extern "C" {
extern void *__malloc (int);
extern void *__realloc (void *, int);
extern void __free (void *);
}

void *
realloc (void *ptr, size_t size)
{
  void *ret;
  int s = splhigh ();

  if (!size) {
    if (ptr)
      __free (ptr);
    ret = NULL;
  }
  else
    ret = ptr ? __realloc (ptr, size) : __malloc (size);

  splx (s);
  return ret;
}

void *
xrealloc (void *ptr, size_t size)
{
  void *ret;
  int s = splhigh ();

  while (!(ret = realloc (ptr, size)) && size)
    outofmemory (s);

  splx (s);
  return ret;
}

void
free (void *ptr)
{
  realloc (ptr, 0);
}

void *
malloc (size_t size)
{
  return realloc (NULL, size);
}

#else /* DMALLOC */

/* Dmalloc doesn't like the ANSI C specifications for realloc and free
 * (for instance it won't let you free NULL pointers).  We fix that
 * here.  */
static inline void *
dorealloc (char *file, int line, void *ptr, size_t size)
{
  void *ret = NULL;
  int s = splhigh ();

  if (ptr && size)
    ret = _realloc_leap (file, line, ptr, size);
  else if (size)
    ret = _malloc_leap (file, line, size);
  else if (ptr)
    _free_leap (file, line, ptr);

  splx (s);
  return ret;
}

void *
threadsafe_realloc_leap (char *file, int line, void *ptr, size_t size)
{
  return dorealloc (file, line, ptr, size);
}

void *
threadsafe_xrealloc_leap (char *file, int line, void *ptr, size_t size)
{
  void *ret;
  int s = splhigh ();

  while (!(ret = dorealloc (file, line, ptr, size)) && size)
    outofmemory (s);

  splx (s);
  return ret;
}

void *
realloc_r (void *ptr, size_t size)
{
  return threadsafe_realloc_leap ("unknown file", 0, ptr, size);
}

void *
xrealloc (void *ptr, size_t size)
{
  return threadsafe_xrealloc_leap ("unknown file", 0, ptr, size);
}

void
free_r (void *ptr)
{
  realloc_r (ptr, 0);
}

void *
malloc_r (size_t size)
{
  return realloc_r (NULL, size);
}

void *
operator new (size_t size, dmalloc_t, char *file, int line)
{
  /* As per the C++ standard, allocating 0 bytes must not return NULL,
   * and must return a different pointer each time... */
  if (!size)
    size = 1;
  return threadsafe_xrealloc_leap (file, line, NULL, size);
}

void *
operator new[] (size_t size, dmalloc_t, char *file, int line)
{
  if (!size)
    size = 1;
  return threadsafe_xrealloc_leap (file, line, NULL, size);
}

#endif /* DMALLOC */

void
xfree (void *ptr)
{
  xrealloc (ptr, 0);
}
void *
xmalloc (size_t size)
{
  return xrealloc (NULL, size);
}

void *
operator new (size_t size)
{
  if (!size)
    size = 1;
  return xmalloc (size);
}

void *
operator new[] (size_t size)
{
  if (!size)
    size = 1;
  return xmalloc (size);
}

void
operator delete (void *ptr)
{
  xfree (ptr);
}

void
operator delete[] (void *ptr)
{
  xfree (ptr);
}
